!==========================================================================================!
!==========================================================================================!
!    This module contains subroutines and functions that will apply disturbances to        !
! patches.  This is usually done once a year, and the main disturbance driver will decide  !
! which kind of disturbance should be applied.                                             !
!------------------------------------------------------------------------------------------!
module disturbance_utils

   use ed_state_vars   , only : allocate_patchtype    & ! subroutine
                              , copy_patchtype        & ! subroutine
                              , deallocate_patchtype  & ! subroutine
                              , allocate_sitetype     & ! subroutine
                              , deallocate_sitetype   & ! subroutine
                              , copy_sitetype_mask    ! ! subroutine
   use fuse_fiss_utils , only : fuse_cohorts          & ! subroutine
                              , terminate_cohorts     & ! subroutine
                              , split_cohorts         ! ! subroutine
   !=======================================================================================!
   !=======================================================================================!


   contains



   !=======================================================================================!
   !=======================================================================================!
   !     This is the main disturbance driver.  It will be called every New Year day, and   !
   ! it will decide whether a new patch should be created.  Disturbances can be natural    !
   ! and anthropogenic, and the new patch will be always assigned an identification flag   !
   ! that will tell how that patch was created.  Three categories are currently possible:  !
   ! 1 - agriculture: conversion to agriculture by land clearing;                          !
   ! 2 - secondary forest: logging, land abandonment, and harvest create this patch;       !
   ! 3 - primary forest: natural disturbances (treefall or fire).                          !
   !---------------------------------------------------------------------------------------!
   subroutine apply_disturbances(cgrid)
     use phenology_startup, only: initialize_paw
      use ed_state_vars, only : edtype                  & ! structure
                              , polygontype             & ! structure
                              , sitetype                & ! structure
                              , patchtype               ! ! structure
      use ed_misc_coms , only : current_time            ! ! intent(in)
      use disturb_coms , only : min_new_patch_area      & ! intent(in)
                              , mature_harvest_age      & ! intent(in)
                              , plantation_rotation     & ! intent(in)
                              , ianth_disturb           & ! intent(in)
                              , time2canopy,fire_year   & ! intent(in)
                              , treefall_disturbance_rate
      use ed_max_dims  , only : n_dist_types            & ! intent(in)
                              , n_pft                   & ! intent(in)
                              , n_dbh                   ! ! intent(in)
      use mem_polygons , only : maxcohort               ! ! intent(in)
      use grid_coms    , only : nzg                     & ! intent(in)
                              , nzs                     ! ! intent(in)
      use pft_coms,      only: c2n_leaf                 &
                              ,c2n_structural           &
                              ,c2n_slow,resprout
      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(edtype)                   , target      :: cgrid
      !----- Local variables. -------------------------------------------------------------!
      type(polygontype)              , pointer     :: cpoly
      type(sitetype)                 , pointer     :: csite
      type(sitetype)                 , pointer     :: tsite
      type(patchtype)                , pointer     :: qpatch
      integer                                      :: ipy
      integer                                      :: isi
      integer                                      :: ipa
      integer                                      :: ipft
      integer                                      :: onsp
      integer                                      :: old_lu
      integer                                      :: new_lu
      integer                                      :: poly_dest_type
      logical, dimension(:)          , allocatable :: disturb_mask
      logical                                      :: ploughed
      logical                                      :: abandoned
      logical                                      :: natural
      logical                                      :: logged
      logical                                      :: is_plantation
      logical                                      :: mature_plantation
      logical                                      :: mature_primary
      logical                                      :: mature_secondary
      real   , dimension(n_pft,n_dbh)              :: initial_agb
      real   , dimension(n_pft,n_dbh)              :: initial_basal_area
      real   , dimension(n_pft)                    :: mindbh_harvest
      real                                         :: area
      real                                         :: area_fac
      real                                         :: dA
      real                                         :: elim_nplant
      real                                         :: elim_lai
      real          , dimension(n_pft)              :: n_resprout
      real          , dimension(n_pft)              :: nplant_resprout

     
      real :: plantN1, soilN1, totalN1, plantN2, soilN2,totalN2
      type(patchtype), pointer :: cpatch
      integer :: ico
      !------------------------------------------------------------------------------------!
      !----- Allocating the temporary site that will host the original patches. -----------!
      nullify(tsite)
      allocate(tsite)

      plantN1=0.;plantN2=0.;soilN1=0.;soilN2=0.;totalN1=0.;totalN2=0.;nplant_resprout(:) = 0.
      polyloop: do ipy = 1,cgrid%npolygons
         
         cpoly => cgrid%polygon(ipy)
         siteloop: do isi = 1,cpoly%nsites

            csite => cpoly%site(isi)


            do ipa = 1, csite%npatches

               cpatch => csite%patch(ipa)

               soilN1 = soilN1 + (csite%fast_soil_N(ipa) + csite%slow_soil_N(ipa) +       &!ATT
                        csite%structural_soil_C(ipa)/c2n_structural +                     &
                        csite%mineralized_soil_N(ipa)) * csite%area(ipa)    

               do ico = 1, cpatch%ncohorts
                  plantN1 = plantN1 + csite%area(ipa) * cpatch%nplant(ico) *               &
                            (cpatch%balive(ico)/c2n_leaf(cpatch%pft(ico)) +                &
                            cpatch%bdead(ico)/c2n_structural + cpatch%nstorage(ico))      !JL
               enddo
            enddo
           
            !----- Store AGB, basal area profiles in memory. ------------------------------!
            call update_site_derived_props(cpoly, 1,isi)
            initial_agb(1:n_pft,1:n_dbh)        = cpoly%agb(1:n_pft,1:n_dbh,isi)
            initial_basal_area(1:n_pft,1:n_dbh) = cpoly%basal_area(1:n_pft,1:n_dbh,isi)

            !------------------------------------------------------------------------------!
            !      First take care of harvesting, i.e., secondary -> secondary and         !
            ! primary -> secondary.                                                        !
            !------------------------------------------------------------------------------!
            call apply_forestry(cpoly,isi, current_time%year)

            !----- Update the cut output variables. ---------------------------------------!
            call update_site_derived_props(cpoly, 1,isi)
            cpoly%agb_cut(1:n_pft,1:n_dbh,isi) = cpoly%agb_cut(1:n_pft, 1:n_dbh,isi)       &
                                               + initial_agb(1:n_pft, 1:n_dbh)             &
                                               - cpoly%agb(1:n_pft, 1:n_dbh,isi)

            cpoly%basal_area_cut(1:n_pft,1:n_dbh,isi) =                                    &
                                                 cpoly%basal_area_cut(1:n_pft,1:n_dbh,isi) &
                                               + initial_basal_area(1:n_pft,1:n_dbh)       &
                                               - cpoly%basal_area(1:n_pft,1:n_dbh,isi)

            !----- Save the Original Number (of) Site Patches, onsp... --------------------!
            onsp = csite%npatches


            !------------------------------------------------------------------------------!
            !     Create a temporary site with vectors containing all current patches as   !
            ! well as n_dist_types patches.  Create the newly disturbed patches in here,   !
            ! and depending on how many are created, repopulate the existing site's patch  !
            ! vectors.                                                                     !
            !------------------------------------------------------------------------------!
            call allocate_sitetype(tsite,onsp)

            allocate(disturb_mask(onsp + n_dist_types))
            disturb_mask         = .false.
            disturb_mask(1:onsp) = .true.

            !------------------------------------------------------------------------------!
            !     Transfer the origial patch values into the front end of the temp's       !
            ! space.                                                                       !
            !------------------------------------------------------------------------------!
            call copy_sitetype_mask(csite,tsite,disturb_mask(1:onsp),count(disturb_mask)   &
                                   ,count(disturb_mask))

            !----- Reallocate and transfer them back. -------------------------------------!
            call deallocate_sitetype(csite)
            call allocate_sitetype(csite,onsp + n_dist_types)
            call copy_sitetype_mask(tsite,csite,disturb_mask(1:onsp),count(disturb_mask)   &
                                   ,count(disturb_mask))
            call deallocate_sitetype(tsite)

            !------------------------------------------------------------------------------!
            !      Initialize all the potential as well as implemented disturbance         !
            ! patches.  n_dist_types new patches will be created, each one containing a    !
            ! different patch type.  In case no conversion to that kind of patch has       !
            ! happened, or if the newly created patch is tiny, it will be removed soon.    !                                                                   !
            !------------------------------------------------------------------------------!
            do new_lu = onsp+1, onsp+n_dist_types
               call initialize_disturbed_patch(csite,cpoly%met(isi)%atm_tmp,new_lu,1       &
                                              ,cpoly%lsl(isi))
            end do
            !----- Loop over q, the *destination* landuse type. ---------------------------!
            new_lu_loop: do new_lu = 1, n_dist_types
               !----- Set up area to zero, in case no conversion happens. -----------------!
               area = 0.0
               do ipa=1,onsp
                  !----- Save the old land use in a shorter variable for convenience. -----!
                  old_lu        = csite%dist_type(ipa)
                  is_plantation = csite%plantation(ipa) == 1
                  !------------------------------------------------------------------------!
                  !    Now we add the area associated with each kind of possible disturb-  !
                  ! ance that can happen.  Types of conversion that are solved here are:   !
                  ! * ploughed  - conversion from primary/secondary land to agriculture.   !
                  ! * abandoned - conversion from agriculture to secondary land.           !
                  ! * natural   - natural disturbance from primary/secondary land to       !
                  !               primary land (fires or tree fall)                        !
                  ! * logged    - conversion from primary/secondary land to secondary      !
                  !               land due to logging.                                     !
                  !------------------------------------------------------------------------!
                  ploughed  = new_lu == 1 .and. old_lu /= 1
                  abandoned = new_lu == 2 .and. old_lu == 1
                  !----- Natural disturbance, either trees are old or there is a fire. ----!
                  natural   = new_lu == 3 .and. old_lu /= 1 !.and.                          &
                             ! ( (fire_year))
                  print*,'fire_year1',fire_year,'natural',natural,'new_lu',new_lu
                  if(natural)then
                     cpoly%disturbance_rates(new_lu,old_lu,isi)  = treefall_disturbance_rate
                  endif
                  

                  !----- Check whether the patch is ready  be harvested. ------------------!
                  mature_primary    = old_lu == 3                                  .and.   &
                                      csite%age(ipa)         > mature_harvest_age
                  mature_plantation = is_plantation                                .and.   &
                                      csite%age(ipa)         > plantation_rotation
                  mature_secondary  = old_lu == 2 .and. (.not. is_plantation)      .and.   &
                                      csite%age(ipa)         > mature_harvest_age
                  !------------------------------------------------------------------------!
                  logged    = new_lu == 2 .and.                                            &
                              ( mature_primary .or. mature_plantation .or.                 &
                                mature_secondary)

                  !------------------------------------------------------------------------!
                  !    Add area if any of the disturbances that produce of type new_lu has !
                  ! happened.  The ones that produce type other than new_lu will be always !
                  ! false for new_lu.                                                      !
                  !------------------------------------------------------------------------!
                  if  (ploughed .or. abandoned .or. natural .or. logged) then
                     dA = csite%area(ipa)                                                  &
                        * (1. - exp(- ( cpoly%disturbance_rates(new_lu,old_lu,isi)         &
                                      + cpoly%disturbance_memory(new_lu,old_lu,isi) ) ) )
                     area = area + dA
                     print*,'cpoly%disturbance_rates(new_lu,old_lu,isi)',cpoly%disturbance_rates(new_lu,old_lu,isi)
                     print*,'cpoly%disturbance_memory(new_lu,old_lu,isi)',cpoly%disturbance_memory(new_lu,old_lu,isi)
                     
                  end if
                  !------------------------------------------------------------------------!


               end do
              ! print*,'area',area,'min_new_patch_area',min_new_patch_area
               if (area > min_new_patch_area) then
                  write(unit=*,fmt='(a,1x,es12.5,1x,a,1x,i5)')                             &
                      ' ---> Making new patch, with area=',area,' for dist_type=',new_lu

                  !------------------------------------------------------------------------!
                  !     Set the flag that this patch should be kept as a newly created     !
                  ! transition patch.                                                      !
                  !------------------------------------------------------------------------!
                  disturb_mask(onsp+new_lu)     = .true.
  
                  csite%dist_type(onsp+new_lu)  = new_lu
                  csite%plantation(onsp+new_lu) = 0
                  csite%area(onsp+new_lu)       = area
                  
                  !----- Initialize to zero the new trasitioned patches. ------------------!
                  call initialize_disturbed_patch(csite,cpoly%met(isi)%atm_tmp,onsp+new_lu &
                                                 ,1,cpoly%lsl(isi))

                  !----- Now go through patches, adding its contribution to the new patch. !
                  do ipa=1,onsp
                     !----- Save the old land use in a shorter variable for convenience. --!
                     cpatch => csite%patch(ipa)
                     old_lu        = csite%dist_type(ipa)
                     is_plantation = csite%plantation(ipa) == 1

                     !----- Check whether this patch can be disturbed. --------------------!
                     ploughed  = new_lu == 1 .and. old_lu /= 1
                     abandoned = new_lu == 2 .and. old_lu == 1
                     natural   = new_lu == 3 .and. old_lu /= 1 !.and.                       &
                         ! ( ( fire_year) )
                                !       print*,'fire_year2',fire_year,'natural',natural,'new_lu',new_lu,'old_lu',old_lu

                     !----- Check whether the patch is ready  be harvested. ---------------!
                     mature_primary    = old_lu == 3                                 .and. &
                                         csite%age(ipa)         > mature_harvest_age
                     mature_plantation = is_plantation                               .and. &
                                         csite%age(ipa)         > plantation_rotation
                     mature_secondary  = old_lu == 2 .and. (.not. is_plantation)     .and. &
                                         csite%age(ipa)         > mature_harvest_age
                     !---------------------------------------------------------------------!
                     logged    = new_lu == 2 .and.                                         &
                                 ( mature_primary .or. mature_plantation .or.              &
                                   mature_secondary)

                     !---------------------------------------------------------------------!
                     !     Adjust some information to be sent to the disturbance routine.  !
                     !---------------------------------------------------------------------!
                     if (natural) then
                        poly_dest_type          = cpoly%nat_dist_type(isi)
                        mindbh_harvest(1:n_pft) = huge(1.)
                     elseif (logged .and. mature_primary) then
                        poly_dest_type          = 2
                        mindbh_harvest(1:n_pft) = cpoly%mindbh_primary(1:n_pft,isi)
                     elseif (logged) then
                        poly_dest_type          = 2
                        mindbh_harvest(1:n_pft) = cpoly%mindbh_secondary(1:n_pft,isi)
                     else
                        poly_dest_type          = 0
                        mindbh_harvest(1:n_pft) = huge(1.)
                     end if
                     !---------------------------------------------------------------------!



                     !---------------------------------------------------------------------!
                     !    If the patch is going to be disturbed, compute the area of the   !
                     ! disturbed patch to be added to the new destination patch and update !
                     ! the litter layer.                                                   !
                     !---------------------------------------------------------------------!
                     if (ploughed .or. abandoned .or. natural .or. logged) then
                        dA  = csite%area(ipa)                                              &
                            * (1. - exp(- ( cpoly%disturbance_rates(new_lu,old_lu,isi)     &
                                       + cpoly%disturbance_memory(new_lu,old_lu,isi) ) ) ) 

                        area_fac = dA / csite%area(onsp+new_lu)
                      
                        call increment_patch_vars(csite,new_lu+onsp,ipa,area_fac)

                        !ATT combined accum_dist_litt with insert_survivors so that survivorship()
                        !didnt have to be called seperately

                        print*,'IPA',ipa,'onsp',onsp

                        call insert_survivors(csite,new_lu+onsp,ipa,new_lu,area_fac              &
                             ,poly_dest_type,mindbh_harvest,cpoly%loss_fraction(new_lu,isi),cpoly%lsl(isi),n_resprout)

                        
                        ! call accum_dist_litt(csite,new_lu+onsp,ipa,new_lu,area_fac         &
                        !                    ,cpoly%loss_fraction(new_lu,isi)               &
                         !                   ,poly_dest_type,mindbh_harvest,survivors)
                       
                        !----- Update patch area. -----------------------------------------!
                        csite%area(ipa) = csite%area(ipa) - dA
                        do ipft=1,n_pft
                          ! print*,'nplant_resprout(ipft)',nplant_resprout(ipft),'n_resprout(ipft)',n_resprout(ipft)
                           !print*,'csite%area(ipa)',csite%area(ipa),'dA',dA
                           nplant_resprout(ipft) = nplant_resprout(ipft) + n_resprout(ipft) * dA
                        enddo
                     end if
                  end do

                  !------------------------------------------------------------------------!
                  !      Update temperature and density.  This must be done before plant-  !
                  ! ing, since the leaf temperature is initially assigned as the canopy    !
                  ! air temperature.                                                       !
                  !------------------------------------------------------------------------!
                  call update_patch_thermo_props(csite,new_lu+onsp,new_lu+onsp,nzg,nzs)         
                  !----- If the new patch is agriculture, plant it with grasses. ----------!
                  call initialize_paw(csite,new_lu+onsp,csite%ntext_soil(:,new_lu+onsp))           

                  if(.false.)then
                  if (fire_year) then
                     do ipft=1,n_pft
                        if(resprout(ipft) .and. nplant_resprout(ipft) > 0.)then
                           print*,'Resprouting, pft:',ipft,', at density:', nplant_resprout(ipft)
                           call plant_patch(csite,new_lu+onsp,new_lu+onsp,nzg,ipft   &
                                ,nplant_resprout(ipft)                     &
                                ,csite%ntext_soil(:,new_lu+onsp)                              &
                                ,csite%green_leaf_factor(:,new_lu+onsp), 1.0, cpoly%lsl(isi))
                        endif
                     enddo
                  end if
                  endif
                  qpatch => csite%patch(new_lu+onsp)

                  !----- Fuse then terminate cohorts. -------------------------------------!
                  if (csite%patch(new_lu+onsp)%ncohorts > 0 .and. maxcohort >= 0) then
                     call fuse_cohorts(csite,new_lu+onsp,csite%green_leaf_factor(:,new_lu+onsp)    &
                                      ,cpoly%lsl(isi))
                     call terminate_cohorts(csite,new_lu+onsp,elim_nplant,elim_lai)
                     call split_cohorts(qpatch,csite%green_leaf_factor(:,new_lu+onsp)              &
                                       ,cpoly%lsl(isi))
                  end if
              
                  !----- Store AGB, basal area profiles in memory. ------------------------!
                  initial_agb(1:n_pft,1:n_dbh) = cpoly%agb(1:n_pft, 1:n_dbh,isi)
                  initial_basal_area(1:n_pft,1:n_dbh) =                                    &
                                                      cpoly%basal_area(1:n_pft,1:n_dbh,isi)

                  !------------------------------------------------------------------------!
                  !     Update the derived properties including veg_height, and patch-     !
                  ! -level LAI, WAI, WPA.                                                  !
                  !------------------------------------------------------------------------!
                  call update_patch_derived_props(csite,cpoly%lsl(isi),cpoly%met(isi)%prss &
                                                 ,new_lu+onsp)
                  !----- Update soil temperature, liquid fraction, etc. -------------------!

                  call new_patch_sfc_props(csite,new_lu+onsp,nzg,nzs                       &
                                          ,csite%ntext_soil(:,new_lu+onsp))                        !
                  !----- Update budget properties. ----------------------------------------!
                  call update_budget(csite,cpoly%lsl(isi),new_lu+onsp,new_lu+onsp)

                  !----- Update AGB, basal area. ------------------------------------------!
                  call update_site_derived_props(cpoly,1,isi)


                  !----- Update either cut or mortality. ----------------------------------!
                  if (new_lu /= 3) then
                     cpoly%agb_cut(1:n_pft,1:n_dbh,isi) =                                  &
                            cpoly%agb_cut(1:n_pft, 1:n_dbh,isi)                            &
                          + initial_agb(1:n_pft, 1:n_dbh)                                  &
                          - cpoly%agb(1:n_pft, 1:n_dbh,isi)
                     cpoly%basal_area_cut(1:n_pft, 1:n_dbh,isi) =                          &
                            cpoly%basal_area_cut(1:n_pft, 1:n_dbh,isi)                     &
                          + initial_basal_area(1:n_pft, 1:n_dbh)                           &
                          - cpoly%basal_area(1:n_pft, 1:n_dbh,isi)
                  else
                     cpoly%agb_mort(1:n_pft,1:n_dbh,isi) =                                 &
                            cpoly%agb_mort(1:n_pft,1:n_dbh,isi)                            &
                          + initial_agb(1:n_pft,1:n_dbh)                                   &
                          - cpoly%agb(1:n_pft,1:n_dbh,isi)
                     cpoly%basal_area_mort(1:n_pft, 1:n_dbh,isi) =                         &
                            cpoly%basal_area_mort(1:n_pft, 1:n_dbh,isi)                    &
                          + initial_basal_area(1:n_pft, 1:n_dbh)                           &
                          - cpoly%basal_area(1:n_pft, 1:n_dbh,isi)
                  endif
                  
                  !----- Clear the disturbance memory for this disturbance type. ----------!
                  cpoly%disturbance_memory(new_lu,1:n_dist_types,isi) = 0.0

               elseif(area > 0.0)then
                  !------------------------------------------------------------------------!
                  !     The patch creation has been skipped because the area was too       !
                  ! small.  Put the current disturbance rates in memory to be added at the !
                  ! next timestep.                                                         !
                  !------------------------------------------------------------------------!
                  cpoly%disturbance_memory(new_lu,1:n_dist_types,isi) =                    &
                         cpoly%disturbance_memory(new_lu,1:n_dist_types,isi)               &
                       + cpoly%disturbance_rates(new_lu,1:n_dist_types,isi)
               end if
            end do new_lu_loop

            !------------------------------------------------------------------------------!
            !      Reallocate the current site to fit the original patches and whatever    !
            ! was generated in disturbance (ie, make it non sparse).  Populate the         !
            ! original site with both the modified original patches, and the newly created !
            ! patches.  The index of all of these are disturb_mask.  This mask should be   !
            ! ones for all original patches, and sparse from there after.                  !
            !------------------------------------------------------------------------------!
            call allocate_sitetype(tsite,count(disturb_mask))
            call copy_sitetype_mask(csite,tsite,disturb_mask,size(disturb_mask)            &
                                   ,count(disturb_mask))
            call deallocate_sitetype(csite)
            call allocate_sitetype(csite,count(disturb_mask))
          
            disturb_mask = .false.
            disturb_mask(1:csite%npatches) = .true.
            call copy_sitetype_mask(tsite,csite,disturb_mask(1:csite%npatches)             &
                                   ,count(disturb_mask),count(disturb_mask))
          
            call deallocate_sitetype(tsite)
            deallocate(disturb_mask)
            !------------------------------------------------------------------------------!

            do ipa = 1, csite%npatches

               cpatch => csite%patch(ipa)

               soilN2 = soilN2 + (csite%fast_soil_N(ipa) + csite%slow_soil_N(ipa) +         &!ATT
                        csite%structural_soil_C(ipa)/c2n_structural +                       &
                        csite%mineralized_soil_N(ipa)) * csite%area(ipa)
                    
               do ico = 1, cpatch%ncohorts
                  plantN2 = plantN2 + csite%area(ipa) * cpatch%nplant(ico) *                &
                           (cpatch%balive(ico)/c2n_leaf(cpatch%pft(ico)) +                  &
                           cpatch%bdead(ico)/c2n_structural  + cpatch%nstorage(ico)) !JL!
               enddo
            enddo

!            print*,'NITROGREN, FINAL',soilN2,plantN2,soilN2+plantN2
          
         end do siteloop
      end do polyloop

      !----- Free memory before leaving... ------------------------------------------------!
      deallocate(tsite)
      return
   end subroutine apply_disturbances
   !=======================================================================================!
   !=======================================================================================!






   !=======================================================================================!
   !=======================================================================================!
   subroutine site_disturbance_rates(month, year, cgrid)

      use ed_state_vars, only : edtype                    & ! structure
                              , polygontype               & ! structure
                              , sitetype                  & ! structure
                              , patchtype                 ! ! structure
      use disturb_coms , only : treefall_disturbance_rate & ! intent(in)
                              , lutime                    & ! structure
                              , ianth_disturb             & ! intent(in)
                              , include_fire              & ! intent(in)
                              , plantation_rotation       & ! intent(in)
                              , mature_harvest_age        ! ! intent(in)
      use pft_coms     , only : agf_bs                    ! ! intent(in)
      use ed_max_dims  , only : n_pft                     ! ! intent(in)

      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(edtype)     , target           :: cgrid
      integer          , intent(in)       :: month
      integer          , intent(in)       :: year
      !----- Local variables. -------------------------------------------------------------!
      type(polygontype), pointer          :: cpoly
      type(sitetype)   , pointer          :: csite
      type(patchtype)  , pointer          :: cpatch
      type(lutime)     , pointer          :: clutime
      integer                             :: ipy
      integer                             :: isi
      integer                             :: ipa
      integer                             :: ico
      integer                             :: ipft
      integer                             :: iyear
      integer                             :: useyear
      logical                             :: mature_plantation
      logical                             :: mature_secondary
      real                                :: weight
      real                                :: sumweight
      real                                :: pharvest
      real                                :: fire_dist_rate
      !----- Local constants. -------------------------------------------------------------!
      real             , parameter        :: max_pharvest = 1.-2.*epsilon(1.)
      !------------------------------------------------------------------------------------!


      !------------------------------------------------------------------------------------!
      !      Loop over sites and polygons.                                                 !
      !------------------------------------------------------------------------------------!
      polyloop: do ipy = 1,cgrid%npolygons
         cpoly => cgrid%polygon(ipy)

         siteloop: do isi = 1,cpoly%nsites
            csite => cpoly%site(isi)

            !------ Calculate fire disturbance rates only when fire is on. ----------------!
            select case (include_fire)
            case (0) 
               fire_dist_rate = 0.0
            case (1,2)
               fire_dist_rate = sum(cpoly%lambda_fire(1:12,isi)) / 12.0
            end select
            cpoly%fire_disturbance_rate(isi) = fire_dist_rate
            !------------------------------------------------------------------------------!


            !------------------------------------------------------------------------------!
            !      For natural disturbance, use largest disturbance mode.                  !
            !------------------------------------------------------------------------------!
            if (fire_dist_rate > treefall_disturbance_rate) then
               cpoly%nat_disturbance_rate(isi) = fire_dist_rate
               cpoly%nat_dist_type(isi)        = 1
            else
               cpoly%nat_disturbance_rate(isi) = treefall_disturbance_rate
               cpoly%nat_dist_type(isi)        = 0
            end if

            !----- Set disturbance rates assuming only natural disturbance. ---------------!
            cpoly%disturbance_rates(1:2,1:3,isi) = 0.0
            cpoly%disturbance_rates(3,1,isi)     = 0.0
            cpoly%disturbance_rates(3,2:3,isi)   = cpoly%nat_disturbance_rate(isi)

            !------------------------------------------------------------------------------!
            !      Now it is time for anthropogenic disturbance rates.  We no longer need  !
            ! to check the year because all years in this simulation are assigned          !
            ! prescribed disturbance rates (even if that means zero disturbance).          !
            !------------------------------------------------------------------------------!
            useyear = cpoly%num_landuse_years(isi)
            !----- Loop over years. -------------------------------------------------------!
            find_lu_year: do iyear = 1,cpoly%num_landuse_years(isi)
               
               if (year == cpoly%clutimes(iyear,isi)%landuse_year) then
                  useyear = iyear
                  exit find_lu_year
               end if
            end do find_lu_year

            !------------------------------------------------------------------------------!
            !    Update land-use transition matrix.  The matrix (and array for each        !
            ! site) has two indices.  The leftmost one is the destination, and the         !
            ! middle one is the source.  Land use types are:                               !
            !    - 1. Agriculture                                                          !
            !    - 2. Secondary forest                                                     !
            !    - 3. Primary forest                                                       !
            !------------------------------------------------------------------------------!
            clutime => cpoly%clutimes(useyear,isi)

            !----- Agriculture to agriculture (1 => 1). -----------------------------------!
            cpoly%disturbance_rates(1,1,isi) = 0.0
            !----- Secondary forest to agriculture (2 => 1). ------------------------------!
            cpoly%disturbance_rates(1,2,isi) = clutime%landuse(7) + clutime%landuse(9)
            !----- Primary forest to agriculture (3 => 1). --------------------------------!
            cpoly%disturbance_rates(1,3,isi) = clutime%landuse(4) + clutime%landuse(5)

            !------------------------------------------------------------------------------!
            !     Agriculture to secondary land (1 => 2).   Here we also account for       !
            ! transitions to primary land due to abandonment.  In ED definition, a land    !
            ! can is considered primary only when the last disturbance.                    ! 
            !------------------------------------------------------------------------------!
            cpoly%disturbance_rates(2,1,isi) = clutime%landuse(8) + clutime%landuse(10)    &
                                             + clutime%landuse(3) + clutime%landuse(6)

            !------------------------------------------------------------------------------!
            !     Secondary forest to secondary forest (2 => 2).  Here we must convert the !
            ! harvest probability of being cut given that the DBH exceeds the minimum DBH. !
            ! This is done only when anthropogenic disturbance is on and we are not seek-  !
            ! ing the biomass target, otherwise we set it to zero.                         !
            !------------------------------------------------------------------------------!
            sumweight = 0.
            pharvest  = 0.
            if (ianth_disturb == 1 .and. clutime%landuse(12) < 0) then
               patchloop: do ipa=1,csite%npatches
                  cpatch => csite%patch(ipa)

                  !------------------------------------------------------------------------!
                  !    Check whether anything is going to be harvested from this patch.    !
                  !------------------------------------------------------------------------!
                  mature_plantation = csite%plantation(ipa) == 1                   .and.   &
                                      csite%age(ipa)         > plantation_rotation
                  mature_secondary  = csite%dist_type(ipa)  == 2                   .and.   &
                                      csite%plantation(ipa) /= 1                   .and.   &
                                      csite%age(ipa)         > mature_harvest_age
                  !------------------------------------------------------------------------!



                  !------------------------------------------------------------------------!
                  !     If this patch is secondary and mature, compute the weighted aver-  !
                  ! age of the probability of harvest, using the number of plants as the   !
                  ! weight.                                                                !
                  !------------------------------------------------------------------------!
                  if (mature_plantation .or. mature_secondary) then
                     cohortloop: do ico=1,cpatch%ncohorts
                        ipft = cpatch%pft(ico)
                        if (cpatch%dbh(ico) >= cpoly%mindbh_secondary(ipft,isi)) then
                           weight    = cpatch%nplant(ico) * csite%area(ipa)
                           pharvest  = pharvest                                            &
                                     + cpoly%probharv_secondary(ipft,isi) * weight
                           sumweight = sumweight + weight
                        end if  
                     end do cohortloop
                  end if
                  !----- Normalise the probability, unless it's zero. ---------------------!
                  if (sumweight > 0.) then
                     pharvest = min(max_pharvest,pharvest/sumweight)
                  else
                     pharvest = 0.
                  end if
               end do patchloop
            end if
            cpoly%disturbance_rates(2,2,isi) = -0.5 * log(1.0 - pharvest)
            !------------------------------------------------------------------------------!



            !------------------------------------------------------------------------------!
            !     Primary forest to secondary forest (3 => 2).  Here we must decide        !
            ! whether we want a biomass target harvesting, or a selective logging.  When   !
            ! harvesting is biomass-based, then we set the disturbance rate to zero,       !
            ! otherwise we read the disturbance rate.                                      !
            !------------------------------------------------------------------------------!
            if (clutime%landuse(12) > 0.) then
               cpoly%disturbance_rates(2,3,isi) = 0.0
            else
               cpoly%disturbance_rates(2,3,isi) = clutime%landuse(11)
            end if
            !------------------------------------------------------------------------------!


            !------------------------------------------------------------------------------!
            !    Agriculture to primary forest (1 => 3).  This must be zero, even though   !
            ! the landuse(3) and landuse(6) may not be (these will be added to the         !
            ! secondary instead).  This is because abandoned land is always considered     !
            ! secondary forest since the "disturbance" that creates this patch is not      !
            ! natural.                                                                     !
            !------------------------------------------------------------------------------!
            cpoly%disturbance_rates(3,1,isi) = 0.0


            !------------------------------------------------------------------------------!
            !    Secondary forest to primary forest (2 => 3).  This is the natural         !
            ! disturbance rate, because the definition of primary or secondary forest is   !
            ! defined based on the last disturbance.                                       !
            !------------------------------------------------------------------------------!
            cpoly%disturbance_rates(3,2,isi)= cpoly%nat_disturbance_rate(isi)
            !----- Primary forest to primary forest (3 => 3).  Natural disturbance. -------!
            cpoly%disturbance_rates(3,3,isi)= cpoly%nat_disturbance_rate(isi)
            !------------------------------------------------------------------------------!

            !------------------------------------------------------------------------------!
            !      Fraction of above ground litter from disturbance that is removed from   !
            ! patch.                                                                       !
            !------------------------------------------------------------------------------!
            cpoly%loss_fraction(1,isi) = agf_bs
            cpoly%loss_fraction(2,isi) = agf_bs
            cpoly%loss_fraction(3,isi) = 0.0
            !------------------------------------------------------------------------------!

         end do siteloop
      end do polyloop

      return
   end subroutine site_disturbance_rates
   !=======================================================================================!
   !=======================================================================================!






   !=======================================================================================!
   !=======================================================================================!
   !     This subroutine assigns initial conditions to a patch that has been disturbed     !
   !---------------------------------------------------------------------------------------!
   subroutine initialize_disturbed_patch(csite,atm_tmp,np,dp,lsl)

      use ed_state_vars, only : sitetype  & ! structure
                              , patchtype ! ! structure
      use consts_coms  , only : t3ple     ! ! intent(in)
      use grid_coms    , only : nzs       & ! intent(in)
                              , nzg       ! ! intent(in)

      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(sitetype), target      :: csite
      real          , intent(in)  :: atm_tmp
      integer       , intent(in)  :: np
      integer       , intent(in)  :: dp
      integer       , intent(in)  :: lsl
      !----- Local variables. -------------------------------------------------------------!
      integer                     :: k
      !------------------------------------------------------------------------------------!


      !----- Start with an empty patch.  Surviving cohorts will be added later. -----------!
      csite%patch(np)%ncohorts = 0

      !----- Initialise the surface water diagnostic variables with some defaults. --------!
      do k=1,nzs
         csite%sfcwater_tempk(k,np) = atm_tmp
         if (atm_tmp >= t3ple) then
            csite%sfcwater_fracliq(k,np) = 1.0
         else
            csite%sfcwater_fracliq(k,np) = 1.0
         end if
      end do

      !------------------------------------------------------------------------------------!
      !     Initialise most variables, except dist_type, plantation, and area, which will  !
      ! be defined outside this subroutine.  Most of the following variables will receive  !
      ! properties from the donor patches.                                                 !
      !------------------------------------------------------------------------------------!
      csite%age                        (np) = 0.0
      csite%fast_soil_C                (np) = 0.0
      csite%slow_soil_C                (np) = 0.0
      csite%structural_soil_C          (np) = 0.0
      csite%structural_soil_L          (np) = 0.0
      csite%mineralized_soil_N         (np) = 0.0
      csite%fast_soil_N                (np) = 0.0
      csite%slow_soil_N                (np) = 0.0!ATT
      csite%sum_dgd                    (np) = 0.0
      csite%sum_chd                    (np) = 0.0
      csite%can_depth                  (np) = 0.0
      csite%can_theta                  (np) = 0.0
      csite%can_theiv                  (np) = 0.0
      csite%can_prss                   (np) = 0.0
      csite%can_shv                    (np) = 0.0
      csite%can_co2                    (np) = 0.0
      csite%ggbare                     (np) = 0.0
      csite%ggveg                      (np) = 0.0
      csite%soil_energy          (1:nzg,np) = 0.0
      csite%soil_water           (1:nzg,np) = 0.0
      csite%sfcwater_mass        (1:nzs,np) = 0.0
      csite%sfcwater_energy      (1:nzs,np) = 0.0
      csite%sfcwater_depth       (1:nzs,np) = 0.0
      csite%rough                      (np) = 0.0
      csite%fsc_in                     (np) = 0.0
      csite%slsc_in                    (np) = 0.0
      csite%slsn_in                    (np) = 0.0
      csite%ssc_in                     (np) = 0.0
      csite%ssl_in                     (np) = 0.0
      csite%fsn_in                     (np) = 0.0
      csite%total_plant_nitrogen_uptake(np) = 0.0
      !------------------------------------------------------------------------------------!

      !----- Initialise all fast variables. -----------------------------------------------!
      call init_ed_patch_vars(csite,np,np,lsl)

      return
   end subroutine initialize_disturbed_patch
   !=======================================================================================!
   !=======================================================================================!






   !=======================================================================================!
   !=======================================================================================!
   !     This subroutine will include the contribution of each contributing patch (cp) to  !
   ! the new, disturbed patch (np).                                                        !
   !---------------------------------------------------------------------------------------!
   subroutine increment_patch_vars(csite,np, cp, area_fac)
      use ed_state_vars, only : sitetype  & ! structure
                              , patchtype ! ! structure
      use ed_max_dims  , only : n_pft     ! ! intent(in)
      use grid_coms    , only : nzg       ! ! intent(in)

      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(sitetype), target      :: csite
      integer       , intent(in)  :: np
      integer       , intent(in)  :: cp
      real          , intent(in)  :: area_fac
      !----- Local variables. -------------------------------------------------------------!
      integer                     :: k
      !------------------------------------------------------------------------------------!

      csite%fast_soil_C                (np) = csite%fast_soil_C                (np)        &
                                            + csite%fast_soil_C                (cp)        &
                                            * area_fac
      csite%slow_soil_C                (np) = csite%slow_soil_C                (np)        &
                                            + csite%slow_soil_C                (cp)        &
                                            * area_fac
      csite%structural_soil_C          (np) = csite%structural_soil_C          (np)        &
                                            + csite%structural_soil_C          (cp)        &
                                            * area_fac
      csite%structural_soil_L          (np) = csite%structural_soil_L          (np)        &
                                            + csite%structural_soil_L          (cp)        &
                                            * area_fac
      csite%mineralized_soil_N         (np) = csite%mineralized_soil_N         (np)        &
                                            + csite%mineralized_soil_N         (cp)        &
                                            * area_fac
      csite%fast_soil_N                (np) = csite%fast_soil_N                (np)        &
                                            + csite%fast_soil_N                (cp)        &
                                            * area_fac
      csite%slow_soil_N                (np) = csite%slow_soil_N                (np)        &!ATT
                                            + csite%slow_soil_N                (cp)        & 
                                            * area_fac

      csite%sum_dgd                    (np) = csite%sum_dgd                    (np)        &
                                            + csite%sum_dgd                    (cp)        &
                                            * area_fac
      csite%sum_chd                    (np) = csite%sum_chd                    (np)        &
                                            + csite%sum_chd                    (cp)        &
                                            * area_fac
      csite%can_theta                  (np) = csite%can_theta                  (np)        &
                                            + csite%can_theta                  (cp)        &
                                            * area_fac
      csite%can_theiv                  (np) = csite%can_theiv                  (np)        &
                                            + csite%can_theiv                  (cp)        &
                                            * area_fac
      csite%can_prss                   (np) = csite%can_prss                   (np)        &
                                            + csite%can_prss                   (cp)        &
                                            * area_fac
      csite%can_shv                    (np) = csite%can_shv                    (np)        &
                                            + csite%can_shv                    (cp)        &
                                            * area_fac
      csite%can_co2                    (np) = csite%can_co2                    (np)        &
                                            + csite%can_co2                    (cp)        &
                                            * area_fac
      csite%can_depth                  (np) = csite%can_depth                  (np)        &
                                            + csite%can_depth                  (cp)        &
                                            * area_fac
      csite%ggbare                     (np) = csite%ggbare                     (np)        &
                                            + csite%ggbare                     (cp)        &
                                            * area_fac
      csite%ggveg                      (np) = csite%ggveg                      (np)        &
                                            + csite%ggveg                      (cp)        &
                                            * area_fac
      csite%rough                      (np) = csite%rough                      (np)        &
                                            + csite%rough                      (cp)        &
                                            * area_fac
      csite%mean_rh                    (np) = csite%mean_rh                    (np)        &
                                            + csite%mean_rh                    (cp)        &
                                            * area_fac
      csite%today_A_decomp             (np) = csite%today_A_decomp             (np)        &
                                            + csite%today_A_decomp             (cp)        &
                                            * area_fac
      csite%today_Af_decomp            (np) = csite%today_Af_decomp            (np)        &
                                            + csite%today_Af_decomp            (cp)        &
                                            * area_fac
      csite%fsc_in                     (np) = csite%fsc_in                     (np)        &
                                            + csite%fsc_in                     (cp)        &
                                            * area_fac
      csite%slsc_in                    (np) = csite%slsc_in                    (np)        &
                                            + csite%slsc_in                    (cp)        &
                                            * area_fac
      csite%slsn_in                    (np) = csite%slsn_in                    (np)        &
                                            + csite%slsn_in                    (cp)        &
                                            * area_fac
      csite%ssc_in                     (np) = csite%ssc_in                     (np)        &
                                            + csite%ssc_in                     (cp)        &
                                            * area_fac
      csite%ssl_in                     (np) = csite%ssl_in                     (np)        &
                                            + csite%ssl_in                     (cp)        &
                                            * area_fac
      csite%fsn_in                     (np) = csite%fsn_in                     (np)        &
                                            + csite%fsn_in                     (cp)        &
                                            * area_fac
      csite%total_plant_nitrogen_uptake(np) = csite%total_plant_nitrogen_uptake(np)        &
                                            + csite%total_plant_nitrogen_uptake(cp)        &
                                            * area_fac

      !----- Do the same thing for the multiple-level variables. --------------------------!
      do k=1,n_pft
         csite%repro                 (k,np) = csite%repro                    (k,np)        &
                                            + csite%repro                    (k,cp)        &
                                            * area_fac
      end do
      do k = 1, csite%nlev_sfcwater(cp)
         csite%sfcwater_mass         (k,np) = csite%sfcwater_mass            (k,np)        &
                                            + csite%sfcwater_mass            (k,cp)        &
                                            * area_fac
         csite%sfcwater_energy       (k,np) = csite%sfcwater_energy          (k,np)        &
                                            + csite%sfcwater_energy          (k,cp)        &
                                            * csite%sfcwater_mass            (k,cp)        &
                                            * area_fac
         csite%sfcwater_depth        (k,np) = csite%sfcwater_depth           (k,np)        &
                                            + csite%sfcwater_depth           (k,cp)        &
                                            * area_fac
      end do
      do k = 1, nzg
         csite%soil_energy           (k,np) = csite%soil_energy              (k,np)        &
                                            + csite%soil_energy              (k,cp)        &
                                            * area_fac
         csite%soil_water(k,np)             = csite%soil_water               (k,np)        &
                                            + csite%soil_water               (k,cp)        &
                                            * area_fac
         csite%ntext_soil(k,np) = csite%ntext_soil(k,cp)
      end do
!      print*,'np',np,'csite%ntext_soil',csite%ntext_soil(:,:)
      return
   end subroutine increment_patch_vars
   !=======================================================================================!
   !=======================================================================================!






   !=======================================================================================!
   !=======================================================================================!
   !     This subroutine will populate the disturbed patch with the cohorts that were      !
   ! disturbed but did not go extinct.                                                     !
   !---------------------------------------------------------------------------------------!
   subroutine insert_survivors(csite, np, cp, q, area_fac,poly_dest_type,mindbh_harvest,loss_fraction,lsl,nplant_resprout)

      use ed_state_vars, only : sitetype   & ! structure
                              , patchtype  & ! structure
                              , polygontype  ! ! structure
      use ed_misc_coms , only : idoutput   & ! intent(in)
                              , iqoutput   & ! intent(in)
                              , imoutput   ! ! intent(in)
      use ed_max_dims  , only : n_pft,n_dbh ! ! intent(in)
      use grid_coms    , only : nzg                     & ! intent(in)
                              , nzs
      use pft_coms     , only : resprout,bark_thickness_beta   &
                              , resprout_fraction &
                              , c2n_storage  & ! intent(in)
                              , c2n_leaf     & ! intent(in)
                              , c2n_recruit  & ! intent(in)
                              , c2n_stem     & ! intent(in)
                              , l2n_stem,hgt_min   
      use decomp_coms  , only : f_labile     &! ! intent(in)
                              , f_fast
      use pft_coms     , only : c2n_storage  & ! intent(in)
                              , c2n_leaf     & ! intent(in)
                              , c2n_recruit  & ! intent(in)
                              , c2n_stem     & ! intent(in)
                              , l2n_stem     ! ! intent(in)
      use grid_coms    , only : nzg          ! ! intent(in)
      use disturb_coms , only : fire_year
      use phenology_startup, only: initialize_paw
      use allometry    , only : h2dbh                    & ! function
                              , dbh2bd                   & ! function
                              , area_indices             & ! function
                              , ed_biomass
      use consts_coms  , only : t3ple                    & ! intent(in)
                              , pio4
      use fuse_fiss_utils, only : sort_cohorts             ! ! sub-routine
      use ed_therm_lib   , only : calc_veg_hcap            ! ! function

      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(sitetype)                  , target      :: csite
      integer                         , intent(in)  :: q
      integer                         , intent(in)  :: poly_dest_type
      integer                         , intent(in)  :: np
      integer                         , intent(in)  :: cp
      integer                         , intent(in)  :: lsl
      real          , dimension(n_pft), intent(in)  :: mindbh_harvest
      real                            , intent(in)  :: area_fac
      real                            , intent(in) :: loss_fraction
      real          , dimension(n_pft), intent(out):: nplant_resprout

      !----- Local variables. -------------------------------------------------------------!
      type(patchtype)                 , pointer     :: cpatch
      type(patchtype)                 , pointer     :: npatch
      type(patchtype)                 , pointer     :: tpatch
      logical        , dimension(:)   , allocatable :: mask
      integer                                       :: ico
      integer                                       :: ipft
      integer                                       :: nco
      integer                                       :: nc
      integer                                       :: oc
      real                                          :: agb_grass
      real                                          :: n_survivors
      real                                          :: survival_fac
      real                                          :: fast_litter
      real                                          :: slow_litter!ATT
      real                                          :: slow_litter_n!ATT
      real                                          :: struct_litter
      real                                          :: struct_lignin
      real                                          :: fast_litter_n
      real                                          :: struct_cohort
      real          , dimension(:)   , allocatable  :: survivors
      !------------------------------------------------------------------------------------!

      !------------------------------------------------------------------------------------!
      ! cpatch => Old patches we are looping                                               !
      ! npatch => New transition patch                                                     !
      ! tpatch => temporary patch                                                          !
      !------------------------------------------------------------------------------------!
      cpatch => csite%patch(cp)
      npatch => csite%patch(np)
      nullify(tpatch)
      allocate(tpatch)

      !----- Mask: flag to decide whether the cohort survived or not. ---------------------!
      allocate(mask(cpatch%ncohorts))
      mask(:) = .false.
      allocate(survivors(cpatch%ncohorts))

      !compute grass agb for patch to determine fire intensity
      agb_grass = 0.
      do ico = 1,cpatch%ncohorts
         if(cpatch%pft(ico) == 1) then
            agb_grass = agb_grass + cpatch%agb(ico) * cpatch%nplant(ico)
         endif
      end do
      print*,'agb_grass',agb_grass
      nplant_resprout(:) = 0.
      survivalloop: do ico = 1,cpatch%ncohorts
         survivors(ico) = survivorship(q,poly_dest_type, mindbh_harvest, csite, cp, ico,agb_grass)
         survival_fac = survivors(ico) * area_fac
         n_survivors     = cpatch%nplant(ico) * survival_fac
         nplant_resprout(cpatch%pft(ico)) = nplant_resprout(cpatch%pft(ico)) +                     &
              cpatch%nplant(ico) * (1 - survival_fac) * resprout_fraction(cpatch%pft(ico))

         !----- If something survived, make a new cohort. ---------------------------------!
         mask(ico) = n_survivors > 0.0
      end do survivalloop
      print*,'survivors',survivors

      !------------------------------------------------------------------------------------!
      !     If the new patch has received survivors from a donor already, then it should   !
      ! have cohorts.  So the temporary patch vector will be the sum of the new cohorts    !
      ! found here, plus those already applied previously in the loop calling this sub-    !
      ! routine.                                                                           !
      !------------------------------------------------------------------------------------!
      if (npatch%ncohorts > 0) then
         nco = npatch%ncohorts
         call allocate_patchtype(tpatch,count(mask) + npatch%ncohorts)
         call copy_patchtype(npatch,tpatch,1,npatch%ncohorts,1,npatch%ncohorts)
         call deallocate_patchtype(npatch)
      else
         nco = 0
         call allocate_patchtype(tpatch,count(mask))
      end if

      cohortloop: do ico = 1,cpatch%ncohorts

         if(ico <= cpatch%ncohorts)then
            survival_fac = survivors(ico) * area_fac
         else
            survival_fac = 1.
         endif
                     
         n_survivors  = cpatch%nplant(ico) * survival_fac

         !----- If mask is true, at least some of this cohort survived. -------------------!
         if (mask(ico)) then
            nco = nco + 1
            call copy_patchtype(cpatch,tpatch,ico,ico,nco,nco)

            !------------------------------------------------------------------------------!
            !    Scale the total area based on the new population density and new area.    !
            ! We must also rescale all extensive properties from cohorts, since they are   !
            ! per unit area and we are effectively changing the area.                      !
            ! IMPORTANT: Only cohort-level variables that have units per area (m2) should  !
            !            be rescaled.  Variables whose units are per plant should _NOT_ be !
            !            included here.                                                    !
            !------------------------------------------------------------------------------!
            tpatch%lai                (nco) = tpatch%lai              (nco) * survival_fac
            tpatch%wai                (nco) = tpatch%wai              (nco) * survival_fac
            tpatch%wpa                (nco) = tpatch%wpa              (nco) * survival_fac
            tpatch%nplant             (nco) = tpatch%nplant           (nco) * survival_fac
            tpatch%mean_gpp           (nco) = tpatch%mean_gpp         (nco) * survival_fac
            tpatch%mean_leaf_resp     (nco) = tpatch%mean_leaf_resp   (nco) * survival_fac
            tpatch%mean_root_resp     (nco) = tpatch%mean_root_resp   (nco) * survival_fac
            tpatch%mean_growth_resp   (nco) = tpatch%mean_growth_resp (nco) * survival_fac
            tpatch%mean_storage_resp  (nco) = tpatch%mean_storage_resp(nco) * survival_fac
            tpatch%mean_vleaf_resp    (nco) = tpatch%mean_vleaf_resp  (nco) * survival_fac
            tpatch%today_gpp          (nco) = tpatch%today_gpp        (nco) * survival_fac
            tpatch%today_nppleaf      (nco) = tpatch%today_nppleaf    (nco) * survival_fac
            tpatch%today_nppfroot     (nco) = tpatch%today_nppfroot   (nco) * survival_fac
            tpatch%today_nppsapwood   (nco) = tpatch%today_nppsapwood (nco) * survival_fac
            tpatch%today_nppcroot     (nco) = tpatch%today_nppcroot   (nco) * survival_fac
            tpatch%today_nppseeds     (nco) = tpatch%today_nppseeds   (nco) * survival_fac
            tpatch%today_nppwood      (nco) = tpatch%today_nppwood    (nco) * survival_fac
            tpatch%today_nppdaily     (nco) = tpatch%today_nppdaily   (nco) * survival_fac
            tpatch%today_gpp_pot      (nco) = tpatch%today_gpp_pot    (nco) * survival_fac
            tpatch%today_gpp_max      (nco) = tpatch%today_gpp_max    (nco) * survival_fac
            tpatch%today_leaf_resp    (nco) = tpatch%today_leaf_resp  (nco) * survival_fac
            tpatch%today_root_resp    (nco) = tpatch%today_root_resp  (nco) * survival_fac
            tpatch%Psi_open           (nco) = tpatch%Psi_open         (nco) * survival_fac
            tpatch%gpp                (nco) = tpatch%gpp              (nco) * survival_fac
            tpatch%leaf_respiration   (nco) = tpatch%leaf_respiration (nco) * survival_fac
            tpatch%root_respiration   (nco) = tpatch%root_respiration (nco) * survival_fac
            tpatch%monthly_dndt       (nco) = tpatch%monthly_dndt     (nco) * survival_fac
            tpatch%leaf_water         (nco) = tpatch%leaf_water       (nco) * survival_fac
            tpatch%leaf_hcap          (nco) = tpatch%leaf_hcap        (nco) * survival_fac
            tpatch%leaf_energy        (nco) = tpatch%leaf_energy      (nco) * survival_fac
            tpatch%wood_water         (nco) = tpatch%wood_water       (nco) * survival_fac
            tpatch%wood_hcap          (nco) = tpatch%wood_hcap        (nco) * survival_fac
            tpatch%wood_energy        (nco) = tpatch%wood_energy      (nco) * survival_fac
            !----- Crown area shall not exceed 1. -----------------------------------------!
            tpatch%crown_area         (nco) = min(1.,tpatch%crown_area(nco) * survival_fac)
            !----- Carbon flux monthly means are extensive, we must convert them. ---------!
            if (idoutput > 0 .or. imoutput > 0 .or. iqoutput > 0) then
               tpatch%dmean_par_l     (nco) = tpatch%dmean_par_l      (nco) * survival_fac
               tpatch%dmean_par_l_beam(nco) = tpatch%dmean_par_l_beam (nco) * survival_fac
               tpatch%dmean_par_l_diff(nco) = tpatch%dmean_par_l_diff (nco) * survival_fac
            end if
            if (imoutput > 0 .or. iqoutput > 0) then
               tpatch%mmean_par_l     (nco) = tpatch%mmean_par_l      (nco) * survival_fac
               tpatch%mmean_par_l_beam(nco) = tpatch%mmean_par_l_beam (nco) * survival_fac
               tpatch%mmean_par_l_diff(nco) = tpatch%mmean_par_l_diff (nco) * survival_fac
            end if
            if (iqoutput > 0) then
               tpatch%qmean_par_l     (:,nco) = tpatch%qmean_par_l      (:,nco)            &
                                              * survival_fac
               tpatch%qmean_par_l_beam(:,nco) = tpatch%qmean_par_l_beam (:,nco)            &
                                              * survival_fac
               tpatch%qmean_par_l_diff(:,nco) = tpatch%qmean_par_l_diff (:,nco)            &
                                              * survival_fac
            end if
         end if

      end do cohortloop


      !----- Copy the temporary patch into the newpatch. ----------------------------------!
      call allocate_patchtype(npatch,tpatch%ncohorts)
      call copy_patchtype(tpatch,npatch,1,tpatch%ncohorts,1,tpatch%ncohorts)
      call deallocate_patchtype(tpatch)

      deallocate(tpatch)
      deallocate(mask)

      !ATT accumulate litter subroutine added
      !---- Initialise the non-scaled litter pools. ---------------------------------------!
      fast_litter   = 0.0
      slow_litter   = 0.0
      struct_litter = 0.0
      struct_lignin = 0.0
      fast_litter_n = 0.0
      slow_litter_n = 0.0

      !------------------------------------------------------------------------------------!
      ! cpatch => contributing patch                                                       !
      ! npatch => new patch.                                                               !
      !------------------------------------------------------------------------------------!
      cpatch => csite%patch(cp)
      npatch => csite%patch(np)

      do ico = 1,cpatch%ncohorts
         ipft = cpatch%pft(ico)
         slow_litter   = slow_litter                                                       &
              ! + (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico)) &
              + (1. - survivors(ico)) &
              * (f_labile(ipft) * cpatch%balive(ico) + cpatch%bstorage(ico))      &
              * cpatch%nplant(ico) * (1 - f_fast(ipft))
         slow_litter_n = slow_litter_n                                                     &
              !+ (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico)) &
              + (1. - survivors(ico)) &
              * (f_labile(ipft) * cpatch%balive(ico) / c2n_leaf(ipft)             &
              + cpatch%nstorage(ico)) * cpatch%nplant(ico) * (1 - f_fast(ipft))
 
         fast_litter   = fast_litter                                                       &
              !+ (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico)) &
              + (1. - survivors(ico)) &
              * ( f_labile(ipft) * cpatch%balive(ico) + cpatch%bstorage(ico))     &
              * cpatch%nplant(ico) * f_fast(ipft)
         fast_litter_n = fast_litter_n                                                     &
              + (1. - survivors(ico)) &
              !+ (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico)) &
              * ( f_labile(ipft) * cpatch%balive(ico) / c2n_leaf(ipft)            &
              + cpatch%nstorage(ico)) * cpatch%nplant(ico) * f_fast(ipft)   
         
         struct_cohort = cpatch%nplant(ico)                                                &
              ! * (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico)) &
              * (1. - survivors(ico)) &
              * ( (1. - loss_fraction ) * cpatch%bdead(ico)                       &
              + (1. - f_labile(ipft)) * cpatch%balive(ico) )

         struct_litter = struct_litter + struct_cohort
         struct_lignin = struct_lignin + struct_cohort * l2n_stem / c2n_stem(ipft)
         
      end do

      !----- Load disturbance litter directly into carbon and N pools. --------------------!
      csite%fast_soil_C(np)       = csite%fast_soil_C(np)       + fast_litter   * area_fac
      csite%slow_soil_C(np)       = csite%slow_soil_C(np)       + slow_litter   * area_fac
      csite%structural_soil_C(np) = csite%structural_soil_C(np) + struct_litter * area_fac
      csite%structural_soil_L(np) = csite%structural_soil_L(np) + struct_lignin * area_fac
      csite%fast_soil_N(np)       = csite%fast_soil_N(np)       + fast_litter_n * area_fac
      csite%slow_soil_N(np)       = csite%slow_soil_N(np)       + slow_litter_n * area_fac!ATT
      
      deallocate(survivors)

      return
   end subroutine insert_survivors
   !=======================================================================================!
   !=======================================================================================!






   !=======================================================================================!
   !=======================================================================================!
   !     This subroutine updates the litter pools after a disturbance takes place.         !
   !---------------------------------------------------------------------------------------!
   subroutine accum_dist_litt(csite,np,cp,q,area_fac,loss_fraction,poly_dest_type          &
                             ,mindbh_harvest)
      use ed_state_vars, only : sitetype     & ! structure
                              , patchtype    & ! structure
                              , polygontype  ! ! structure
      use decomp_coms  , only : f_labile     &! ! intent(in)
                              , f_fast
      use ed_max_dims  , only : n_pft        ! ! intent(in)
      use pft_coms     , only : c2n_storage  & ! intent(in)
                              , c2n_leaf     & ! intent(in)
                              , c2n_recruit  & ! intent(in)
                              , c2n_stem     & ! intent(in)
                              , l2n_stem     ! ! intent(in)
      use grid_coms    , only : nzg          ! ! intent(in)

      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(sitetype)                   , target     :: csite
      integer                          , intent(in) :: np
      integer                          , intent(in) :: cp
      real           , dimension(n_pft), intent(in) :: mindbh_harvest
      real                             , intent(in) :: loss_fraction
      integer                          , intent(in) :: q
      real                             , intent(in) :: area_fac
      integer                          , intent(in) :: poly_dest_type

      !----- Local variables. -------------------------------------------------------------!
      type(patchtype)                  , pointer    :: cpatch
      type(patchtype)                  , pointer    :: npatch
      integer                                       :: ico
      integer                                       :: ipft
      real                                          :: fast_litter
      real                                          :: slow_litter!ATT
      real                                          :: slow_litter_n!ATT
      real                                          :: struct_litter
      real                                          :: struct_lignin
      real                                          :: fast_litter_n
      real                                          :: struct_cohort
      !------------------------------------------------------------------------------------!

      !---- Initialise the non-scaled litter pools. ---------------------------------------!
      fast_litter   = 0.0
      slow_litter   = 0.0
      struct_litter = 0.0
      struct_lignin = 0.0
      fast_litter_n = 0.0
      slow_litter_n = 0.0

      !------------------------------------------------------------------------------------!
      ! cpatch => contributing patch                                                       !
      ! npatch => new patch.                                                               !
      !------------------------------------------------------------------------------------!
      cpatch => csite%patch(cp)
      npatch => csite%patch(np)

      do ico = 1,cpatch%ncohorts
         ipft = cpatch%pft(ico)
         slow_litter   = slow_litter                                                       &
              + (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico,0.)) &
              * (f_labile(ipft) * cpatch%balive(ico) + cpatch%bstorage(ico))      &
              * cpatch%nplant(ico) * (1 - f_fast(ipft))
         slow_litter_n = slow_litter_n                                                     &
              + (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico,0.)) &
              * (f_labile(ipft) * cpatch%balive(ico) / c2n_leaf(ipft)             &
              + cpatch%nstorage(ico)) * cpatch%nplant(ico) * (1 - f_fast(ipft))
 
         fast_litter   = fast_litter                                                       &
              + (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico,0.)) &
              * ( f_labile(ipft) * cpatch%balive(ico) + cpatch%bstorage(ico))     &
              * cpatch%nplant(ico) * f_fast(ipft)
         fast_litter_n = fast_litter_n                                                     &
              + (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico,0.)) &
              * ( f_labile(ipft) * cpatch%balive(ico) / c2n_leaf(ipft)            &
              + cpatch%nstorage(ico)) * cpatch%nplant(ico) * f_fast(ipft)   
         
         struct_cohort = cpatch%nplant(ico)                                                &
              * (1. - survivorship(q,poly_dest_type,mindbh_harvest,csite,cp,ico,0.)) &
              * ( (1. - loss_fraction ) * cpatch%bdead(ico)                       &
              + (1. - f_labile(ipft)) * cpatch%balive(ico) )

         struct_litter = struct_litter + struct_cohort
         struct_lignin = struct_lignin + struct_cohort * l2n_stem / c2n_stem(ipft)
      end do

      !----- Load disturbance litter directly into carbon and N pools. --------------------!
      csite%fast_soil_C(np)       = csite%fast_soil_C(np)       + fast_litter   * area_fac
      csite%slow_soil_C(np)       = csite%slow_soil_C(np)       + slow_litter   * area_fac
      csite%structural_soil_C(np) = csite%structural_soil_C(np) + struct_litter * area_fac
      csite%structural_soil_L(np) = csite%structural_soil_L(np) + struct_lignin * area_fac
      csite%fast_soil_N(np)       = csite%fast_soil_N(np)       + fast_litter_n * area_fac
      csite%slow_soil_N(np)       = csite%slow_soil_N(np)       + slow_litter_n * area_fac!ATT
      return
   end subroutine accum_dist_litt
   !=======================================================================================!
   !=======================================================================================!






   !=======================================================================================!
   !=======================================================================================!
   !     This function computes the survivorship rate after some disturbance happens.      !
   !---------------------------------------------------------------------------------------!
   real function survivorship(dest_type,poly_dest_type,mindbh_harvest,csite,ipa,ico,grass_agb)
      use ed_state_vars, only : patchtype                & ! structure
                              , sitetype                 ! ! structure
      use disturb_coms , only : treefall_hite_threshold  & ! intent(in)
                              , fire_residence_time      &
                              , grass_biomass_threshold  &
                              , fire_year, fire_area
      use pft_coms     , only : treefall_s_ltht          & ! intent(in)
                              , treefall_s_gtht          & ! intent(in)
                              , bark_thickness_beta      &
                              , is_grass
      use ed_max_dims  , only : n_pft                    ! ! intent(in)
      
      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(sitetype)                  , target     :: csite
      real          , dimension(n_pft), intent(in) :: mindbh_harvest
      real                            , intent(in) :: grass_agb
      integer                         , intent(in) :: ico
      integer                         , intent(in) :: ipa
      integer                         , intent(in) :: dest_type
      integer                         , intent(in) :: poly_dest_type
      !----- Local variables. -------------------------------------------------------------!
      type(patchtype)                 , pointer    :: cpatch
      integer                                      :: ipft
      real :: bark_thickness, cambial_damage_tao, P_survive
      !------------------------------------------------------------------------------------!

     ! print*,'fire_residence_time',fire_residence_time,'dest_type',dest_type
      cpatch => csite%patch(ipa)
      ipft = cpatch%pft(ico)

      !----- Base the survivorship rates on the destination type. -------------------------!
      select case(dest_type)
      case (1) !----- Agriculture/cropland. -----------------------------------------------!
         survivorship = 0.0

      case (2) !----- Secondary land or forest plantation. --------------------------------!

         !----- Decide the fate based on the type of secondary disturbance. ---------------!
         select case (poly_dest_type)
         case (0) !----- Land abandonment, assume this is the last harvest. ---------------!
            survivorship = 0.0
         case (1) !----- Biomass logging, assume that nothing stays. ----------------------!
            survivorship = 0.0
         case (2) !----- Selective logging. -----------------------------------------------!
            !------------------------------------------------------------------------------!
            !     If the PFT DBH exceeds the minimum PFT for harvesting, the survivorship  !
            ! should be zero, otherwise, we assume survivorship similar to the treefall    !
            ! disturbance rate for short trees.                                            ! 
            !------------------------------------------------------------------------------!
            if (cpatch%dbh(ico) >= mindbh_harvest(ipft)) then
               survivorship = 0.0
            else
               survivorship = treefall_s_ltht(ipft)
            end if
         end select

      case (3) !----- Primary land. -------------------------------------------------------!

         !----- Decide the fate based on the type of natural disturbance. -----------------!
         select case (poly_dest_type)


         !Hyjack Treefall disturbance to use for statistical fire mortality
         case (0) !----- Treefall, we must check the cohort height. -----------------------!
            !fire survivorship based on dbh and bark thickness 
            !equations taken from Thonick et al. 2010 Biogeosciences
            bark_thickness = bark_thickness_beta(cpatch%pft(ico)) * cpatch%dbh(ico) !in cm
            cambial_damage_tao = 2.9 * bark_thickness ** 2 !in minutes
            print*,' bark_thickness', bark_thickness,'dbh',cpatch%dbh(ico)!,'bark_thickness_beta',bark_thickness_beta(cpatch%pft(ico))

            !-----------------SPITFIRE MODEL-------------------------------------------------------------!
            !print*,'fire_residence_time/cambial_damage_tao',fire_residence_time/cambial_damage_tao
            !print*,'cambial_damage_tao',cambial_damage_tao
            !if (fire_residence_time/cambial_damage_tao < 0.22) then
            !   survivorship = 1.
            !elseif ((fire_residence_time/cambial_damage_tao > 0.22) .and.                    &
            !     (fire_residence_time/cambial_damage_tao < 2.)) then
            !   survivorship = 1. - 0.563 * (fire_residence_time/cambial_damage_tao)
            !else
            !   survivorship = 0.
            !endif
            !--------------------------------------------------------------------------------------------!

            !------------------Adam's Neotropical Savanna Survivorship-----------------------------------!
            !binomial survivorship scheme, also include random seed in ed_driver.f90 if you want to use
            ! P_survive = 1. / (1. + exp(4.2026 - 0.4986 * bark_thickness * 10.))
             !survivorship = real(random_binomial2(1, P_survive, .true.))!binomial survivorship scheme
             
           ! survivorship = 1. / (1. + exp(4.2026 - 0.4986 * bark_thickness * 10.))
           ! print*,'P_survive',P_survive,'survivorship',survivorship
           ! if (is_grass(cpatch%pft(ico))) then
           !    P_survive = 0.
           !    survivorship = 0.
           !    print*,'all grass burns'
           ! endif
 
            
            !----Adam's Neotropical Savanna Survivorship With Fire Intensity Dependent on Grass Biomass----!
            !binomial survivorship scheme, also include random seed in ed_driver.f90 if you want to use
            if(fire_year)then!if there is fire, no treefall
               if (grass_agb > grass_biomass_threshold .and. (.not. is_grass(cpatch%pft(ico)))) then
                  !high intensity fire
                  P_survive = min(1. - (-0.0618 * bark_thickness * 10. + 0.9617),1.)
                  survivorship = min(1. - (-0.0618 * bark_thickness * 10. + 0.9617),1.)
                  print*,'Grass biomass: ',grass_agb,', High Intensity Fire ','grass_biomass_threshold',grass_biomass_threshold
               elseif  (grass_agb <= grass_biomass_threshold .and. (.not. is_grass(cpatch%pft(ico)))) then
                  !low intensity fire
                  P_survive = min(1. - (-0.0602 * bark_thickness * 10. + 0.8516),1.)
                  survivorship = min(1. - (-0.0602 * bark_thickness * 10. + 0.8516),1.)
                  print*,'Grass biomass: ',grass_agb,', Low Intensity Fire ','grass_biomass_threshold',grass_biomass_threshold
               elseif (is_grass(cpatch%pft(ico))) then
                  P_survive = 0.
                  survivorship = 0.
                  print*,'all grass burns'
               endif
               !treefall mortality
               if (cpatch%hite(ico) < treefall_hite_threshold) then
                  survivorship =  survivorship * (0.01 / fire_area * treefall_s_ltht(ipft)    &
                       + (fire_area - 0.01) / fire_area) 
               else
                  survivorship = survivorship * (0.01 / fire_area * treefall_s_gtht(ipft)    &
                       + (fire_area - 0.01) / fire_area)                                       
               end if
               print*,'P_survive',P_survive,'survivorship',survivorship
            else! if there is no fire, treefall disturbance occurs
               if (cpatch%hite(ico) < treefall_hite_threshold) then
                  survivorship =  treefall_s_ltht(ipft)
                  print*, 'not fire year, survivorship: ',survivorship,'dbh',cpatch%dbh(ico)
               else
                  survivorship = treefall_s_gtht(ipft)
                  print*, 'not fire year, survivorship: ',survivorship,'dbh',cpatch%dbh(ico)
               end if
            endif

         case (1) !----- Fire, no survival. -----------------------------------------------!
            survivorship = 0.0
         end select
      end select

      return
   end function survivorship
   !=======================================================================================!
   !=======================================================================================!






   !=======================================================================================!
   !=======================================================================================!
   !    Add a cohort of the appropriate PFT type to populate a plantation/cropland/pasture !
   ! patch.                                                                                !
   !---------------------------------------------------------------------------------------!
   subroutine plant_patch(csite,np,cp,mzg,pft,density,ntext_soil,green_leaf_factor            &
                         ,height_factor,lsl)
      use ed_state_vars , only  : sitetype                 & ! structure
                                , patchtype                ! ! structure
      use pft_coms       , only : q                        & ! intent(in)
                                , qsw                      & ! intent(in)
                                , sla                      & ! intent(in)
                                , hgt_min                  & ! intent(in)
                                , max_dbh                  & ! intent(in)
                                , is_grass,bark_thickness_beta ! ! intent(in)
      use ed_misc_coms   , only : dtlsm                    ! ! intent(in)
      use fuse_fiss_utils, only : sort_cohorts             ! ! sub-routine
      use ed_therm_lib   , only : calc_veg_hcap            ! ! function
      use consts_coms    , only : t3ple                    & ! intent(in)
                                , pio4                     ! ! intent(in)
      use allometry      , only : h2dbh                    & ! function
                                , dbh2bd                   & ! function
                                , area_indices             & ! function
                                , ed_biomass               ! ! function
      use ed_max_dims    , only : n_pft                    ! ! intent(in)
      use phenology_coms , only : retained_carbon_fraction ! ! intent(in)

      implicit none
      !----- Arguments. -------------------------------------------------------------------!
      type(sitetype)                  , target     :: csite
      integer                         , intent(in) :: mzg
      integer                         , intent(in) :: np
      integer                         , intent(in) :: cp
      integer                         , intent(in) :: pft
      integer                         , intent(in) :: lsl
      integer       , dimension(mzg)  , intent(in) :: ntext_soil
      real          , dimension(n_pft), intent(in) :: green_leaf_factor
      real                            , intent(in) :: density
      real                            , intent(in) :: height_factor
      !----- Local variables. -------------------------------------------------------------!
      type(patchtype)                 , pointer    :: cpatch
      type(patchtype)                 , pointer    :: tpatch
      integer                                      :: nc
      real                                         :: salloc
      real                                         :: salloci
      real                                         :: bleaf_max
      real                                         :: balive_max
      !------------------------------------------------------------------------------------!



      cpatch => csite%patch(np)


      !------------------------------------------------------------------------------------!
      !      Reallocate the current cohort with an extra space for the planted plantation  !
      ! cohort.  If the patch was previously empty, we simply create the first cohort.     !
      !------------------------------------------------------------------------------------!
      nc = cpatch%ncohorts + 1
      if (cpatch%ncohorts > 0) then
         nullify(tpatch)
         allocate(tpatch)
         call allocate_patchtype(tpatch,cpatch%ncohorts)
         call copy_patchtype(cpatch,tpatch,1,cpatch%ncohorts,1,cpatch%ncohorts)
         call deallocate_patchtype(cpatch)
         call allocate_patchtype(cpatch,tpatch%ncohorts + 1)
         call copy_patchtype(tpatch,cpatch,1,tpatch%ncohorts,1,tpatch%ncohorts)
         call deallocate_patchtype(tpatch)
         deallocate(tpatch)
      else
         call allocate_patchtype(cpatch,1)
      end if
         
      cpatch%ncohorts = nc
      csite%paco_n(np)= nc

      !------------------------------------------------------------------------------------!
      !      We can add all the biomass needed in a single cohort, even if it exceeds the  !
      ! safe range (i.e., with maximum LAI exceeding the unity).  In case it is a large    !
      ! cohort, it will soon be split at the splitting call of apply_disturbances().  This !
      ! new cohort will always be the last here, they will be sorted afterwards too.       !
      !------------------------------------------------------------------------------------!
      cpatch%pft(nc)    = pft
      cpatch%nplant(nc) = density
      cpatch%hite(nc)   = hgt_min(cpatch%pft(nc)) * min(1.0,height_factor)
     ! print*,'RESPROUT','pft',pft,'density',density
      !------------------------------------------------------------------------------------!

      !----- Initialise other cohort-level variables. -------------------------------------!
      call init_ed_cohort_vars(cpatch, nc, lsl)

      !-------   by ATT   Initialize cohort_id --------------------------------------------!
      cpatch%cohort_id(nc) = nc

      !----- Find DBH and the maximum leaf biomass. ---------------------------------------!
      cpatch%dbh(nc)   = h2dbh(cpatch%hite(nc),cpatch%pft(nc)) !- 2. * bark_thickness_beta(cpatch%pft(nc))
      cpatch%dbh(nc)   = cpatch%dbh(nc) - 2. * bark_thickness_beta(cpatch%pft(nc)) * cpatch%dbh(nc)

      cpatch%bdead(nc) = dbh2bd(cpatch%dbh(nc),cpatch%pft(nc))
     ! print*,' cpatch%dbh(nc)', cpatch%dbh(nc)
      !------------------------------------------------------------------------------------!
      !      Initialise the active and storage biomass scaled by the leaf drought          !
      !      phenology (or start with 1.0 if the plant doesn't                             !
      !      shed their leaves due to water stress.                                        !
      !------------------------------------------------------------------------------------!
      call pheninit_balive_bstorage(mzg,csite,np,nc,ntext_soil,green_leaf_factor)
      !dont need to put in a patch loop because ntext soil is not actually used in this routine
      !------------------------------------------------------------------------------------!



      !----- Compute all area indices needed. ---------------------------------------------!
      call area_indices(cpatch%nplant(nc),cpatch%bleaf(nc),cpatch%bdead(nc)                &
                       ,cpatch%balive(nc),cpatch%dbh(nc),cpatch%hite(nc),cpatch%pft(nc)    &
                       ,cpatch%sla(nc),cpatch%lai(nc),cpatch%wpa(nc),cpatch%wai(nc)        &
                       ,cpatch%crown_area(nc),cpatch%bsapwood(nc))


      !----- Find the new basal area and above-ground biomass. ----------------------------!
      cpatch%basarea(nc) = pio4 * cpatch%dbh(nc) * cpatch%dbh(nc)
      cpatch%agb(nc)     = ed_biomass(cpatch%bdead(nc),cpatch%balive(nc),cpatch%bleaf(nc)  &
                                     ,cpatch%pft(nc),cpatch%hite(nc) ,cpatch%bstorage(nc)  &
                                     ,cpatch%bsapwood(nc))
      print*,'cpatch%can_temp(nc)',csite%can_temp(np)
      cpatch%leaf_temp(nc)  = csite%can_temp(np)
      print*,'cpatch%leaf_temp(nc)',cpatch%leaf_temp(nc)
      cpatch%leaf_water(nc) = 0.0
      cpatch%leaf_fliq(nc)  = 0.0
      cpatch%wood_temp(nc)  = csite%can_temp(np)
      cpatch%wood_water(nc) = 0.0
      cpatch%wood_fliq(nc)  = 0.0

      !----- Because we assigned no water, the internal energy is simply hcap*T. ----------!
      call calc_veg_hcap(cpatch%bleaf(nc),cpatch%bdead(nc),cpatch%bsapwood(nc)             &
                        ,cpatch%nplant(nc),cpatch%pft(nc)                                  &
                        ,cpatch%leaf_hcap(nc),cpatch%wood_hcap(nc))

      cpatch%leaf_energy(nc) = cpatch%leaf_hcap(nc) * cpatch%leaf_temp(nc)
      cpatch%wood_energy(nc) = cpatch%wood_hcap(nc) * cpatch%wood_temp(nc)

      call is_resolvable(csite,np,nc,green_leaf_factor)

      !----- Should plantations be considered recruits? -----------------------------------!
      cpatch%new_recruit_flag(nc) = 1

      !----- Sort the cohorts so that the new cohort is at the correct height bin. --------!
      call sort_cohorts(cpatch)

      return
   end subroutine plant_patch
   !=======================================================================================!
   !=======================================================================================!

   FUNCTION random_binomial2(n, pp, first) RESULT(ival)
     !**********************************************************************
     !     Translated to Fortran 90 by Alan Miller from:
     !                              RANLIB
     !
     !     Library of Fortran Routines for Random Number Generation
     !
     !                      Compiled and Written by:
     !
     !                           Barry W. Brown
     !                            James Lovato
     !
     !               Department of Biomathematics, Box 237
     !               The University of Texas, M.D. Anderson Cancer Center
     !               1515 Holcombe Boulevard
     !               Houston, TX      77030
     !
     ! This work was supported by grant CA-16672 from the National Cancer Institute.

     !                    GENerate BINomial random deviate

     !                              Function

     !     Generates a single random deviate from a binomial
     !     distribution whose number of trials is N and whose
     !     probability of an event in each trial is P.

     !                              Arguments

     !     N  --> The number of trials in the binomial distribution
     !            from which a random deviate is to be generated.
     !                              INTEGER N

     !     P  --> The probability of an event in each trial of the
     !            binomial distribution from which a random deviate
     !            is to be generated.
     !                              REAL P

     !     FIRST --> Set FIRST = .TRUE. for the first call to perform initialization
     !               the set FIRST = .FALSE. for further calls using the same pair
     !               of parameter values (N, P).
     !                              LOGICAL FIRST

     !     random_binomial2 <-- A random deviate yielding the number of events
     !                from N independent trials, each of which has
     !                a probability of event P.
     !                              INTEGER random_binomial

     !                              Method

     !     This is algorithm BTPE from:

     !         Kachitvichyanukul, V. and Schmeiser, B. W.
     !         Binomial Random Variate Generation.
     !         Communications of the ACM, 31, 2 (February, 1988) 216.

     !**********************************************************************

     !*****DETERMINE APPROPRIATE ALGORITHM AND WHETHER SETUP IS NECESSARY

     !     ..
     !     .. Scalar Arguments ..
     REAL, INTENT(IN)    :: pp
     INTEGER, INTENT(IN) :: n
     LOGICAL, INTENT(IN) :: first
     INTEGER             :: ival
     !     ..
     !     .. Local Scalars ..
     REAL            :: alv, amaxp, f, f1, f2, u, v, w, w2, x, x1, x2, ynorm, z, z2
     REAL, PARAMETER :: zero = 0.0, half = 0.5, one = 1.0
     INTEGER         :: i, ix, ix1, k, mp
     INTEGER, SAVE   :: m
     REAL, SAVE      :: p, q, xnp, ffm, fm, xnpq, p1, xm, xl, xr, c, al, xll,  &
          xlr, p2, p3, p4, qn, r, g

     !     ..
     !     .. Executable Statements ..

     !*****SETUP, PERFORM ONLY WHEN PARAMETERS CHANGE

     IF (first) THEN
        p = MIN(pp, one-pp)
        q = one - p
        xnp = n * p
     END IF

     IF (xnp > 30.) THEN
        IF (first) THEN
           ffm = xnp + p
           m = ffm
           fm = m
           xnpq = xnp * q
           p1 = INT(2.195*SQRT(xnpq) - 4.6*q) + half
           xm = fm + half
           xl = xm - p1
           xr = xm + p1
           c = 0.134 + 20.5 / (15.3 + fm)
           al = (ffm-xl) / (ffm - xl*p)
           xll = al * (one + half*al)
           al = (xr - ffm) / (xr*q)
           xlr = al * (one + half*al)
           p2 = p1 * (one + c + c)
           p3 = p2 + c / xll
           p4 = p3 + c / xlr
        END IF

        !*****GENERATE VARIATE, Binomial mean at least 30.
        
20      CALL RANDOM_NUMBER(u)
        u = u * p4
        
        CALL RANDOM_NUMBER(v)

        !     TRIANGULAR REGION

        IF (u <= p1) THEN
           ix = xm - p1 * v + u
           GO TO 110
        END IF

        !     PARALLELOGRAM REGION

        IF (u <= p2) THEN
           x = xl + (u-p1) / c
           v = v * c + one - ABS(xm-x) / p1
           IF (v > one .OR. v <= zero) GO TO 20
           ix = x
        ELSE

           !     LEFT TAIL

           IF (u <= p3) THEN
              ix = xl + LOG(v) / xll
              IF (ix < 0) GO TO 20
              v = v * (u-p2) * xll
           ELSE

              !     RIGHT TAIL

              ix = xr - LOG(v) / xlr
              IF (ix > n) GO TO 20
              v = v * (u-p3) * xlr
           END IF
        END IF

        !*****DETERMINE APPROPRIATE WAY TO PERFORM ACCEPT/REJECT TEST

        k = ABS(ix-m)
        IF (k <= 20 .OR. k >= xnpq/2-1) THEN

           !     EXPLICIT EVALUATION

           f = one
           r = p / q
           g = (n+1) * r
           IF (m < ix) THEN
              mp = m + 1
              DO i = mp, ix
                 f = f * (g/i-r)
              END DO

           ELSE IF (m > ix) THEN
              ix1 = ix + 1
              DO i = ix1, m
                 f = f / (g/i-r)
              END DO
           END IF

           IF (v > f) THEN
              GO TO 20
           ELSE
              GO TO 110
           END IF
        END IF

        !     SQUEEZING USING UPPER AND LOWER BOUNDS ON LOG(F(X))

        amaxp = (k/xnpq) * ((k*(k/3. + .625) + .1666666666666)/xnpq + half)
        ynorm = -k * k / (2.*xnpq)
        alv = LOG(v)
        IF (alv<ynorm - amaxp) GO TO 110
        IF (alv>ynorm + amaxp) GO TO 20

        !     STIRLING'S (actually de Moivre's) FORMULA TO MACHINE ACCURACY FOR
        !     THE FINAL ACCEPTANCE/REJECTION TEST

        x1 = ix + 1
        f1 = fm + one
        z = n + 1 - fm
        w = n - ix + one
        z2 = z * z
        x2 = x1 * x1
        f2 = f1 * f1
        w2 = w * w
        IF (alv - (xm*LOG(f1/x1) + (n-m+half)*LOG(z/w) + (ix-m)*LOG(w*p/(x1*q)) +    &
             (13860.-(462.-(132.-(99.-140./f2)/f2)/f2)/f2)/f1/166320. +               &
             (13860.-(462.-(132.-(99.-140./z2)/z2)/z2)/z2)/z/166320. +                &
             (13860.-(462.-(132.-(99.-140./x2)/x2)/x2)/x2)/x1/166320. +               &
             (13860.-(462.-(132.-(99.-140./w2)/w2)/w2)/w2)/w/166320.) > zero) THEN
           GO TO 20
        ELSE
           GO TO 110
        END IF

     ELSE
        !     INVERSE CDF LOGIC FOR MEAN LESS THAN 30
        IF (first) THEN
           qn = q ** n
           r = p / q
           g = r * (n+1)
        END IF

90      ix = 0
        f = qn
        CALL RANDOM_NUMBER(u)
100     IF (u >= f) THEN
           IF (ix > 110) GO TO 90
           u = u - f
           ix = ix + 1
           f = f * (g/ix - r)
           GO TO 100
        END IF
     END IF

110  IF (pp > half) ix = n - ix
     ival = ix
     RETURN

   END FUNCTION random_binomial2

   !=======================================================================================!
   !=======================================================================================!
   
end module disturbance_utils
!==========================================================================================!
!==========================================================================================!
