!==========================================================================================!
!==========================================================================================!
!     This is the main loader of ecosystem parameters.  Since some compilers do not under- !
! stand the assignment in the modules when the variable is not a constant (parameter),     !
! this is the safest way to guarantee it will read something (not to mention that makes    !
! compilation much faster when you want to test the sensitivity of one number).            !
!------------------------------------------------------------------------------------------!
subroutine load_ed_ecosystem_params()

   use ed_max_dims , only : n_pft               ! ! intent(in)
   use pft_coms    , only : include_these_pft   & ! intent(in)
                          , include_pft         & ! intent(out)
                          , include_pft_ag      & ! intent(out)
                          , C2B                 & ! intent(out)
                          , frost_mort          & ! intent(out)
                          , grass_pft           & ! intent(out)
                          , pft_name16          ! ! intent(out)
   use disturb_coms, only : ianth_disturb       ! ! intent(in)

   implicit none
   !----- Arguments -----------------------------------------------------------------------!
   integer :: p
   !---------------------------------------------------------------------------------------!
   !----- Loading several parameters ------------------------------------------------------!
   call init_decomp_params()
   call init_ff_coms()
   call init_disturb_params()
   call init_physiology_params()
   call init_met_params()
   call init_lapse_params()
   call init_hydro_coms()
   call init_soil_coms()
   call init_phen_coms()
   call init_ed_misc_coms()

   !---------------------------------------------------------------------------------------!
   !      Main table of Plant functional types.  If you add some PFT, please make sure     !
   ! that you assign values for all PFT-dependent variables.  Below is a summary table of  !
   ! the main characteristics of the currently available PFTs.                             !
   !---------------------------------------------------------------------------------------!
   !  PFT | Name                                       | Grass   | Tropical | agriculture? !
   !------+--------------------------------------------+---------+----------+--------------!
   !    1 | C4 grass                                   |     yes |      yes |          yes !
   !    2 | Early tropical                             |      no |      yes |           no !
   !    3 | Mid tropical                               |      no |      yes |           no !
   !    4 | Late tropical                              |      no |      yes |           no !
   !    5 | C3 grass                                   |     yes |       no |          yes !
   !    6 | Northern pines                             |      no |       no |           no !
   !    7 | Southern pines                             |      no |       no |           no !
   !    8 | Late conifers                              |      no |       no |           no !
   !    9 | Early temperate deciduous                  |      no |       no |           no !
   !   10 | Mid temperate deciduous                    |      no |       no |           no !
   !   11 | Late temperate deciduous                   |      no |       no |           no !
   !   12 | C3 pasture                                 |     yes |       no |          yes !
   !   13 | C3 crop (e.g.,wheat, rice, soybean)        |     yes |       no |          yes !
   !   14 | C4 pasture                                 |     yes |      yes |          yes !
   !   15 | C4 crop (e.g.,corn/maize)                  |     yes |      yes |          yes !
   !   16 | Subtropical C3 grass                       |     yes |      yes |          yes !
   !   17 | Araucaria                                  |      no |      yes |           no !
   !   21 | Aspen                                      |      no |       no |           no !
   !   22 | Black Spruce                               |      no |       no |           no !
   !30-32 | Savanna                                    |      no |      yes |           no !
   !33-35 | Tropical Forest                            |      no |      yes |           no !
   !------+--------------------------------------------+---------+----------+--------------!

   !----- Name the PFTs (no spaces, please). ----------------------------------------------!
   pft_name16( 1) = 'C4_grass        '
   pft_name16( 2) = 'Early_tropical  '
   pft_name16( 3) = 'Mid_tropical    '
   pft_name16( 4) = 'Late_tropical   '
   pft_name16( 5) = 'C3_grass        '
   pft_name16( 6) = 'North_pine      '
   pft_name16( 7) = 'South_pine      '
   pft_name16( 8) = 'Late_conifer    '
   pft_name16( 9) = 'Early_hardwood  '
   pft_name16(10) = 'Mid_hardwood    '
   pft_name16(11) = 'Late_hardwood   '
   pft_name16(12) = 'C3_pasture      '
   pft_name16(13) = 'C3_crop         '
   pft_name16(14) = 'C4_pasture      '
   pft_name16(15) = 'C4_crop         '
   pft_name16(16) = 'Subtrop_C3_grass'
   pft_name16(17) = 'Araucaria       '
   pft_name16(18) = 'Early Trop LD   '
   pft_name16(19) = 'Mid Trop LD     '
   pft_name16(20) = 'Late Trop LD    '
   pft_name16(21) = 'Trembling Aspen '
   pft_name16(22) = 'Black Spruce    '
   pft_name16(23) = 'XXX             '
   pft_name16(24) = 'XXX             '
   pft_name16(25) = 'XXX             '
   pft_name16(26) = 'Xiangtao        '
   pft_name16(27) = 'Xiangtao        '
   pft_name16(28) = 'Xiangtao        '
   pft_name16(29) = 'XXX             '
   pft_name16(30) = 'Savanna         '
   pft_name16(31) = 'Savanna         '
   pft_name16(32) = 'Savanna         '
   pft_name16(33) = 'Tropical Forest '
   pft_name16(34) = 'Tropical Forest '
   pft_name16(35) = 'Tropical Forest '

   !----- Define the grass PFTs -----------------------------------------------------------!
   grass_pft=huge(1)
   grass_pft(1)=1
   grass_pft(2)=5
   grass_pft(3)=12
   grass_pft(4)=13
   grass_pft(5)=14
   grass_pft(6)=15
   grass_pft(7)=16

   !---------------------------------------------------------------------------------------!
   !    Include_pft: flag specifying to whether you want to include a plant functional     !
   !                 type (T) or whether you want it excluded (F) from the simulation.     !
   !---------------------------------------------------------------------------------------!
   include_pft    = .false.
   include_pft_ag = .false.
   do p=1,n_pft
      if (include_these_pft(p) >  0 .and. include_these_pft(p) <= n_pft) then
         include_pft(include_these_pft(p)) = .true.
      end if
   end do

   !----- Grasses can grow anywhere, including agricultural patches -----------------------!
   p=1
   do while (grass_pft(p) > 0 .and. grass_pft(p) <= n_pft)
      include_pft_ag(grass_pft(p)) = include_pft(grass_pft(p))
      p = p+1
   end do
   if (count(include_pft_ag) == 0 .and. ianth_disturb == 1) then
      call warning ('No grass included in include_these_pft,'//                            &
                    ' you should have at least one kind of grass...'                       &
                   ,'load_ecosystem_params','ed_params.f90')
   end if

   !----- Assign many PFT-dependent parameters. -------------------------------------------!
   call init_pft_resp_params()
   call init_pft_alloc_params()
   call init_pft_photo_params()
   call init_pft_mort_params()
   call init_pft_nitro_params()
   call init_pft_leaf_params()
   call init_pft_repro_params()
   call init_pft_derived_params()

   !----- This must be done after defining some PFT parameters. ---------------------------!
   call init_can_air_params()
   call init_can_rad_params()
   call init_can_lyr_params()

   !---------------------------------------------------------------------------------------!
   !     This should be always the last one, since it depends on variables assigned in     !
   ! the previous init_????_params.                                                        !
   !---------------------------------------------------------------------------------------!
   call init_rk4_params()
   !---------------------------------------------------------------------------------------!

   return
end subroutine load_ed_ecosystem_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!     This subroutine assigns values for some variables that are in ed_misc_coms, which    !
! wouldn't fit in any of the other categories.                                             !
!------------------------------------------------------------------------------------------!
subroutine init_ed_misc_coms
   use ed_max_dims  , only : n_pft               & ! intent(in)
                           , n_dbh               & ! intent(in)
                           , n_age               ! ! intent(in)
   use consts_coms  , only : erad                & ! intent(in)
                           , pio180              ! ! intent(in)
   use ed_misc_coms , only : burnin              & ! intent(out)
                           , outputMonth         & ! intent(out)
                           , restart_target_year & ! intent(out)
                           , use_target_year     & ! intent(out)
                           , maxage              & ! intent(out)
                           , dagei               & ! intent(out)
                           , maxdbh              & ! intent(out)
                           , ddbhi               & ! intent(out)
                           , vary_elev           & ! intent(out)
                           , vary_hyd            & ! intent(out)
                           , vary_rad            & ! intent(out)
                           , max_thsums_dist     & ! intent(out)
                           , max_poihist_dist    & ! intent(out)
                           , max_poi99_dist      ! ! intent(out)
   implicit none


   !----- Flags that allow components of subgrid heterogeneity to be turned on/off --------!
   vary_elev = 1
   vary_rad  = 1
   vary_hyd  = 1
   !---------------------------------------------------------------------------------------!


   !----- Number of years to ignore demography when starting a run. -----------------------!
   burnin = 0
   !---------------------------------------------------------------------------------------!


   !----- Month to output the yearly files. -----------------------------------------------!
   outputMonth = 6
   !---------------------------------------------------------------------------------------!


   !----- Year to read when parsing pss/css with multiple years. --------------------------!
   restart_target_year = 2000
   !---------------------------------------------------------------------------------------!


   !----- Flag specifying whether to search for a target year in pss/css. -----------------!
   use_target_year = 0    
   !---------------------------------------------------------------------------------------!



   !----- Maximum age [yr] to split into classes. -----------------------------------------!
   maxage = 200.
   !---------------------------------------------------------------------------------------!



   !----- Maximum DBH [cm] to be split into classes. --------------------------------------!
   maxdbh = 100.
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     The inverse of bin classes will depend on max??? and n_???, leaving one class for !
   ! when the number exceeds the maximum.                                                  !
   !---------------------------------------------------------------------------------------!
   dagei = real(n_age-1) / maxage
   ddbhi = real(n_dbh-1) / maxdbh
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !    Maximum distance to the current polygon that we still consider the file grid point !
   ! to be representative of the polygon.  The value below is 1.25 degree at the Equator.  !
   !---------------------------------------------------------------------------------------!
   max_thsums_dist    = 1.25 * erad * pio180
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      Alternative method for mixing 1 grid and POI's.  Only use the grid if their is   !
   ! NOT an POI  within a user specified resolution.  Remember, this assumes there is only !
   ! 1 gridded file, and it is the first file when ied_init_mode is set to 99  (Developer  !
   ! use only).                                                                            !
   !---------------------------------------------------------------------------------------!
   max_poi99_dist     = 5.0 * erad * pio180
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      This variable is used for the history start initialisation.  This sets the       !
   ! maximum acceptable distance between the expected polygon and the polygon found in the !
   ! history file.  Units: m.                                                              !
   !---------------------------------------------------------------------------------------!
   max_poihist_dist   = 250.
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_ed_misc_coms
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!     This subroutine defines the minimum and maximum acceptable values in the meteoro-    !
! logical forcing.                                                                         !
!------------------------------------------------------------------------------------------!
subroutine init_met_params()

   use met_driver_coms, only : rshort_min   & ! intent(out)
                             , rshort_max   & ! intent(out)
                             , rlong_min    & ! intent(out)
                             , rlong_max    & ! intent(out)
                             , dt_radinterp & ! intent(out)
                             , atm_tmp_min  & ! intent(out)
                             , atm_tmp_max  & ! intent(out)
                             , atm_shv_min  & ! intent(out)
                             , atm_shv_max  & ! intent(out)
                             , atm_rhv_min  & ! intent(out)
                             , atm_rhv_max  & ! intent(out)
                             , atm_co2_min  & ! intent(out)
                             , atm_co2_max  & ! intent(out)
                             , prss_min     & ! intent(out)
                             , prss_max     & ! intent(out)
                             , pcpg_min     & ! intent(out)
                             , pcpg_max     & ! intent(out)
                             , vels_min     & ! intent(out)
                             , vels_max     & ! intent(out)
                             , geoht_min    & ! intent(out)
                             , geoht_max    ! ! intent(out)

   !----- Minimum and maximum acceptable shortwave radiation [W/m]. ----------------------!
   rshort_min  = 0.
   rshort_max  = 1500.
   !----- Minimum and maximum acceptable longwave radiation [W/m]. -----------------------!
   rlong_min   = 40.
   rlong_max   = 600.
   !----- Minimum and maximum acceptable air temperature    [   K]. -----------------------!
   atm_tmp_min = 184.     ! Lowest temperature ever measured, in Vostok Basin, Antarctica
   atm_tmp_max = 331.     ! Highest temperature ever measured, in El Azizia, Libya
   !----- Minimum and maximum acceptable air specific humidity [kg_H2O/kg_air]. -----------!
   atm_shv_min = 1.e-6    ! That corresponds to a relative humidity of 0.1% at 1000hPa
   atm_shv_max = 10.2e-2   ! That is saturated air at 55degC.  That corresponds to a dew point of 32C at 1000hPa.
   !----- Minimum and maximum acceptable CO2 mixing ratio [mol/mol]. ---------------------!
   atm_co2_min = 100.     ! 
   atm_co2_max = 1100.    ! 
   !----- Minimum and maximum acceptable pressure [Pa]. -----------------------------------!
   prss_min =  45000. ! It may crash if you run a simulation in Mt. Everest.
   prss_max = 110000. ! It may crash if you run a simulation under water.
   !----- Minimum and maximum acceptable precipitation rates [kg/m/s]. -------------------!
   pcpg_min     = 0.0     ! No negative precipitation is allowed
   pcpg_max     = 0.1111  ! This is a precipitation rate of 400mm/hr.
   !----- Minimum and maximum acceptable wind speed [m/s]. --------------------------------!
   vels_min     =  0.0    ! No negative wind is acceptable.
   vels_max     = 85.0    ! Maximum sustained winds recorded during Typhoon Tip (1970).
   !----- Minimum and maximum reference heights [m]. --------------------------------------!
   geoht_min    =   1.0   ! This should be above-canopy measurement, but 1.0 is okay for
                          !     grasslands...
   geoht_max    = 350.0   ! This should be not that much above the canopy, but tall towers
                          !     do exist...
   !---------------------------------------------------------------------------------------!

   !---------------------------------------------------------------------------------------!
   !     Minimum and maximum acceptable relative humidity (fraction).  This is not going   !
   ! cause the simulation to crash, instead it will just impose these numbers to the       !
   ! meteorological forcing.                                                               !
   !---------------------------------------------------------------------------------------!
   atm_rhv_min = 5.e-3 ! 0.5%
   atm_rhv_max = 1.0   ! 100.0%.  Although canopy air space can experience super-
                       !    saturation, we don't allow the air above to be super-saturated.
   !---------------------------------------------------------------------------------------!




   !---------------------------------------------------------------------------------------!
   !     Time step used to perform the daytime average of the secant of the zenith angle.  !
   !---------------------------------------------------------------------------------------!
   dt_radinterp = 30.0    ! Value in seconds.
   !---------------------------------------------------------------------------------------!


   return
end subroutine init_met_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!     This subroutine defines defaults for lapse rates.                                    !
!------------------------------------------------------------------------------------------!
subroutine init_lapse_params()

   use met_driver_coms, only : lapse             & ! intent(out)
                             , atm_tmp_intercept & ! intent(out)
                             , atm_tmp_slope     & ! intent(out)
                             , prec_intercept    & ! intent(out)
                             , prec_slope        & ! intent(out)
                             , humid_scenario    ! ! intent(out)

   lapse%geoht        = 0.0
   lapse%vels         = 0.0
   lapse%atm_tmp      = 0.0
   lapse%atm_theta    = 0.0
   lapse%atm_theiv    = 0.0
   lapse%atm_shv      = 0.0
   lapse%prss         = 0.0
   lapse%pcpg         = 0.0
   lapse%atm_co2      = 0.0
   lapse%rlong        = 0.0
   lapse%nir_beam     = 0.0
   lapse%nir_diffuse  = 0.0
   lapse%par_beam     = 0.0
   lapse%par_diffuse  = 0.0
   lapse%pptnorm      = 0.0

   atm_tmp_intercept = 0.0
   atm_tmp_slope     = 1.0
   prec_intercept    = 0.0
   prec_slope        = 1.0
   humid_scenario    = 0

   return
end subroutine init_lapse_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!    This subroutine will assign some radiation related parameters.                        !
!------------------------------------------------------------------------------------------!
subroutine init_can_rad_params()

   use canopy_radiation_coms , only : leaf_reflect_nir            & ! intent(out)
                                    , leaf_trans_nir              & ! intent(out)
                                    , leaf_scatter_nir            & ! intent(out)
                                    , leaf_reflect_vis            & ! intent(out)
                                    , leaf_trans_vis              & ! intent(out)
                                    , leaf_scatter_vis            & ! intent(out)
                                    , leaf_reflect_vis            & ! intent(out)
                                    , leaf_trans_vis              & ! intent(out)
                                    , leaf_backscatter_vis        & ! intent(out)
                                    , leaf_backscatter_nir        & ! intent(out)
                                    , leaf_emis                   & ! intent(out)
                                    , wood_reflect_nir            & ! intent(out)
                                    , wood_trans_nir              & ! intent(out)
                                    , wood_scatter_nir            & ! intent(out)
                                    , wood_reflect_vis            & ! intent(out)
                                    , wood_trans_vis              & ! intent(out)
                                    , wood_scatter_vis            & ! intent(out)
                                    , wood_reflect_vis            & ! intent(out)
                                    , wood_trans_vis              & ! intent(out)
                                    , wood_backscatter_vis        & ! intent(out)
                                    , wood_backscatter_nir        & ! intent(out)
                                    , wood_emis                   & ! intent(out)
                                    , mubar                       & ! intent(out)
                                    , fvis_beam_def               & ! intent(out)
                                    , fvis_diff_def               & ! intent(out)
                                    , fnir_beam_def               & ! intent(out)
                                    , fnir_diff_def               & ! intent(out)
                                    , rshort_twilight_min         & ! intent(out)
                                    , cosz_min                    & ! intent(out)
                                    , cosz_min8                   ! ! intent(out)
   use consts_coms           , only : pio180                      ! ! intent(out)
   implicit none



   !---------------------------------------------------------------------------------------!
   !     Leaf angle distribution parameter (dimensionless).  Let mu' be the cosine of leaf !
   ! angle and G(mu') be the distribution of mu'.  Then, mubar = (integral from 0 to 1)    !
   ! (d mu'   mu' / G(mu')).  See, for example, Dickinson 1983.                            !
   !---------------------------------------------------------------------------------------!
   mubar                      = 1.0d0 
   !---------------------------------------------------------------------------------------!




   !---------------------------------------------------------------------------------------!
   !      The following parameters are used to split the shortwave radiation into visible  !
   ! and near-infrared radiation.                                                          !
   !---------------------------------------------------------------------------------------!
   fvis_beam_def = 0.43
   fnir_beam_def = 1.0 - fvis_beam_def
   fvis_diff_def = 0.52
   fnir_diff_def = 1.0 - fvis_diff_def
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      Leaf reflectance.                                                                !
   !      Values for temperate PFTs were left as they were.  Tropical and sub-tropical     !
   ! PFTs use the parameters from CLM.  I checked the values against some published and    !
   ! they seem similar at a first glance, at least closer than the original values, which  !
   ! looked like the visible ignoring the green band.                                      !
   !---------------------------------------------------------------------------------------!
   !----- Visible (PAR). ------------------------------------------------------------------!
   leaf_reflect_vis(1)     = 1.3d-1 ! 6.2d-2
   leaf_reflect_vis(2:4)   = 1.3d-1 ! 6.2d-2
   leaf_reflect_vis(5)     = 1.1d-1 ! 6.2d-2
   leaf_reflect_vis(6:11)  = 1.1d-1 ! 1.1d-1
   leaf_reflect_vis(12:13) = 1.1d-1 ! 1.1d-1
   leaf_reflect_vis(14:15) = 1.1d-1 ! 1.1d-1
   leaf_reflect_vis(16)    = 1.3d-1 ! 0.062
   leaf_reflect_vis(17)    = 9.0d-2 ! 9.0d-2
   leaf_reflect_vis(18:20) = 1.3d-1 ! 6.2d-2
   leaf_reflect_vis(21:29) = 1.1d-1
   leaf_reflect_vis(23:35)   = 0.061 ! from Milly et al. 2014 same as LM3.1 !1.3d-1 ! 6.2d-2
   									 ! also Castro-Esau et al. 2006
									 ! Avalos et al. 1998
   !----- Near infrared. ------------------------------------------------------------------!
   leaf_reflect_nir(1)     = 6.30d-1
   leaf_reflect_nir(2:4)   = 6.30d-1
   leaf_reflect_nir(5)     = 5.77d-1
   leaf_reflect_nir( 6:11) = 5.77d-1
   leaf_reflect_nir(12:13) = 5.77d-1
   leaf_reflect_nir(14:15) = 5.80d-1
   leaf_reflect_nir(16)    = 6.30d-1
   leaf_reflect_nir(17)    = 4.60d-1
   leaf_reflect_nir(18:20) = 6.30d-1
   leaf_reflect_nir(21:29) = 5.77d-1
   leaf_reflect_nir(23:35)   = 5.20d-1    !  Castro-Esau et al. 2006, Kalascka et al. 2005
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      Wood reflectance, using CLM parameters.                                          !
   !---------------------------------------------------------------------------------------!
   !----- Visible (PAR). ------------------------------------------------------------------!
   wood_reflect_vis(1)     = 3.60d-1
   wood_reflect_vis(2:4)   = 1.60d-1
   wood_reflect_vis(5)     = 3.60d-1
   wood_reflect_vis(6:11)  = 1.60d-1
   wood_reflect_vis(12:13) = 3.60d-1
   wood_reflect_vis(14:15) = 3.60d-1
   wood_reflect_vis(16)    = 3.60d-1
   wood_reflect_vis(17)    = 1.60d-2
   wood_reflect_vis(18:20) = 1.60d-1
   wood_reflect_vis(21:35) = 1.60d-1
   !----- Near infrared. ------------------------------------------------------------------!
   wood_reflect_nir(1)     = 5.80d-1
   wood_reflect_nir(2:4)   = 3.90d-1
   wood_reflect_nir(5)     = 5.80d-1
   wood_reflect_nir( 6:11) = 3.90d-1
   wood_reflect_nir(12:13) = 5.80d-1
   wood_reflect_nir(14:15) = 5.80d-1
   wood_reflect_nir(16)    = 5.80d-1
   wood_reflect_nir(17)    = 3.90d-1
   wood_reflect_nir(18:20) = 3.90d-1
   wood_reflect_nir(21:35) = 3.90d-1
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      Leaf transmittance.                                                              !
   !      Values for temperate PFTs were left as they were.  Tropical and sub-tropical     !
   ! PFTs use the parameters from CLM.  I checked the values against some published and    !
   ! they seem similar at a first glance, at least closer than the original values, which  !
   ! looked like the visible ignoring the green band.                                      !
   !---------------------------------------------------------------------------------------!
   !----- Visible (PAR). ------------------------------------------------------------------!
   leaf_trans_vis(    1) = 2.80d-2  ! 7.00d-2
   leaf_trans_vis(  2:4) = 2.80d-2  ! 5.00d-2
   leaf_trans_vis(    5) = 1.60d-1  ! 0.160
   leaf_trans_vis( 6:11) = 1.60d-1  ! 0.160
   leaf_trans_vis(12:13) = 1.60d-1  ! 0.160
   leaf_trans_vis(14:15) = 7.00d-2  ! 0.028
   leaf_trans_vis(   16) = 2.80d-2  ! 7.00d-2
   leaf_trans_vis(   17) = 5.00d-2  ! 0.160
   leaf_trans_vis(18:20) = 2.80d-2  ! 5.00d-2
   leaf_trans_vis(21:29) = 1.60d-1
   leaf_trans_vis(23:35) = 6.0d-2  ! Castro-Esau et al. 2006
   !----- Near infrared. ------------------------------------------------------------------!
   leaf_trans_nir(    1) = 2.48d-1
   leaf_trans_nir(  2:4) = 2.48d-1
   leaf_trans_nir(    5) = 2.48d-1
   leaf_trans_nir( 6:11) = 2.48d-1
   leaf_trans_nir(12:13) = 2.48d-1
   leaf_trans_nir(14:15) = 2.48d-1
   leaf_trans_nir(   16) = 2.48d-1
   leaf_trans_nir(   17) = 1.80d-1
   leaf_trans_nir(18:20) = 2.48d-1
   leaf_trans_nir(21:35) = 2.48d-1
   leaf_trans_nir(23:35) = 4.0d-1  ! Castro-Esau et al. 2006
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      Wood transmittance, using the parameters from CLM.                               !
   !---------------------------------------------------------------------------------------!
   !----- Visible (PAR). ------------------------------------------------------------------!
   wood_trans_vis(    1) = 2.80d-2
   wood_trans_vis(  2:4) = 1.00d-3
   wood_trans_vis(    5) = 2.80d-2
   wood_trans_vis( 6:11) = 1.00d-3
   wood_trans_vis(12:13) = 2.20d-1
   wood_trans_vis(14:15) = 2.20d-1
   wood_trans_vis(   16) = 2.80d-2
   wood_trans_vis(   17) = 1.00d-3
   wood_trans_vis(18:20) = 1.00d-3
   wood_trans_vis(21:35) = 1.00d-3
   !----- Near infrared. ------------------------------------------------------------------!
   wood_trans_nir(    1) = 2.48d-1
   wood_trans_nir(  2:4) = 1.00d-3
   wood_trans_nir(    5) = 2.48d-1
   wood_trans_nir( 6:11) = 1.00d-3
   wood_trans_nir(12:13) = 2.48d-1
   wood_trans_nir(14:15) = 2.48d-1
   wood_trans_nir(   16) = 2.48d-1
   wood_trans_nir(   17) = 1.00d-3
   wood_trans_nir(18:20) = 1.00d-3
   wood_trans_nir(21:35) = 1.00d-3
   !---------------------------------------------------------------------------------------!




   !---------------------------------------------------------------------------------------!
   !      Scattering coefficient.  For all PFTs this is just the sum of reflectance and    !
   ! transmittance.                                                                        !
   !---------------------------------------------------------------------------------------!
   !----- Visible (PAR). ------------------------------------------------------------------!
   leaf_scatter_vis(1:35) = leaf_reflect_vis(1:35) + leaf_trans_vis(1:35) 
   wood_scatter_vis(1:35) = wood_reflect_vis(1:35) + wood_trans_vis(1:35)
   !----- Near infrared. ------------------------------------------------------------------!
   leaf_scatter_nir(1:35) = leaf_reflect_nir(1:35) + leaf_trans_nir(1:35)
   wood_scatter_nir(1:35) = wood_reflect_nir(1:35) + wood_trans_nir(1:35)
   !---------------------------------------------------------------------------------------!




   !---------------------------------------------------------------------------------------!
   !      Back-scattering coefficients.                                                    !
   !---------------------------------------------------------------------------------------!
   !----- Visible (PAR). ------------------------------------------------------------------!
   leaf_backscatter_vis(1:35) = (2.d0 * leaf_reflect_vis(1:35) - leaf_trans_vis(1:35))     &
                              / (3.d0 * leaf_scatter_vis(1:35))
   wood_backscatter_vis(1:35) = (2.d0 * wood_reflect_vis(1:35) - wood_trans_vis(1:35))     &
                              / (3.d0 * wood_scatter_vis(1:35))
   !----- Near infrared. ------------------------------------------------------------------!
   leaf_backscatter_nir(1:35) = (2.d0 * leaf_reflect_nir(1:35) - leaf_trans_nir(1:35))     &
                              / (3.d0 * leaf_scatter_nir(1:35))
   wood_backscatter_nir(1:35) = (2.d0 * wood_reflect_nir(1:35) - wood_trans_nir(1:35))     &
                              / (3.d0 * wood_scatter_nir(1:35))
   !---------------------------------------------------------------------------------------!






   !---------------------------------------------------------------------------------------!
   !      Emissivity.                                                                      !
   !---------------------------------------------------------------------------------------!
   leaf_emis(1)     = 9.60d-1
   leaf_emis(2:4)   = 9.50d-1
   leaf_emis(5)     = 9.60d-1
   leaf_emis(6:8)   = 9.70d-1
   leaf_emis(9:11)  = 9.50d-1
   leaf_emis(12:15) = 9.60d-1
   leaf_emis(16)    = 9.60d-1
   leaf_emis(17)    = 9.70d-1
   leaf_emis(18:20) = 9.50d-1
   leaf_emis(21)    = 9.50d-1
   leaf_emis(22)    = 9.70d-1
   leaf_emis(23:35) = 9.50d-1

   wood_emis(1)     = 9.60d-1
   wood_emis(2:4)   = 9.00d-1
   wood_emis(5)     = 9.60d-1
   wood_emis(6:8)   = 9.00d-1
   wood_emis(9:11)  = 9.00d-1
   wood_emis(12:15) = 9.60d-1
   wood_emis(16)    = 9.60d-1
   wood_emis(17)    = 9.00d-1
   wood_emis(18:20) = 9.00d-1
   wood_emis(21)    = 9.00d-1
   wood_emis(22)    = 9.00d-1
   wood_emis(23:35) = 9.00d-1
   !---------------------------------------------------------------------------------------!




   !---------------------------------------------------------------------------------------!
   !     These variables are the thresholds for things that should be computed during the  !
   ! day time hours only.                                                                  !
   !---------------------------------------------------------------------------------------!
   rshort_twilight_min = 0.5
   cosz_min            = cos(89.*pio180) !cos(89.5*pio180)
   cosz_min8           = dble(cosz_min)
   !---------------------------------------------------------------------------------------!


   return
end subroutine init_can_rad_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!    This subroutine will assign some canopy air related parameters.                       !
!------------------------------------------------------------------------------------------!
subroutine init_can_air_params()
   use consts_coms    , only : onethird              & ! intent(in)
                             , twothirds             & ! intent(in)
                             , vonk                  ! ! intent(in)
   use pft_coms       , only : hgt_min               & ! intent(in)
                             , hgt_max               ! ! intent(in)
   use canopy_air_coms, only : i_blyr_condct         & ! intent(in)
                             , isfclyrm              & ! intent(in)
                             , ustmin                & ! intent(in)
                             , ggfact                & ! intent(in)
                             , gamm                  & ! intent(in)
                             , gamh                  & ! intent(in)
                             , tprandtl              & ! intent(in)
                             , vkopr                 & ! intent(in)
                             , vh2vr                 & ! intent(in)
                             , vh2dh                 & ! intent(in)
                             , ribmax                & ! intent(out)
                             , leaf_drywhc           & ! intent(out)
                             , leaf_maxwhc           & ! intent(out)
                             , rb_inter              & ! intent(out)
                             , rb_slope              & ! intent(out)
                             , veg_height_min        & ! intent(out)
                             , minimum_canopy_depth  & ! intent(out)
                             , minimum_canopy_depth8 & ! intent(out)
                             , exar                  & ! intent(out)
                             , covr                  & ! intent(out)
                             , ugbmin                & ! intent(out)
                             , ubmin                 & ! intent(out)
                             , exar8                 & ! intent(out)
                             , ez                    & ! intent(out)
                             , ustmin8               & ! intent(out)
                             , ugbmin8               & ! intent(out)
                             , ubmin8                & ! intent(out)
                             , ggfact8               & ! intent(out)
                             , ez8                   & ! intent(out)
                             , vh2dh8                & ! intent(out)
                             , cdrag0                & ! intent(out)
                             , pm0                   & ! intent(out)
                             , c1_m97                & ! intent(out)
                             , c2_m97                & ! intent(out)
                             , c3_m97                & ! intent(out)
                             , kvwake                & ! intent(out)
                             , alpha_m97             & ! intent(out)
                             , alpha_mw99            & ! intent(out)
                             , gamma_mw99            & ! intent(out)
                             , nu_mw99               & ! intent(out)
                             , infunc                & ! intent(out)
                             , cdrag08               & ! intent(out)
                             , pm08                  & ! intent(out)
                             , c1_m978               & ! intent(out)
                             , c2_m978               & ! intent(out)
                             , c3_m978               & ! intent(out)
                             , kvwake8               & ! intent(out)
                             , alpha_m97_8           & ! intent(out)
                             , alpha_mw99_8          & ! intent(out)
                             , gamma_mw99_8          & ! intent(out)
                             , nu_mw99_8             & ! intent(out)
                             , infunc_8              & ! intent(out)
                             , bl79                  & ! intent(out)
                             , csm                   & ! intent(out)
                             , csh                   & ! intent(out)
                             , dl79                  & ! intent(out)
                             , bbeta                 & ! intent(out)
                             , abh91                 & ! intent(out)
                             , bbh91                 & ! intent(out)
                             , cbh91                 & ! intent(out)
                             , dbh91                 & ! intent(out)
                             , ebh91                 & ! intent(out)
                             , fbh91                 & ! intent(out)
                             , cod                   & ! intent(out)
                             , bcod                  & ! intent(out)
                             , fm1                   & ! intent(out)
                             , ate                   & ! intent(out)
                             , atetf                 & ! intent(out)
                             , z0moz0h               & ! intent(out)
                             , z0hoz0m               & ! intent(out)
                             , bl798                 & ! intent(out)
                             , csm8                  & ! intent(out)
                             , csh8                  & ! intent(out)
                             , dl798                 & ! intent(out)
                             , bbeta8                & ! intent(out)
                             , gamm8                 & ! intent(out)
                             , gamh8                 & ! intent(out)
                             , ribmax8               & ! intent(out)
                             , tprandtl8             & ! intent(out)
                             , vkopr8                & ! intent(out)
                             , abh918                & ! intent(out)
                             , bbh918                & ! intent(out)
                             , cbh918                & ! intent(out)
                             , dbh918                & ! intent(out)
                             , ebh918                & ! intent(out)
                             , fbh918                & ! intent(out)
                             , cod8                  & ! intent(out)
                             , bcod8                 & ! intent(out)
                             , fm18                  & ! intent(out)
                             , ate8                  & ! intent(out)
                             , atetf8                & ! intent(out)
                             , z0moz0h8              & ! intent(out)
                             , z0hoz0m8              & ! intent(out)
                             , aflat_turb            & ! intent(out)
                             , aflat_lami            & ! intent(out)
                             , bflat_turb            & ! intent(out)
                             , bflat_lami            & ! intent(out)
                             , nflat_turb            & ! intent(out)
                             , nflat_lami            & ! intent(out)
                             , mflat_turb            & ! intent(out)
                             , mflat_lami            & ! intent(out)
                             , ocyli_turb            & ! intent(out)
                             , ocyli_lami            & ! intent(out)
                             , acyli_turb            & ! intent(out)
                             , acyli_lami            & ! intent(out)
                             , bcyli_turb            & ! intent(out)
                             , bcyli_lami            & ! intent(out)
                             , ncyli_turb            & ! intent(out)
                             , ncyli_lami            & ! intent(out)
                             , mcyli_turb            & ! intent(out)
                             , mcyli_lami            & ! intent(out)
                             , beta_r1               & ! intent(out)
                             , beta_r2               & ! intent(out)
                             , beta_re0              & ! intent(out)
                             , beta_g1               & ! intent(out)
                             , beta_g2               & ! intent(out)
                             , beta_gr0              & ! intent(out)
                             , aflat_turb8           & ! intent(out)
                             , aflat_lami8           & ! intent(out)
                             , bflat_turb8           & ! intent(out)
                             , bflat_lami8           & ! intent(out)
                             , nflat_turb8           & ! intent(out)
                             , nflat_lami8           & ! intent(out)
                             , mflat_turb8           & ! intent(out)
                             , mflat_lami8           & ! intent(out)
                             , ocyli_turb8           & ! intent(out)
                             , ocyli_lami8           & ! intent(out)
                             , acyli_turb8           & ! intent(out)
                             , acyli_lami8           & ! intent(out)
                             , bcyli_turb8           & ! intent(out)
                             , bcyli_lami8           & ! intent(out)
                             , ncyli_turb8           & ! intent(out)
                             , ncyli_lami8           & ! intent(out)
                             , mcyli_turb8           & ! intent(out)
                             , mcyli_lami8           & ! intent(out)
                             , beta_r18              & ! intent(out)
                             , beta_r28              & ! intent(out)
                             , beta_re08             & ! intent(out)
                             , beta_g18              & ! intent(out)
                             , beta_g28              & ! intent(out)
                             , beta_gr08             & ! intent(out)
                             , ggsoil0               & ! intent(out)
                             , kksoil                & ! intent(out)
                             , ggsoil08              & ! intent(out)
                             , kksoil8               ! ! intent(out)
   implicit none
   !----- Arguments. ----------------------------------------------------------------------!
   integer :: ican
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !    Minimum leaf water content to be considered.  Values smaller than this will be     !
   ! flushed to zero.  This value is in kg/[m2 tree], so it will be scaled by (LAI+WAI)    !
   ! where needed be.                                                                      !
   !---------------------------------------------------------------------------------------!
   leaf_drywhc = 5.e-4 * leaf_maxwhc
   !---------------------------------------------------------------------------------------!

   !---------------------------------------------------------------------------------------!
   !      Variables to define the vegetation aerodynamic resistance.  They are currently   !
   ! not PFT dependent.                                                                    !
   !---------------------------------------------------------------------------------------!
   select case (i_blyr_condct)
   case (-1)
      rb_slope =  25.0
      rb_inter =   0.0
   case default
      rb_slope =   0.0
      rb_inter =   1.e9
   end select

   !---------------------------------------------------------------------------------------!
   ! veg_height_min       - This is the minimum vegetation height allowed [m].  Vegetation !
   !                        height is used to calculate drag coefficients and patch        !
   !                        roughness.                                                     !
   ! minimum_canopy_depth - This is the minimum canopy depth allowed [m].  Canopy depth    !
   !                        is used to calculate the heat and moisture storage capacity in !
   !                        the canopy air space.                                          !
   !---------------------------------------------------------------------------------------!
   veg_height_min        = minval(hgt_min) ! alternative: minval(hgt_min) 
   minimum_canopy_depth  = veg_height_min  ! alternative: minval(hgt_min) 

   !----- This is the dimensionless exponential wind atenuation factor. -------------------!
   exar  = 2.5

   !----- This is the scaling factor of tree area index (not sure if it is used...) -------!
   covr = 2.16


   !---------------------------------------------------------------------------------------!
   !      Parameters for surface layer models.                                             !
   !---------------------------------------------------------------------------------------!
   !----- This is the minimum wind speed for boundary layer conductivity. -----------------!
   ugbmin    = 0.10
   !----- This is the minimum wind scale under stable and unstable conditions. ------------!
   ubmin     = 0.10
   !---------------------------------------------------------------------------------------!

   !----- Louis (1979) model. -------------------------------------------------------------!
   bl79        = 5.0    ! b prime parameter
   csm         = 7.5    ! C* for momentum (eqn. 20, not co2 char. scale)
   csh         = 5.0    ! C* for heat (eqn.20, not co2 char. scale)
   dl79        = 5.0    ! ???
   !----- Oncley and Dudhia (1995) model. -------------------------------------------------!
   bbeta       = 5.0           ! Beta 
   !----- Beljaars and Holtslag (1991) model. ---------------------------------------------!
   abh91       = -1.00         ! -a from equation  (28) and (32)
   bbh91       = -twothirds    ! -b from equation  (28) and (32)
   cbh91       =  5.0          !  c from equations (28) and (32)
   dbh91       =  0.35         !  d from equations (28) and (32)
   ebh91       = -twothirds    ! - factor multiplying a*zeta in equation (32)
   fbh91       =  1.50         ! exponent in equation (32)
   cod         = cbh91/dbh91   ! c/d
   bcod        = bbh91 * cod   ! b*c/d
   fm1         = fbh91 - 1.0   ! f-1
   ate         = abh91 * ebh91 ! a * e
   atetf       = ate   * fbh91 ! a * e * f
   z0moz0h     = 1.0           ! z0(M)/z0(h)
   z0hoz0m     = 1. / z0moz0h  ! z0(M)/z0(h)
   !---------------------------------------------------------------------------------------!


   
   !----- Legacy variable, we can probably remove it. -------------------------------------!
   ez  = 0.172
   !---------------------------------------------------------------------------------------!







   !---------------------------------------------------------------------------------------!
   !      Parameters for the aerodynamic resistance between the leaf (flat surface) and    !
   ! wood (kind of cylinder surface), and the canopy air space.  These are the A, B, n,    !
   ! and m parameters that define the Nusselt number for forced and free convection, at    !
   ! equations 10.7 and 10.9.  The parameters are found at the appendix A, table A.5(a)    !
   ! and A.5(b).                                                                           !
   !                                                                                       !
   ! M08 - Monteith, J. L., M. H. Unsworth, 2008. Principles of Environmental Physics,     !
   !       3rd. edition, Academic Press, Amsterdam, 418pp.  (Mostly Chapter 10).           !
   !---------------------------------------------------------------------------------------!
   aflat_lami = 0.600    ! A (forced convection), laminar   flow
   nflat_lami = 0.500    ! n (forced convection), laminar   flow
   aflat_turb = 0.032    ! A (forced convection), turbulent flow
   nflat_turb = 0.800    ! n (forced convection), turbulent flow
   bflat_lami = 0.500    ! B (free   convection), laminar   flow
   mflat_lami = 0.250    ! m (free   convection), laminar   flow
   bflat_turb = 0.130    ! B (free   convection), turbulent flow
   mflat_turb = onethird ! m (free   convection), turbulent flow
   ocyli_lami = 0.320    ! intercept (forced convection), laminar   flow
   acyli_lami = 0.510    ! A (forced convection), laminar   flow
   ncyli_lami = 0.520    ! n (forced convection), laminar   flow
   ocyli_turb = 0.000    ! intercept (forced convection), turbulent flow
   acyli_turb = 0.240    ! A (forced convection), turbulent flow
   ncyli_turb = 0.600    ! n (forced convection), turbulent flow
   bcyli_lami = 0.480    ! B (free   convection), laminar   flow
   mcyli_lami = 0.250    ! m (free   convection), laminar   flow
   bcyli_turb = 0.090    ! B (free   convection), turbulent flow
   mcyli_turb = onethird ! m (free   convection), turbulent flow
   !---------------------------------------------------------------------------------------!




   !---------------------------------------------------------------------------------------!
   !     Both free and forced convection tend to underestimate the Nusselt number under    !
   ! different conditions.  Based on M08 review on the subject, I wrote the following      !
   ! functional form to expand the Nusselt number by a factor beta:                        !
   ! - beta_forced = R1 + R2 * tanh[log(Re/Re0)]                                           !
   ! - beta_free   = G1 + G2 * tanh[log(Gr/Gr0)]                                           !
   !     The values of beta change depending on the boundary layer conductance method.     !
   ! Currently only the forced convection varies, as we believe that the Reynolds number   !
   ! is the one that influences the most.                                                  ! 
   !---------------------------------------------------------------------------------------!
   select case (i_blyr_condct)
   case (-1,0)
      beta_r1  = 1.
      beta_r2  = 0.
      beta_re0 = 2000.
      beta_g1  = 1.
      beta_g2  = 0.
      beta_gr0 = 100000.
   case (1)
      beta_r1  =   7./4.
      beta_r2  =   3./4.
      beta_re0 =   2000.
      beta_g1  =   3./2.
      beta_g2  =  -1./2.
      beta_gr0 = 100000.
   case (2)
      beta_r1  =   11./2.
      beta_r2  =   9. /2.
      beta_re0 =   2000.
      beta_g1  =   3./2.
      beta_g2  =  -1./2.
      beta_gr0 = 100000.
   end select
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     Define the variables that are going to be used by Massman (1997) and Massman and  !
   ! Weil (1999).  Full reference:                                                         !
   !                                                                                       !
   ! Massman, W. J., 1997: An analytical one-dimensional model of momentum transfer by     !
   !    vegetation of arbitrary structure.  Boundary Layer Meteorology, 83, 407-421.       !
   !                                                                                       !
   ! Massman, W. J., and J. C. Weil, 1999: An analytical one-dimension second-order clos-  !
   !    ure model turbulence statistics and the Lagrangian time scale within and above     !
   !    plant canopies of arbitrary structure.  Boundary Layer Meteorology, 91, 81-107.    !
   !---------------------------------------------------------------------------------------!
   !----- Fluid drag coefficient for turbulent flow in leaves. ----------------------------!
   cdrag0    = 0.2
   !----- Sheltering factor of fluid drag on canopies. ------------------------------------!
   pm0       = 1.0
   !----- Surface drag parameters (Massman 1997). -----------------------------------------!
   c1_m97    = 0.320 
   c2_m97    = 0.264
   c3_m97    = 15.1
   !----- Eddy diffusivity due to Von Karman Wakes in gravity flows. ----------------------!
   kvwake    = 0.001
   !---------------------------------------------------------------------------------------!
   !     Alpha factors to produce the profile of sheltering factor and within canopy drag, !
   ! as suggested by Massman (1997) and Massman and Weil (1999).                           !
   !---------------------------------------------------------------------------------------!
   alpha_m97  = 5.00
   alpha_mw99 = 0.03
   !---------------------------------------------------------------------------------------!
   !      Parameter to represent the roughness sublayer effect.  According to Massman,     !
   ! assuming this to be zero means that the sublayer effects will be ignored.  Otherwise  !
   ! Raupach (1994) tried values up to 0.316.                                              !
   !---------------------------------------------------------------------------------------!
   infunc    = 0.193
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !  Gamma and nu are the parameters that close equation 10 in Massman and Weil (1999).   !
   !  VERY IMPORTANT: If you mess with gamma, you must recompute nu!                       !
   !---------------------------------------------------------------------------------------!
   gamma_mw99 = (/2.4, 1.9, 1.25/)
   nu_mw99    = (/0.3024,3.4414,36.1476/)
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !   Soil conductance terms, from:                                                       !
   !                                                                                       !
   ! Passerat de Silans, A., 1986: Transferts de masse et de chaleur dans un sol stratifi !
   !     soumis  une excitation amtosphrique naturelle. Comparaison: Modles-exprience. !
   !     Thesis, Institut National Polytechnique de Grenoble. (P86)                        !
   !                                                                                       !
   ! retrieved from:                                                                       !
   ! Mahfouf, J. F., J. Noilhan, 1991: Comparative study of various formulations of        !
   !     evaporation from bare soil using in situ data. J. Appl. Meteorol., 30, 1354-1365. !
   !     (MN91)                                                                            !
   !                                                                                       !
   !     Please notice that the values are inverted because we compute conductance, not    !
   ! resistance.                                                                           !
   !---------------------------------------------------------------------------------------!
   ggsoil0 = 1. / 38113.
   kksoil  = 13.515
   !---------------------------------------------------------------------------------------!



   !----- Set the double precision variables. ---------------------------------------------!
   minimum_canopy_depth8 = dble(minimum_canopy_depth)
   exar8                 = dble(exar                )
   ubmin8                = dble(ubmin               )
   ugbmin8               = dble(ugbmin              )
   ustmin8               = dble(ustmin              )
   ggfact8               = dble(ggfact              )
   ez8                   = dble(ez                  )
   vh2dh8                = dble(vh2dh               )
   bl798                 = dble(bl79                )
   csm8                  = dble(csm                 )
   csh8                  = dble(csh                 )
   dl798                 = dble(dl79                )
   bbeta8                = dble(bbeta               )
   gamm8                 = dble(gamm                )
   gamh8                 = dble(gamh                )
   ribmax8               = dble(ribmax              )
   tprandtl8             = dble(tprandtl            )
   vkopr8                = dble(vkopr               )
   abh918                = dble(abh91               )
   bbh918                = dble(bbh91               )
   cbh918                = dble(cbh91               )
   dbh918                = dble(dbh91               )
   ebh918                = dble(ebh91               )
   fbh918                = dble(fbh91               )
   cod8                  = dble(cod                 )
   bcod8                 = dble(bcod                )
   fm18                  = dble(fm1                 )
   ate8                  = dble(ate                 )
   atetf8                = dble(atetf               )
   z0moz0h8              = dble(z0moz0h             )
   z0hoz0m8              = dble(z0hoz0m             )
   aflat_lami8           = dble(aflat_lami          ) 
   nflat_lami8           = dble(nflat_lami          ) 
   aflat_turb8           = dble(aflat_turb          ) 
   nflat_turb8           = dble(nflat_turb          ) 
   bflat_lami8           = dble(bflat_lami          ) 
   mflat_lami8           = dble(mflat_lami          ) 
   bflat_turb8           = dble(bflat_turb          ) 
   mflat_turb8           = dble(mflat_turb          ) 
   ocyli_lami8           = dble(ocyli_lami          ) 
   acyli_lami8           = dble(acyli_lami          ) 
   ncyli_lami8           = dble(ncyli_lami          ) 
   ocyli_turb8           = dble(ocyli_turb          ) 
   acyli_turb8           = dble(acyli_turb          ) 
   ncyli_turb8           = dble(ncyli_turb          ) 
   bcyli_lami8           = dble(bcyli_lami          ) 
   mcyli_lami8           = dble(mcyli_lami          ) 
   bcyli_turb8           = dble(bcyli_turb          ) 
   mcyli_turb8           = dble(mcyli_turb          ) 
   beta_r18              = dble(beta_r1             )
   beta_r28              = dble(beta_r2             )
   beta_re08             = dble(beta_re0            )
   beta_g18              = dble(beta_g1             )
   beta_g28              = dble(beta_g2             )
   beta_gr08             = dble(beta_gr0            )
   cdrag08               = dble(cdrag0              )
   pm08                  = dble(pm0                 )
   c1_m978               = dble(c1_m97              )
   c2_m978               = dble(c2_m97              )
   c3_m978               = dble(c3_m97              )
   kvwake8               = dble(kvwake              )
   alpha_m97_8           = dble(alpha_m97           )
   alpha_mw99_8          = dble(alpha_mw99          )
   gamma_mw99_8          = dble(gamma_mw99          )
   nu_mw99_8             = dble(nu_mw99             )
   infunc_8              = dble(infunc              )
   ggsoil08              = dble(ggsoil0             )
   kksoil8               = dble(kksoil              )
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_can_air_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!     This sub-routine initialises some of the canopy layer variables.  These are used     !
! by sub-routines that need to calculate the canopy properties by layers rather than by    !
! cohorts (or when both must be considered).                                               !
!------------------------------------------------------------------------------------------!
subroutine init_can_lyr_params()
   use canopy_layer_coms, only : ncanlyr                     & ! intent(out)
                               , ncanlyrp1                   & ! intent(out)
                               , ncanlyrt2                   & ! intent(out)
                               , zztop0                      & ! intent(out)
                               , zztop08                     & ! intent(out)
                               , zztop0i                     & ! intent(out)
                               , zztop0i8                    & ! intent(out)
                               , ehgt                        & ! intent(out)
                               , ehgt8                       & ! intent(out)
                               , ehgti                       & ! intent(out)
                               , ehgti8                      & ! intent(out)
                               , dzcan                       & ! intent(out)
                               , dzcan8                      & ! intent(out)
                               , zztop                       & ! intent(out)
                               , zzmid                       & ! intent(out)
                               , zzbot                       & ! intent(out)
                               , zztop8                      & ! intent(out)
                               , zzmid8                      & ! intent(out)
                               , zzbot8                      & ! intent(out)
                               , alloc_canopy_layer          ! ! subroutine
   use pft_coms         , only : hgt_min                     & ! intent(in)
                               , hgt_max                     ! ! intent(in)
   use consts_coms      , only : onethird                    & ! intent(in)
                               , onethird8                   ! ! intent(in)
   implicit none
   !----- Local variables. ----------------------------------------------------------------!
   integer    :: ilyr
   !---------------------------------------------------------------------------------------!



   !----- Find the layer thickness and the number of layers needed. -----------------------!
   ncanlyr   = 100
   ncanlyrp1 = ncanlyr + 1
   ncanlyrt2 = ncanlyr * 2
   zztop0    = onethird  * minval(hgt_min)
   zztop08   = onethird8 * dble(minval(hgt_min))
   zztop0i   = 1.   / zztop0
   zztop0i8  = 1.d0 / zztop08
   ehgt      = log(maxval(hgt_max)/zztop0)        / log(real(ncanlyr))
   ehgt8     = log(dble(maxval(hgt_max))/zztop08) / log(dble(ncanlyr))
   ehgti     = 1./ ehgt
   ehgti8    = 1.d0 / ehgt8
   !---------------------------------------------------------------------------------------!



   !----- Allocate the variables. ---------------------------------------------------------!
   call alloc_canopy_layer()
   !---------------------------------------------------------------------------------------!



   !----- Define the layer heights. -------------------------------------------------------!
   do ilyr =1,ncanlyr
      zztop (ilyr) = zztop0 * real(ilyr  ) ** ehgt
      zzbot (ilyr) = zztop0 * real(ilyr-1) ** ehgt
      dzcan (ilyr) = zztop(ilyr) - zzbot(ilyr)
      zzmid (ilyr) = 0.5 * (zzbot(ilyr) + zztop(ilyr))
      zztop8(ilyr) = zztop0  * dble(ilyr  ) ** ehgt8
      zzbot8(ilyr) = zztop08 * dble(ilyr-1) ** ehgt8
      dzcan8(ilyr) = zztop8(ilyr) - zzbot8(ilyr)
      zzmid8(ilyr) = 5.d-1 * (zzbot8(ilyr) + zztop8(ilyr))
   end do
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_can_lyr_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_pft_photo_params()

   use ed_max_dims    , only : n_pft                   ! ! intent(in)
   use pft_coms       , only : D0                      & ! intent(out)
                             , Vm_low_temp             & ! intent(out)
                             , Vm_high_temp            & ! intent(out)
                             , Vm_decay_e              & ! intent(out)
                             , Vm_decay_a              & ! intent(out)
                             , Vm_decay_b              & ! intent(out)
                             , Vm0                     & ! intent(out)
                             , Vm_hor                  & ! intent(out)
                             , Vm_q10                  & ! intent(out)
                             , Rd_low_temp             & ! intent(out)
                             , Rd_high_temp            & ! intent(out)
                             , Rd_decay_e              & ! intent(out)
                             , Rd0                     & ! intent(out)
                             , Rd_hor                  & ! intent(out)
                             , Rd_q10                  & ! intent(out)
                             , stomatal_slope          & ! intent(out)
                             , leaf_width              & ! intent(out)
                             , cuticular_cond          & ! intent(out)
                             , quantum_efficiency      & ! intent(out)
                             , photosyn_pathway        & ! intent(out)
                             , dark_respiration_factor & ! intent(out)
                             , water_conductance       &
                             , SLA,rho,TLP	       &
		             , leaf_psi50,psi50	       &
		             , Cap_stem,Cap_leaf       &
                             , lambda0,beta0           &
	                     , SRA              	! ! intent(out)
   use consts_coms    , only : t00                     & ! intent(in)
                             , twothirds               & ! intent(in)
                             , umol_2_mol              & ! intent(in)
                             , yr_sec                  ! ! intent(in)
   use physiology_coms , only: iphysiol                & ! intent(in)
                             , vmfact                  & ! intent(in)
                             , mfact                   & ! intent(in)
                             , gamfact                 & ! intent(in)
                             , d0fact                  & ! intent(in)
                             , alphafact               & ! intent(in)
                             , kfact                   & ! intent(in)
                             , lwfact                  & ! intent(in)
                             , thioff                  &! ! intent(in)
							 , h2o_plant_lim
   implicit none

   integer					 :: ipft
   !---------------------------------------------------------------------------------------!

   D0(1)                     = 0.020  ! 0.010 * d0fact
   D0(2:4)                   = 0.015  ! 0.010 * d0fact
   D0(5)                     = 0.010
   D0(6:8)                   = 0.010
   D0(9:11)                  = 0.010
   D0(12:13)                 = 0.010
   D0(14:15)                 = 0.010
   D0(16)                    = 0.015  ! 0.010 * d0fact
   D0(17)                    = 0.015  ! 0.010 * d0fact
   D0(18:20)                 = 0.015  ! 0.010 * d0fact
   D0(21)                    = 0.010
   D0(22)                    = 0.010
   D0(23:35)                 = 0.015

   Vm_low_temp(1)            =  13.0            ! c4 grass
   Vm_low_temp(2)            =  8.0             ! early tropical
   Vm_low_temp(3)            =  8.0             ! mid tropical
   Vm_low_temp(4)            =  8.0             ! late tropical
   Vm_low_temp(5)            =  4.7137          ! c3 grass
   Vm_low_temp(6)            =  4.7137          ! northern pines ! 5.0
   Vm_low_temp(7)            =  4.7137          ! southern pines ! 5.0
   Vm_low_temp(8)            =  4.7137          ! late conifers  ! 5.0
   Vm_low_temp(9)            =  4.7137          ! early hardwoods
   Vm_low_temp(10)           =  4.7137          ! mid hardwoods
   Vm_low_temp(11)           =  4.7137          ! late hardwoods
   Vm_low_temp(12)           =  4.7137          ! c3 pasture
   Vm_low_temp(13)           =  4.7137          ! c3 crop
   Vm_low_temp(14)           =  13.0            ! c4 pasture
   Vm_low_temp(15)           =  13.0            ! c4 crop
   Vm_low_temp(16)           =  4.7137          ! subtropical C3 grass
   Vm_low_temp(17)           =  4.7137          ! Araucaria
   Vm_low_temp(18:20)        =  8.0             ! late tropical
   Vm_low_temp(21)           =  4.7137
   Vm_low_temp(22)           =  0.5
   Vm_low_temp(23:35)        =  8.0

   Vm_high_temp(1)           =  45.0  + thioff ! C4
   Vm_high_temp(2)           =  45.0  + thioff ! C3
   Vm_high_temp(3)           =  45.0  + thioff ! C3
   Vm_high_temp(4)           =  45.0  + thioff ! C3
   Vm_high_temp(5)           =  45.0  + thioff ! C3
   Vm_high_temp(6)           =  45.0  + thioff ! C3
   Vm_high_temp(7)           =  45.0  + thioff ! C3
   Vm_high_temp(8)           =  45.0  + thioff ! C3
   Vm_high_temp(9)           =  45.0  + thioff ! C3
   Vm_high_temp(10)          =  45.0  + thioff ! C3
   Vm_high_temp(11)          =  45.0  + thioff ! C3
   Vm_high_temp(12)          =  45.0  + thioff ! C3
   Vm_high_temp(13)          =  45.0  + thioff ! C3
   Vm_high_temp(14)          =  45.0  + thioff ! C4
   Vm_high_temp(15)          =  45.0  + thioff ! C4
   Vm_high_temp(16)          =  45.0  + thioff ! C3
   Vm_high_temp(17)          =  45.0  + thioff ! C3
   Vm_high_temp(18:20)       =  45.0  + thioff ! C3
   Vm_high_temp(21:35)       =  45.0  + thioff

   !---------------------------------------------------------------------------------------!
   !    Vm_decay_e is the correction term for high and low temperatures when running the   !
   ! original ED-2.1 correction as in Moorcroft et al. (2001).                             !
   !    Vm_decay_a and Vm_decay_b are the correction terms when running the Collatz et al. !
   ! (1991).  When running Collatz, this is used for both C3 and C4 photosynthesis.        !
   !---------------------------------------------------------------------------------------!
   Vm_decay_e(1:35)          = 0.4     !                                          [    ---]
   Vm_decay_a(1:35)          = 220000. !                                          [  J/mol]
   Vm_decay_b(1:35)          = 690.    !                                          [J/mol/K]
   !---------------------------------------------------------------------------------------!



   !------ Vm0 is the maximum photosynthesis capacity in mol/m2/s. -----------------------!
   Vm0(1)                    = 12.5 !13.75  ! 12.500 * vmfact
   Vm0(2)                    = 18.750  ! 18.750 * vmfact
   Vm0(3)                    = 12.500 ! 12.500 * vmfact
   Vm0(4)                    =  6.250 !  6.250 * vmfact
   Vm0(5)                    = 18.300 
   Vm0(6)                    = 11.350 ! 15.625 * 0.7264
   Vm0(7)                    = 11.350 ! 15.625 * 0.7264
   Vm0(8)                    =  4.540 !  6.250 * 0.7264
   Vm0(9)                    = 20.387 ! 18.250 * 1.1171
   Vm0(10)                   = 17.455 ! 15.625 * 1.1171
   Vm0(11)                   =  6.981 !  6.250 * 1.1171
   Vm0(12:13)                = 18.300 ! 18.300
   Vm0(14:15)                = 15.000 ! 12.500 * vmfact
   Vm0(16)                   = 24.0625 ! 21.875 * vmfact
   Vm0(17)                   = 18.750 ! 15.625 * vmfact
   Vm0(18)                   = 18.750 ! 18.750 * vmfact
   Vm0(19)                   = 12.500  ! 12.500 * vmfact
   Vm0(20)                   =  6.250 !  6.250 * vmfact
   Vm0(21)                   = 21.5!19.0
   Vm0(22)                   = 7.03!10.861
   Vm0(23:35)                = 27 *1.2 !18.750  ! 12.500 * vmfact JL

   ! Farquhar's temperature dependence
   Vm0(29) = 55 * 0.44 !50 * 0.44 !!50 * 0.44 !9.0!25.!22.!27. !6.0 * 2.5 * 1.4
   Vm0(27) = 45 * 0.44 !42 * 0.44 !6.5!14.!14.!17.5 !3.5 * 2.5 * 1.4
   Vm0(23) = 35 * 0.44 !38 * 0.44 !7.5!16.!16.!20. !3.8 * 2.5 * 1.4
   Vm0(26) = 35 * 0.44 !38 * 0.44 !7.5!16.!16.!20. !3.8 * 2.5 * 1.4
   Vm0(24) = 25 * 0.44 !5.5!12.!12.!16. ! !2.8 * 2.5 * 1.4
   Vm0(30:35) = 40.5 * 0.44
   Vm0(1) = 55 * 0.44

   if(vmfact == 0.0)then
      Vm0(30:32) = 40.5 * 0.44
      Vm0(33:35) = 40.5 * 0.44!42.4 * 0.44
   elseif(vmfact == 1.0)then
      Vm0(30:32) = 27.8 * 0.44
      Vm0(33:35) = 39.2 * 0.44
   elseif(vmfact == 2.0)then
      Vm0(30:32) = 27.8 * 0.44
      Vm0(33:35) = 45.7 * 0.44
   elseif(vmfact == 3.0)then
      Vm0(30:32) = 54.6 * 0.44
      Vm0(33:35) = 45.7 * 0.44
   elseif(vmfact == 4.0)then
      Vm0(30:32) = 54.6 * 0.44
      Vm0(33:35) = 39.2 * 0.44
   endif
   print*,'Vm0(30:32)',Vm0(30:32),'Vm0(33:35)',Vm0(33:35)

   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      Vm_hor is the Arrhenius "activation energy" divided by the universal gas         !
   ! constant.  Vm_q10 is the base for the Collatz approach.                               !
   !---------------------------------------------------------------------------------------!
   vm_hor(1:35)              = 3000.
   !----- Here we distinguish between C3 and C4 photosynthesis as in Collatz et al 91/92. -!
   vm_q10(1)                 = 2.0
   vm_q10(2:13)              = 2.4
   vm_q10(14:15)             = 2.0
   vm_q10(16:20)             = 2.4
   vm_q10(21:35)             = 2.4
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !    Dark_respiration_factor is the lower-case gamma in Moorcroft et al. (2001).        !
   !---------------------------------------------------------------------------------------!
   dark_respiration_factor(1)     = 0.04 !0.036
   dark_respiration_factor(2)     = 0.02 !0.02 !0.015 ! 0.020 * gamfact
   dark_respiration_factor(3)     = 0.02 !0.015 ! 0.020 * gamfact
   dark_respiration_factor(4)     = 0.02 !0.015 ! 0.020 * gamfact
   dark_respiration_factor(5)     = 0.020
   dark_respiration_factor(6)     = 0.020
   dark_respiration_factor(7)     = 0.020
   dark_respiration_factor(8)     = 0.020
   dark_respiration_factor(9)     = 0.020
   dark_respiration_factor(10)    = 0.020
   dark_respiration_factor(11)    = 0.020
   dark_respiration_factor(12)    = 0.020
   dark_respiration_factor(13)    = 0.020
   dark_respiration_factor(14)    = 0.035
   dark_respiration_factor(15)    = 0.035
   dark_respiration_factor(16)    = 0.015
   dark_respiration_factor(17)    = 0.020
   dark_respiration_factor(18:20) = 0.02 !0.02 !0.015 ! 0.020 * gamfact
   dark_respiration_factor(21)    = 0.020
   dark_respiration_factor(22)    = 0.040
   dark_respiration_factor(23:35) = 0.015!0.025 / 5.
   dark_respiration_factor(29)    = 0.03
   dark_respiration_factor(30:32) = 0.015
   dark_respiration_factor(33:35) = 0.015

   !---------------------------------------------------------------------------------------!


   !----- Currently we assume most parameters to be the same as the Vm ones. --------------!
   Rd_low_temp (1:35) = Vm_low_temp (1:35)
   Rd_high_temp(1:35) = Vm_high_temp(1:35)
   Rd_decay_e  (1:35) = Vm_decay_e  (1:35)
   Rd_hor      (1:35) = Vm_hor      (1:35)
   Rd_q10      (1:35) = Vm_q10      (1:35)
 
   !---------------------------------------------------------------------------------------!
   !    Respiration terms.  Here we must check whether this will run Foley-based or        !
   ! Collatz-based photosynthesis, because the respiration/Vm ratio is constant in the     !
   ! former but not in the latter.                                                         !
   !---------------------------------------------------------------------------------------!
   select case (iphysiol)
   case (0,1)
      !------ This should be simply gamma times Vm0. --------------------------------------!
      Rd0         (1:35) = dark_respiration_factor(1:35) * Vm0(1:35)

   case (2,3)
      !------------------------------------------------------------------------------------!
      !     Collatz-based method.  They have the gamma at 25 C, but because the Q10 bases  !
      ! are different for Vm and respiration (at least for C3), we must convert it to the  !
      ! ratio at 15C.                                                                      !
      !------------------------------------------------------------------------------------!
      Rd0         (1:35) = dark_respiration_factor(1:35) * Vm0(1:35)                       &
                         * Vm_q10(1:35) / Rd_q10(1:35)
   end select
   !---------------------------------------------------------------------------------------!
   ! Mannuly define Rd0 for TDF,   XXT
   ! micro mol /m2/s at 25degC
!   Rd0(29) = 2.25!2.08!1.17 !0.7353 !0.9649 !1.4285 / 1.54
!   Rd0(27) = 1.25!!1.33!!0.75 !0.6007 !0.7694 / 1.54
!   Rd0(23) = 0.90!0.91!!0.5120 !0.80 / 1.54 !0.8874 / 1.54
!   Rd0(26) = 0.90!!0.5120 !0.80 / 1.54 !0.8874 / 1.54
!   Rd0(24) = 0.70!!0.4029 !0.6401 / 1.54
   
!   Rd0(23:29) = (1.4285+0.6401) / 2. / 1.54
   ! 1.53 is to convert 30deg to 25deg


   !----- Define the stomatal slope (aka the M factor). -----------------------------------!
   stomatal_slope(1)         =  4.0 !5.0
   stomatal_slope(2)         =  10.0! 9.0 ! 8.0    * mfact
   stomatal_slope(3)         =  10.0 ! 9.0 ! 8.0    * mfact
   stomatal_slope(4)         =  10.0 !9.0 ! 8.0    * mfact
   stomatal_slope(5)         =  8.0
   stomatal_slope(6)         =  20.0 !6.3949
   stomatal_slope(7)         =  6.3949
   stomatal_slope(8)         =  6.3949
   stomatal_slope(9)         =  6.3949
   stomatal_slope(10)        =  6.3949
   stomatal_slope(11)        =  6.3949
   stomatal_slope(12)        =  8.0   
   stomatal_slope(13)        =  8.0
   stomatal_slope(14)        =  5.0
   stomatal_slope(15)        =  5.0
   stomatal_slope(16)        =  9.0
   stomatal_slope(17)        =  6.4
   stomatal_slope(18:20)     =  10.0! 9.0 ! 8.0    * mfact
   stomatal_slope(21:29)     =  6.3849
   stomatal_slope(23:35)     =  15. ! 9.0 ! 8.0    * mfact
 
   ! Mazoni et al. 2011
   lambda0(1:35)     		 =  12.


 ! Mazoni et al. 2011
   lambda0(1:35)     		 =  12. ! 9.0 ! 8.0    * mfact
   lambda0(21)               =  2.8 !3.5
   lambda0(22)               =  6.64!46. Manzoni et al 2011 'Optimizing stomotal conductance' divide by 100
   lambda0(23)               =  12.
   lambda0(24)               =  12.
   lambda0(26)               =  9.
   lambda0(27)               =  12.
   lambda0(29)               =  12.
   lambda0(30:35)            =  9.! 9.

   beta0(1:35)				 = -0.13 / 102.
   beta0(24)				 = -0.10 / 102.!-0.10 / 102.
   beta0(26)				 = -0.40 / 102.!-0.40 / 102.
   beta0(27)				 = -0.70 / 102.!-0.70 / 102.
   beta0(29)				 = -1.0 / 102.!-1.00 / 102.
   beta0(30:35)				 = -0.40 / 102.!-0.40 / 102.

   cuticular_cond(1)         =  8000.0
   cuticular_cond(2)         = 10000.0
   cuticular_cond(3)         = 10000.0
   cuticular_cond(4)         = 10000.0
   cuticular_cond(5)         = 10000.0
   cuticular_cond(6)         =  1000.0
   cuticular_cond(7)         =  1000.0
   cuticular_cond(8)         =  1000.0
   cuticular_cond(9)         = 20000.0
   cuticular_cond(10)        = 20000.0
   cuticular_cond(11)        = 20000.0
   cuticular_cond(12)        = 10000.0
   cuticular_cond(13)        = 10000.0
   cuticular_cond(14)        =  8000.0
   cuticular_cond(15)        =  8000.0
   cuticular_cond(16)        = 10000.0
   cuticular_cond(17)        =  1000.0
   cuticular_cond(18:20)     = 10000.0
   cuticular_cond(21)        = 10000.0
   cuticular_cond(22)        = 1000.0
   cuticular_cond(23:35)     = 10000.0

!slope of photosynthesis increasing with light
   quantum_efficiency(1)     = 0.053
   quantum_efficiency(2)     = 0.070 ! 0.08  * alphafact
   quantum_efficiency(3)     = 0.070 ! 0.08  * alphafact
   quantum_efficiency(4)     = 0.070 ! 0.08  * alphafact
   quantum_efficiency(5)     = 0.080
   quantum_efficiency(6)     = 0.080
   quantum_efficiency(7)     = 0.080
   quantum_efficiency(8)     = 0.080
   quantum_efficiency(9)     = 0.080
   quantum_efficiency(10)    = 0.080
   quantum_efficiency(11)    = 0.080
   quantum_efficiency(12)    = 0.080
   quantum_efficiency(13)    = 0.080
   quantum_efficiency(14)    = 0.053
   quantum_efficiency(15)    = 0.053
   quantum_efficiency(16)    = 0.070 ! 0.08  * alphafact
   quantum_efficiency(17)    = 0.070 ! 0.08  * alphafact
   quantum_efficiency(18:20) = 0.070 ! 0.08  * alphafact
   quantum_efficiency(21:29) = 0.080
   quantum_efficiency(30:35) = 0.080 ! 0.08  * alphafact

   !---------------------------------------------------------------------------------------!
   !     The KW parameter. Medvigy et al. (2009) and Moorcroft et al. (2001) give the      !
   ! number in m/yr/kg_C_root.  Here we must define it in m/s/kg_C_root.                 !
   !---------------------------------------------------------------------------------------!
!   water_conductance(1:29)  = 250./ yr_sec!450. / yr_sec ! 150. / yr_sec * kfact
 !  water_conductance(30:35) = 450. / yr_sec ! 150. / yr_sec * kfact
   !---------------------------------------------------------------------------------------!

   do ipft = 1,n_pft
        psi50(ipft)  =  (-3.0028 * rho(ipft) - 0.599) * 102. !min(-100.,max(-400.,-500. * (rho(ipft) - 0.35) - 100.)) 
        water_conductance(ipft) = exp(-2.455 * rho(ipft) + 2.348 + 0.5 * 0.6186) !max(1.0,min(3.5,-3.8 * rho(ipft) + 4.6)) ! kg H2O /m /s/Mpa
        TLP(ipft)  = 102. * (-4.59 + 0.62 * log(SLA(ipft)) - 1.15 * log(rho(ipft)))
        leaf_psi50(ipft) = (TLP(ipft) / 102. * 0.7 - 0.4) * 102.  ! from Brodribb et al. 2002
   enddo

   if (h2o_plant_lim < 3) then ! original ED2 water limitaiton scheme
	   water_conductance(1:35) = 450. / yr_sec * kfact
   endif
!   psi50(24) = -3.16 * 102.
!   psi50(23) = -2.76 * 102.
!   psi50(26) = -2.76 * 102.
!   psi50(27) = -2.81 * 102.
!   psi50(29) = -1.74 * 102.
!   psi50(23:29) = (-3.16-1.74) / 2. * 102.

 !  water_conductance(24) = 1.76
 !  water_conductance(23) = 2.48
 !  water_conductance(26) = 2.48
 !  water_conductance(27) = 2.39
 !  water_conductance(29) = 5.78
  ! water_conductance(23:29) = (1.76+5.78) / 2.

  ! TLP(24) = -258.
  ! TLP(23) = -241.
 !  TLP(26) = -241.
 !  TLP(27) = -217.
 !  TLP(29) = -174.
   !  TLP(23:29) = (-258.-174.) / 2.
   
 Vm0(6) = 11.35 
 psi50(6)= -2.6*102 !Aspinwall 2011 table 3
 water_conductance(6)= 0.92 !Aspinwall 2011 (may need to modify) 
 TLP(6)= -2.3*102 !Bartlett 2012 
 leaf_psi50(6)= (TLP(6)/102. * 0.7 - 0.4) *102. !same as above
 TLP(1) = -2.*102.

print*,'TLP',TLP(23:30),'P50',psi50(23:30)
   photosyn_pathway(1)       = 4
   photosyn_pathway(2:4)     = 3
   photosyn_pathway(5)       = 3
   photosyn_pathway(6:13)    = 3
   photosyn_pathway(14:15)   = 4
   photosyn_pathway(16:17)   = 3
   photosyn_pathway(18:20)   = 3
   photosyn_pathway(21:35)   = 3

   !----- Leaf width [m].  This controls the boundary layer conductance. ------------------!
   leaf_width( 1)    = 0.20
   leaf_width( 2)    = 0.20
   leaf_width( 3)    = 0.20
   leaf_width( 4)    = 0.20
   leaf_width( 5)    = 0.05
   leaf_width( 6)    = 0.05
   leaf_width( 7)    = 0.05
   leaf_width( 8)    = 0.05
   leaf_width( 9)    = 0.05
   leaf_width(10)    = 0.05
   leaf_width(11)    = 0.05
   leaf_width(12)    = 0.05
   leaf_width(13)    = 0.05
   leaf_width(14)    = 0.20
   leaf_width(15)    = 0.20
   leaf_width(16)    = 0.20
   leaf_width(17)    = 0.05
   leaf_width(18:20) = 0.20
   leaf_width(21:22) = 0.05
   leaf_width(23:35) = 0.20
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_pft_photo_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_decomp_params()
   use consts_coms , only : yr_day                      ! ! intent(in)
   use decomp_coms , only : resp_opt_water              & ! intent(in)
                          , resp_water_below_opt        & ! intent(in)
                          , resp_water_above_opt        & ! intent(in)
                          , resp_temperature_increase   & ! intent(in)
                          , N_immobil_supply_scale      & ! intent(in)
                          , cwd_frac                    & ! intent(in)
                          , r_fsc                       & ! intent(in)
                          , r_stsc                      & ! intent(in)
                          , r_ssc                       & ! intent(in)
                          , K1                          & ! intent(in)
                          , K2                          & ! intent(in)
                          , K3                          ! ! intent(in)
!   resp_opt_water            = 0.8938
!   resp_water_below_opt      = 5.0786
!   resp_water_above_opt      = 4.5139
!   resp_temperature_increase = 0.0757

   resp_opt_water            = 0.95!TEM parameterization !0.95 ED2
   resp_water_below_opt      = 2.2 
   resp_water_above_opt      = 2.2
   resp_temperature_increase = 0.092 !0.069!ATT from Euskirchen et al 2009 !0.092 ED2 !

   N_immobil_supply_scale    = 40.0 / yr_day
   cwd_frac                  = 0.2
   r_fsc                     = 1.0
   r_stsc                    = 1.0!0.5!0.3
   r_ssc                     = 1.0
   K1                        = 4.5   / yr_day 
   K2                        = 11.0  / yr_day 
   K3                        = 100.2 / yr_day

   return

end subroutine init_decomp_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_pft_resp_params()

   use pft_coms       , only : growth_resp_factor        & ! intent(out)
                             , leaf_turnover_rate        & ! intent(out)
                             , root_turnover_rate        & ! intent(out)
                             , storage_turnover_rate     & ! intent(out)
                             , root_respiration_factor   & ! intent(out)
                             , bark_turnover_rate
   use decomp_coms    , only : f_labile                  & ! intent(out)
                             , f_fast
   use consts_coms    , only : onesixth                  & ! intent(in)
                             , onethird                  ! ! intent(in)
   implicit none

   growth_resp_factor(1)          = onethird
   growth_resp_factor(2)          = onethird
   growth_resp_factor(3)          = onethird
   growth_resp_factor(4)          = onethird
   growth_resp_factor(5)          = onethird
   growth_resp_factor(6)          = 0.4503 ! 0.333
   growth_resp_factor(7)          = 0.4503
   growth_resp_factor(8)          = 0.4503 ! 0.333
   growth_resp_factor(9)          = 0.0
   growth_resp_factor(10)         = 0.0
   growth_resp_factor(11)         = 0.0
   growth_resp_factor(12)         = onethird
   growth_resp_factor(13)         = onethird
   growth_resp_factor(14)         = onethird
   growth_resp_factor(15)         = onethird
   growth_resp_factor(16)         = onethird
   growth_resp_factor(17)         = 0.4503
   growth_resp_factor(18:20)      = onethird
   growth_resp_factor(21)         = onethird
   growth_resp_factor(22)         = 0.4503
   growth_resp_factor(23:35)      = 0.45!0.30!0.28!0.30

   leaf_turnover_rate(1)          = 2.0
   leaf_turnover_rate(2)          = 1.0
   leaf_turnover_rate(3)          = 0.5
   leaf_turnover_rate(4)          = onethird
   leaf_turnover_rate(5)          = 2.0
   leaf_turnover_rate(6)          = onethird
   leaf_turnover_rate(7)          = onethird
   leaf_turnover_rate(8)          = onethird
   leaf_turnover_rate(9)          = 0.0
   leaf_turnover_rate(10)         = 0.0
   leaf_turnover_rate(11)         = 0.0
   leaf_turnover_rate(12)         = 2.0
   leaf_turnover_rate(13)         = 2.0
   leaf_turnover_rate(14)         = 2.0
   leaf_turnover_rate(15)         = 2.0
   leaf_turnover_rate(16)         = 2.0
   leaf_turnover_rate(17)         = onesixth
   leaf_turnover_rate(18)          = 1.0
   leaf_turnover_rate(19)          = 0.5
   leaf_turnover_rate(20)          = onethird
   leaf_turnover_rate(21)         = 0.0
   leaf_turnover_rate(22)         = 0.143
   leaf_turnover_rate(23:35)      = 0.5

   !----- Root turnover rate.  ------------------------------------------------------------!
   root_turnover_rate(1)          = 2.0
   root_turnover_rate(2)          = 1.0
   root_turnover_rate(3)          = 0.5
   root_turnover_rate(4)          = onethird
   root_turnover_rate(5)          = 2.0
   root_turnover_rate(6)          = 0.7 !Gill 2000 et al !3.9272 ! 0.333
   root_turnover_rate(7)          = 4.117847 ! 0.333
   root_turnover_rate(8)          = 3.800132 ! 0.333
   root_turnover_rate(9)          = 5.772506
   root_turnover_rate(10)         = 5.083700
   root_turnover_rate(11)         = 5.070992
   root_turnover_rate(12)         = onethird
   root_turnover_rate(13)         = onethird
   root_turnover_rate(14)         = 2.0
   root_turnover_rate(15)         = 2.0
   root_turnover_rate(16)         = 2.0
   root_turnover_rate(17)         = onesixth
   root_turnover_rate(18)         = 1.0
   root_turnover_rate(19)         = 0.5
   root_turnover_rate(20)         = onethird
   root_turnover_rate(21)         = 0.65!pinno2010
   root_turnover_rate(22)         = 1.56!ruess2003
   root_turnover_rate(23:35)      = 0.5 ! Gill et al. 2000

   storage_turnover_rate(1)       = 0.00 ! 0.25
   storage_turnover_rate(2)       = 0.00 ! 0.25
   storage_turnover_rate(3)       = 0.00 ! 0.25
   storage_turnover_rate(4)       = 0.00 ! 0.25
   storage_turnover_rate(5)       = 0.00 ! 0.25
   storage_turnover_rate(6)       = 0.00 ! 0.25
   storage_turnover_rate(7)       = 0.00 ! 0.25
   storage_turnover_rate(8)       = 0.00 ! 0.25
   storage_turnover_rate(9)       = 0.6243
   storage_turnover_rate(10)      = 0.6243
   storage_turnover_rate(11)      = 0.6243
   storage_turnover_rate(12)      = 0.00 ! 0.25
   storage_turnover_rate(13)      = 0.00 ! 0.25
   storage_turnover_rate(14)      = 0.00 ! 0.25
   storage_turnover_rate(15)      = 0.00 ! 0.25
   storage_turnover_rate(16)      = 0.00 ! 0.25
   storage_turnover_rate(17)      = 0.00 ! 0.25
   storage_turnover_rate(18:20)   = 0.00 ! 0.25
   storage_turnover_rate(21)      = 0.0 !0.6! 0.6243
   storage_turnover_rate(22)      = 0.00
   storage_turnover_rate(23:35)   = 0.0

   bark_turnover_rate(1:29) = 0.
   bark_turnover_rate(30:35) = 0.0

   root_respiration_factor(1:35)  = 0.528 

   f_labile(1:5)                  = 1.0
   f_labile(6:11)                 = 0.79
   f_labile(12:15)                = 1.0
   f_labile(16)                   = 1.0
   f_labile(17)                   = 0.79
   f_labile(18:20)                = 1.0
   f_labile(21)                   = 1.0!0.79
   f_labile(22)                   = 1.0!0.79
   f_labile(23:35)                = 1.0

   f_fast(1:21)                   = 1.0
   f_fast(22)                     = 0.0
   f_fast(23:35)                  = 1.0

   return
end subroutine init_pft_resp_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!     This subroutine assigns some PFT-dependent parameters that control mortality rates.  !
!------------------------------------------------------------------------------------------!
subroutine init_pft_mort_params()

   use pft_coms    , only : mort1                      & ! intent(out)
                          , mort2                      & ! intent(out)
                          , mort3                      & ! intent(out)
                          , rho                        & ! intent(out)
                          , seedling_mortality         & ! intent(out)
                          , treefall_s_gtht            & ! intent(out)
                          , treefall_s_ltht            & ! intent(out)
                          , plant_min_temp             & ! intent(out)
                          , frost_mort                 & ! intent(out)
                          , resprout, resprout_fraction
   use consts_coms , only : t00                        & ! intent(in)
                          , lnexp_max                  ! ! intent(in)
   use disturb_coms, only : treefall_disturbance_rate  & ! intent(inout)
                          , time2canopy                ! ! intent(in)

   implicit none

   !----- Local variables. ----------------------------------------------------------------!
   real     :: aquad
   real     :: bquad
   real     :: cquad
   real     :: discr
   real     :: lambda_ref
   real     :: lambda_eff
   real     :: leff_neg
   real     :: leff_pos
   !---------------------------------------------------------------------------------------!


   frost_mort(1)     = 3.0
   frost_mort(2:4)   = 3.0
   frost_mort(5)     = 3.0
   frost_mort(6:11)  = 3.0
   frost_mort(12:13) = 3.0
   frost_mort(14:15) = 3.0
   frost_mort(16:17) = 3.0
   frost_mort(18:20) = 3.0
   frost_mort(21:35) = 3.0


   mort1(1)  = 10.0
!   mort1(2)  = 10.0
!   mort1(3)  = 10.0
!   mort1(4)  = 10.0

   mort1(2)  = 0.!0.03
   mort1(3)  = 0.!0.03
   mort1(4)  = 0.!0.03

   mort1(5)  = 1.0
   mort1(6)  = 1.0
   mort1(7)  = 1.0
   mort1(8)  = 1.0
   mort1(9)  = 1.0
   mort1(10) = 1.0
   mort1(11) = 1.0
   mort1(12) = 1.0
   mort1(13) = 1.0
   mort1(14) = 10.0
   mort1(15) = 10.0
   mort1(16) = 10.0
   mort1(17) = 10.0
   mort1(18:20)  = 10.0
   mort1(21) = 0.1
   mort1(22:22) = 0.1
   mort1(23:35) = 0.03 !0.1

   mort2(1)  = 20.0
   mort2(2)  = 20.0
   mort2(3)  = 20.0
   mort2(4)  = 20.0
   mort2(5)  = 20.0
   mort2(6)  = 20.0
   mort2(7)  = 20.0
   mort2(8)  = 20.0
   mort2(9)  = 20.0
   mort2(10) = 20.0
   mort2(11) = 20.0
   mort2(12) = 20.0
   mort2(13) = 20.0
   mort2(14) = 20.0
   mort2(15) = 20.0
   mort2(16) = 20.0
   mort2(17) = 20.0
   mort2(18:20) = 20.0
   mort2(21:35) = 20.0

   mort3(1)  = 0.15 * (1. - rho(1) / rho(4)) 
   mort3(2)  = 0.!0.15 * (1. - rho(2)) *0.9 !JL multiplied by 0.9
   mort3(3)  = 0.!0.15 * (1. - rho(3)) *0.9 !JL multiplied by 0.9
   mort3(4)  = 0.!0.15 * (1. - rho(4)) *0.9 !JL multiplied by 0.9
!   mort3(2)  = 0.15 * (1. - rho(2) / rho(4))  
!   mort3(3)  = 0.15 * (1. - rho(3) / rho(4))
!   mort3(4)  = 0.0
   mort3(5)  = 0.066
   mort3(6)  = 0.0033928
   mort3(7)  = 0.0043
   mort3(8)  = 0.0023568
   mort3(9)  = 0.006144
   mort3(10) = 0.003808
   mort3(11) = 0.00428
   mort3(12) = 0.066
   mort3(13) = 0.066
   mort3(14) = 0.15 * (1. - rho(14) / rho(4))
   mort3(15) = 0.15 * (1. - rho(15) / rho(4))
   mort3(16) = 0.15 * (1. - rho(16) / rho(4))
   mort3(17) = 0.0043 ! Same as pines
   mort3(18)  = 0.15 * (1. - rho(18) / rho(20))  
   mort3(19)  = 0.15 * (1. - rho(19) / rho(20))
   mort3(20)  = 0.0
   mort3(21)  = 0.059!0.0067!0.059 
   mort3(22)  = 0.014! 0.005!0.014
   mort3(23:35) = (-5 * (rho(23:35) - 0.25) + 3.75)  / 100.
   mort3(30:32) = 0.029
   mort3(33:35) = 0.029
   !mort3(30:35) = 0.1



   !---------------------------------------------------------------------------------------!
   !     Here we check whether we need to re-calculate the treefall disturbance rate so it !
   ! is consistent with the time to reach the canopy.                                      !
   !---------------------------------------------------------------------------------------!
   if (treefall_disturbance_rate == 0.) then
      !------ No disturbance rate, set time to reach canopy to infinity. ------------------!
      time2canopy = huge(1.) 
      lambda_ref  = 0.
      lambda_eff  = 0.

   else
      lambda_ref = abs(treefall_disturbance_rate)

      if (time2canopy > 0.) then
         !---------------------------------------------------------------------------------!
         !     We are not going to knock down trees as soon as the patch is created;       !
         ! instead, we will wait until the patch age is older than time2canopy.  We want,  !
         ! however, to make the mean patch age to be 1/treefall_disturbance_rate.  The     !
         ! equation below can be retrieved by integrating the steady-state probability     !
         ! distribution function.  The equation is quadratic and the discriminant will     !
         ! never be zero and the treefall_disturbance_rate will be always positive because !
         ! the values of time2canopy and treefall_disturbance_rate have already been       !
         ! tested in ed_opspec.F90.                                                        !
         !---------------------------------------------------------------------------------!
         aquad    = time2canopy * time2canopy * lambda_ref  - 2. * time2canopy
         bquad    = 2. * time2canopy * lambda_ref - 2.
         cquad    = 2. * lambda_ref
         !------ Find the discriminant. ---------------------------------------------------!
         discr    = bquad * bquad - 4. * aquad * cquad
         leff_neg = - 0.5 * (bquad - sqrt(discr)) / aquad
         leff_pos = - 0.5 * (bquad + sqrt(discr)) / aquad
         !---------------------------------------------------------------------------------!
         !      Use the maximum value, but don't let the value to be too large otherwise   !
         ! the negative exponential will cause underflow.                                  !
         !---------------------------------------------------------------------------------!
         lambda_eff = min(lnexp_max,max(leff_neg,leff_pos))
      else
         lambda_eff = lambda_ref
      end if

   end if
   !---------------------------------------------------------------------------------------!




   !---------------------------------------------------------------------------------------!
   !    Print out the summary.                                                             !
   !---------------------------------------------------------------------------------------!
   write (unit=*,fmt='(a)')           '----------------------------------------'
   write (unit=*,fmt='(a)')           '  Treefall disturbance parameters:'
   write (unit=*,fmt='(a,1x,es12.5)') '  - LAMBDA_REF  =',lambda_ref
   write (unit=*,fmt='(a,1x,es12.5)') '  - LAMBDA_EFF  =',lambda_eff
   write (unit=*,fmt='(a,1x,es12.5)') '  - TIME2CANOPY =',time2canopy
   write (unit=*,fmt='(a)')           '----------------------------------------'
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     Here we check whether patches should be created or the treefall should affect     !
   ! only the mortality (quasi- size-structured approximation; other disturbances may be   !
   ! turned off for a true size-structured approximation).                                 !
   !---------------------------------------------------------------------------------------!
   if (treefall_disturbance_rate < 0.) then
      !------------------------------------------------------------------------------------!
      !      We incorporate the disturbance rate into the density-independent mortality    !
      ! rate and turn off the patch-creating treefall disturbance.                         !
      !------------------------------------------------------------------------------------!
      mort3(:) = mort3(:) + lambda_eff
      treefall_disturbance_rate = 0.
   else
      treefall_disturbance_rate = lambda_eff
   end if
   !---------------------------------------------------------------------------------------!


   seedling_mortality(1)    = 0.97
   seedling_mortality(2:4)  = 1.00!0.95
!   seedling_mortality(2)  = 0.5
!   seedling_mortality(3)  = 0.95
!   seedling_mortality(4)  = 0.97
   seedling_mortality(5)    = 0.95
   seedling_mortality(6:15) = 0.95
   seedling_mortality(16)   = 0.95
   seedling_mortality(17)   = 0.95
   seedling_mortality(18:20)  = 0.95
   seedling_mortality(21)   = 0.95
   seedling_mortality(22)   = 0.95
   seedling_mortality(23:35) = 0.95

   resprout(1:29) = .false.
   resprout(30:35) = .true.

   resprout_fraction(1:29) = 0.
   resprout_fraction(30:35) = 0.5

   !seedling_mortality=0.critical height survivorship
   treefall_s_gtht(1:35)    = 0.0

   treefall_s_ltht(1)       = 0.25
   treefall_s_ltht(2:4)     = 0.10
   treefall_s_ltht(5)       = 0.25
   treefall_s_ltht(6:11)    = 0.10!0.10
   treefall_s_ltht(12:15)   = 0.25
   treefall_s_ltht(16)      = 0.25
   treefall_s_ltht(17)      = 0.10
   treefall_s_ltht(18:20)   = 0.10
   treefall_s_ltht(21)      = 0.10!2.0 !0.10
   treefall_s_ltht(22)      = 0.10!2.0!0.10
   treefall_s_ltht(23:35)   = 0.80



! OVERRIDE
!!!!!treefall_s_ltht(:) = 0.0


   plant_min_temp(1:4)      = t00+2.5
   plant_min_temp(5:6)      = t00-80.0
   plant_min_temp(7)        = t00-10.0
   plant_min_temp(8)        = t00-60.0
   plant_min_temp(9)        = t00-80.0
   plant_min_temp(10:11)    = t00-20.0
   plant_min_temp(12:13)    = t00-80.0
   plant_min_temp(14:15)    = t00+2.5
   plant_min_temp(16)       = t00-10.0
   plant_min_temp(17)       = t00-10.0
   plant_min_temp(18:20)    = t00+2.5
   plant_min_temp(21)       = t00-60.0
   plant_min_temp(22)       = t00-80.0
   plant_min_temp(23:35)    = t00+2.5

   return
end subroutine init_pft_mort_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_pft_alloc_params()

   use pft_coms    , only : leaf_turnover_rate    & ! intent(in)
                          , is_tropical           & ! intent(out)
                          , is_grass              & ! intent(out)
                          , rho                   & ! intent(out)
                          , SLA                   & ! intent(out)
                          , horiz_branch          & ! intent(out)
                          , q                     & ! intent(out)
                          , qsw                   & ! intent(out)
                          , init_density          & ! intent(out)
                          , agf_bs                & ! intent(out)
                          , agf_bsi               & ! intent(out)
                          , hgt_min               & ! intent(out)
                          , hgt_ref               & ! intent(out)
                          , hgt_max               & ! intent(out)
                          , min_dbh               & ! intent(out)
                          , max_dbh               & ! intent(out)
                          , b1Ht                  & ! intent(out)
                          , b2Ht                  & ! intent(out)
                          , b1Bs_small            & ! intent(out)
                          , b2Bs_small            & ! intent(out)
                          , b1Bs_big              & ! intent(out)
                          , b2Bs_big              & ! intent(out)
                          , b1Bsap				  &
                          , b2Bsap				  &
                          , b1Ca                  & ! intent(out)
                          , b2Ca                  & ! intent(out)
                          , b1Rd                  & ! intent(out)
                          , b2Rd                  & ! intent(out)
                          , root_beta			  &
                          , b1Vol                 & ! intent(out)
                          , b2Vol                 & ! intent(out)
                          , bdead_crit            & ! intent(out)
                          , b1Bl                  & ! intent(out)
                          , b2Bl                  & ! intent(out)
                          , C2B                   & ! intent(out)
                          , sapwood_ratio         & ! intent(out)
                          , rbranch               & ! intent(out)
                          , rdiamet               & ! intent(out)
                          , rlength               & ! intent(out)
                          , diammin               & ! intent(out)
                          , ntrunk                & ! intent(out)
                          , conijn_a              & ! intent(out)
                          , conijn_b              & ! intent(out)
                          , conijn_c              & ! intent(out)
                          , conijn_d              &
                          , Cap_stem,Cap_leaf	  &
                          , SRA			  &
                          , root_dens,bark_thickness_beta! ! intent(out)
   use allometry   , only : h2dbh                 & ! function
                          , dbh2h				  &
                          , dbh2bd                ! ! function
   use consts_coms , only : twothirds             & ! intent(in)
                          , pi1                   ! ! intent(in)
   use ed_max_dims , only : n_pft                 & ! intent(in)
                          , str_len               ! ! intent(in)
   use ed_misc_coms, only : iallom                ! ! intent(in)
   use disturb_coms, only : grass_biomass_threshold, next_fire_year
   implicit none
   !----- Local variables. ----------------------------------------------------------------!
   integer                           :: ipft
   integer                           :: n
   real                              :: aux
   !----- Constants shared by both bdead and bleaf (tropical PFTs) ------------------------!
   real                  , parameter :: a1          =  -1.981
   real                  , parameter :: b1          =   1.047
   real                  , parameter :: dcrit       = 100.0
   !----- Constants used by bdead only (tropical PFTs) ------------------------------------!
   real                  , parameter :: c1d         =   0.572
   real                  , parameter :: d1d         =   0.931
   real                  , parameter :: a2d         =  -1.086
   real                  , parameter :: b2d         =   0.876
   real                  , parameter :: c2d         =   0.604
   real                  , parameter :: d2d         =   0.871
   !----- Constants used by bleaf only (tropical PFTs) ------------------------------------!
   real                  , parameter :: c1l         =  -0.584
   real                  , parameter :: d1l         =   0.550
   real                  , parameter :: a2l         =  -4.111
   real                  , parameter :: b2l         =   0.605
   real                  , parameter :: c2l         =   0.848
   real                  , parameter :: d2l         =   0.438
   !---------------------------------------------------------------------------------------!
   !     MLO.   These are the new parameters obtained by adjusting a curve that is similar !
   !            to the modified Chave's equation to include wood density effect on the     !
   !            DBH->AGB allometry as described by:                                        ! 
   !                                                                                       !
   !            Baker, T. R., and co-authors, 2004: Variation in wood density determines   !
   !               spatial patterns in Amazonian forest biomass.  Glob. Change Biol., 10,  !
   !               545-562.                                                                !
   !                                                                                       !
   !     The "a" parameters were obtaining by splitting balive and bdead at the same ratio !
   ! as the original ED-2.1 allometry, and optimising a function of the form               !
   ! B? = (rho / a3) * exp [a1 + a2 * ln(DBH)]                                             !
   !     The "z" parameters were obtaining by using the original balive and computing      !
   ! bdead as the difference between the total biomass and the original balive.            !
   !---------------------------------------------------------------------------------------!
   real, dimension(3)    , parameter :: aleaf       = (/ -1.259299,  1.679213,  4.985562 /)
   real, dimension(3)    , parameter :: adead_small = (/ -1.494639,  2.453309,  1.597272 /)
   real, dimension(3)    , parameter :: adead_big   = (/  2.105856,  2.423031, 50.198984 /)
   real, dimension(3)    , parameter :: zdead_small = (/ -1.113827,  2.440483,  2.180632 /)
   real, dimension(3)    , parameter :: zdead_big   = (/ 0.1362546,  2.421739,  6.9483532/)
   !----- Other constants. ----------------------------------------------------------------!
   logical               , parameter :: write_allom = .false.
   character(len=str_len), parameter :: allom_file  = 'allom_param.txt'
   !---------------------------------------------------------------------------------------!

   !----- Carbon-to-biomass ratio of plant tissues. ---------------------------------------!
   C2B    = 2.0

   !---------------------------------------------------------------------------------------! 
   !    This flag should be used to define whether the plant is tropical/subtropical or    !
   ! not.                                                                                  !
   !---------------------------------------------------------------------------------------! 
   is_tropical(1:4)   = .true.
   is_tropical(5:11)  = .false.
   is_tropical(12:13) = .false.
   is_tropical(14:15) = .true.
   is_tropical(16)    = .true.
   is_tropical(18:20) = .true.
   is_tropical(21:22) = .false.
   is_tropical(23:35) = .true.
   !---------------------------------------------------------------------------------------!
   !     This uses tropical allometry for DBH->Bleaf and DBH->Bdead, but otherwise it uses !
   ! the temperate properties.                                                             !
   !---------------------------------------------------------------------------------------!
   is_tropical(17)    = .true.

   !---------------------------------------------------------------------------------------! 
   !    This flag should be used to define whether the plant is tree or grass              !
   !---------------------------------------------------------------------------------------! 
   is_grass(1)     = .true.
   is_grass(2:4)   = .false.
   is_grass(5)     = .true.
   is_grass(6:11)  = .false.
   is_grass(12:15) = .true.
   is_grass(16)    = .true.
   is_grass(17)    = .false.
   is_grass(18:20) = .false.
   is_grass(21:35) = .false.

   !---------------------------------------------------------------------------------------!
   !     Wood density.  Currently only tropical PFTs need it.  C3 grass density will be    !
   ! used only for branch area purposes.                                                   !
   !---------------------------------------------------------------------------------------!
   !---- [KIM] new tropical parameters. ---------------------------------------------------!
   rho(1)     = 0.40   ! 0.40
   rho(2)     = 0.40   ! 0.40
   rho(3)     = 0.60   ! 0.60
   rho(4)     = 0.87   ! 0.87
   rho(5)     = 0.20   ! Copied from C4 grass
   rho(6:11)  = 0.34500 !Aspinwall 2011  ! Currently not used
   rho(12:13) = 0.20
   rho(14:15) = 0.20
   rho(16)    = 0.20
   rho(17)    = 0.54
   rho(18)    = 0.40   ! 0.40
   rho(19)    = 0.60   ! 0.60
   rho(20)    = 0.87   ! 0.87
   rho(21:22) = 0.40
   rho(23:35) = 0.60
   rho(23)     = 0.67 !0.74 ! one hardwood PFT !0.74
   rho(24)     = 0.87!0.85
   rho(25)     = 0.74!0.67!0.10 + rho_m!0.78!0.73!0.70
   rho(26)     = 0.74!!0.67!0.71!0.61!0.41!0.50
   rho(27)     = 0.52!0.64!0.75! one hardwood PFT 0.74!0.42!0.43!0.53!0.50
   rho(28)     = 0.64!0.75!0.73!0.68!0.77!0.85!0.70
   rho(29)     = 0.25!0.38!0.39
   rho(30:35)  = 0.65
   !---------------------------------------------------------------------------------------!

   !----- Specific leaf area [m leaf / kg C] ---------------------------------------------!
   !----- [KIM] - new tropical parameters. ------------------------------------------------!
   !SLA(1:4)   = 10.0**(2.4-0.46*log10(12.0/leaf_turnover_rate(1:4))) * C2B * 0.1
   !print*,'leaf_turnover_rate',leaf_turnover_rate
   SLA(1:4) = 10.0**(1.6923-0.3305*log10(12.0/leaf_turnover_rate(1:4)))
   SLA(5)     = 22.0
   SLA(6)     = 9.0 ! 6.0
   SLA(7)     =  9.0
   SLA(8)     = 10.0
   SLA(9)     = 30.0
   SLA(10)    = 24.2
   SLA(11)    = 60.0
   SLA(12:13) = 22.0
   SLA(14:15) = 10.0**((2.4-0.46*log10(12.0/leaf_turnover_rate(14:15)))) * C2B * 0.1
   ! SLA(14:15) = 10.0**(1.6923-0.3305*log10(12.0/leaf_turnover_rate(14:15)))
   SLA(16)    = 10.0**(2.4-0.46*log10(12.0/leaf_turnover_rate(16))) * C2B * 0.1
   SLA(17)    = 10.0
   SLA(18:20)   = 10.0**(2.4-0.46*log10(12.0/leaf_turnover_rate(18:20))) * C2B * 0.1
   SLA(21)    = 34.84!60.0!34.84
   SLA(22)    = 9.96!4.87*2.0/(3.14159/4.0)!divide by pi/4 to get cylindrical surface
   SLA(23:35) = 19.66
   SLA(23)    = 18.6!22.1 !17.6 !22.1! one hardwood PFT!
   SLA(24)    = 15.7!16.2
   SLA(25)    = 19.1!16.1!12.7!10.0 + SLA_m!16.8!20.0!19.7
   SLA(26)    = 18.7!16.1!12.7 !20.0!16.6!12.9!20.0!17.6
   SLA(27)    = 20.5!24.5!25.8 ! one hardwood PFT 22.1 !15.4!15.3!20.0!22.0
   SLA(28)    = 19.7!24.5 !25.0!25.3!23.9!20.0!16.3
   SLA(29)    = 20.8!19.6
   SLA(30:32) = 10.53 * 2.!9.52 * 2!5.7 * 2
   SLA(33:35) = 10.53 * 2 !8.7 * 2

   bark_thickness_beta(1:29) = 0.0
   bark_thickness_beta(30:32) = 0.142
   bark_thickness_beta(31) = 0.11
   bark_thickness_beta(33:35) = 0.11!0.0675
   bark_thickness_beta(34) = 0.0771
   bark_thickness_beta(35) = 0.0935!0.0872
   next_fire_year = 2013
   
   ! simulations in Panama, SLA is doubled
   !SLA(24:29) = 2 * SLA(24:29)

   SRA(1:35)     = 24. * 2.  ! m2/kgC Amazon data, fertile soil from Metcalfe et al. 2008
!   SRA(24)		= 15. * 2.
!   SRA(26)		= 24. * 2.
!   SRA(27)		= 33. * 2.
!   SRA(27)		= 42. * 2.
   
   root_dens(1:35) = rho(1:35) * 1000. / 2.   ! convert it to kgC m3

   Cap_stem(1:35) = min(400.,max(50.,-700. * (rho(1:35) - 0.3) + 400)) / 102.
   Cap_leaf(1:35) = 3.0e-3 / 102.         ! kgH2O/m2/m
 
   Cap_leaf(6)= 600*18*1e-6/102 !Domec 2009 for loblloly pine
   !---------------------------------------------------------------------------------------!
   !    Fraction of vertical branches.  Values are from Poorter et al. (2006):             !
   !                                                                                       !
   !    Poorter, L.; Bongers, L.; Bongers, F., 2006: Architecture of 54 moist-forest tree  !
   ! species: traits, trade-offs, and functional groups. Ecology, 87, 1289-1301.           !
   ! For simplicity, we assume similar numbers for temperate PFTs.                         !
   !---------------------------------------------------------------------------------------!
   horiz_branch(1)     = 0.50
   horiz_branch(2)     = 0.57
   horiz_branch(3)     = 0.39
   horiz_branch(4)     = 0.61
   horiz_branch(5)     = 0.50
   horiz_branch(6:8)   = 0.61
   horiz_branch(9)     = 0.57
   horiz_branch(10)    = 0.39
   horiz_branch(11)    = 0.61
   horiz_branch(12:15) = 0.50
   horiz_branch(16)    = 0.50
   horiz_branch(17)    = 0.61
   horiz_branch(18)     = 0.57
   horiz_branch(19)     = 0.39
   horiz_branch(20)     = 0.61
   horiz_branch(21)    = 0.57
   horiz_branch(22)    = 0.61
   horiz_branch(23:35) = 0.57
   !---------------------------------------------------------------------------------------!


   !----- Ratio between fine roots and leaves [kg_fine_roots/kg_leaves] -------------------!
   q(1)     = 1.0
   q(2)     = 0.89
   q(3)     = 0.89
   q(4)     = 0.89
!   q(2)     = 1.0
!   q(3)     = 1.0
!   q(4)     = 1.0
   q(5)     = 1.0
   q(6)     = 0.3463 ! 1.0
   q(7)     = 0.3463 ! 1.0
   q(8)     = 0.3463 ! 1.0
   q(9)     = 1.1274
   q(10)    = 1.1274
   q(11)    = 1.1274
   q(12:15) = 1.0
   q(16)    = 1.0
   q(17)    = 1.0
   q(18:20) = 1.0
   q(21)    = 1.8889!3.0
   q(22)    = 0.6078!1.104
   q(23:35) = 1.0
   q(30:32) = 0.42!0.642
   q(33:35) = 0.42

   sapwood_ratio(1:35) = 3900.0

   !---------------------------------------------------------------------------------------!
   !    Finding the ratio between sapwood and leaves [kg_sapwood/kg_leaves]                !
   !                                                                                       !
   !    KIM: ED1/ED2 codes and Moorcroft et al. had the incorrect ratio.  Since the mid-   !
   ! latitude parameters have been optimized using the wrong SLA, we keep the bug until    !
   ! it is updated...                                                                      !
   !---------------------------------------------------------------------------------------!
   qsw(1:4)    = SLA(1:4)   / sapwood_ratio(1:4)    !new is SLA(1:4)/(3900.0*2.0/1000.0)
   qsw(5:13)   = SLA(5:13)  / sapwood_ratio(5:13)
   qsw(14:15)  = SLA(14:15) / sapwood_ratio(14:15)  !new is SLA(14:15)(3900.0*2.0/1000.0)
   qsw(16)     = SLA(16)    / sapwood_ratio(16)
   qsw(17)     = SLA(17)    / sapwood_ratio(17)
   qsw(18:20)    = SLA(18:20)   / sapwood_ratio(18:20)    !new is SLA(1:4)/(3900.0*2.0/1000.0)
   qsw(21:35)    = SLA(21:35)   / sapwood_ratio(21:35)    !new is SLA(1:4)/(3900.0*2.0/1000.0) !JL
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !    Initial density of plants, for near-bare-ground simulations [# of individuals/m2]  !
   !---------------------------------------------------------------------------------------!
   init_density(1)     = 0.01
   init_density(2)     = 0.244875
   init_density(3)     = 0.697625 
   init_density(4)     = 0.422125 
   init_density(5)     = 0.1
   init_density(6:8)   = 0.1
   init_density(9:11)  = 0.1
   init_density(12:13) = 0.1
   init_density(14:15) = 0.1
   init_density(16)    = 0.1
   init_density(17)    = 0.1
   init_density(18:20) = 0.1
   init_density(21)    = 0.1
   init_density(22)    = 0.1 !0.1
   init_density(23:35) = 0.1 !0.25 
   ! change inti_dens so that the leaf area is 0.1 m2 for each simulation
   init_density(24) = 0.5
   init_density(26) = 1.
   init_density(27) = 0.5
   init_density(29) = 0.5
   init_density(30) = 0.5
   init_density(31) = 0.5
   init_density(32) = 0.5
   init_density(33) = 0.5
   init_density(34) = 0.5
   init_density(35) = 0.5
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !    Minimum height of an individual.                                                   !
   !---------------------------------------------------------------------------------------!
   hgt_min(1)     = 0.15
   hgt_min(2:4)   = 1.50
   hgt_min(5)     = 0.15
   hgt_min(6)     = 1.50
   hgt_min(7)     = 1.50
   hgt_min(8)     = 1.50
   hgt_min(9)     = 1.50
   hgt_min(10)    = 1.50
   hgt_min(11)    = 1.50
   hgt_min(12)    = 0.15
   hgt_min(13)    = 0.15
   hgt_min(14)    = 0.50
   hgt_min(15)    = 0.50
   hgt_min(16)    = 0.50
   hgt_min(17)    = 0.50
   hgt_min(18:20) = 1.50
   hgt_min(21) = 1.0
   hgt_min(22) = 1.0
   hgt_min(23:35) = 1.5 
   ! start with the min dbh so that the leaf area is 0.1 m2
  ! hgt_min(24) = 0.911
  ! hgt_min(26) = 2.434
  ! hgt_min(27) = 0.608
  ! hgt_min(29) = 2.374
   !---------------------------------------------------------------------------------------!



   !----- Reference height for diameter/height allometry (temperates only). ---------------!
   hgt_ref(1:5)   = 0.0
   hgt_ref(6:11)  = 1.3
   hgt_ref(12:15) = 0.0
   hgt_ref(16)    = 0.0
   hgt_ref(17)    = 0.0
   hgt_ref(18:20) = 0.0
   hgt_ref(21)    = 0.2111!Yarie+BL fit!0.74 Bond Lamberty!0.5 Yarie
   hgt_ref(22)    = 0.5601!Yarie+BL fit! 0.41 Bond Lamberty!0.64 Yarie
   hgt_ref(23:35) = 0.0

   !---------------------------------------------------------------------------------------!



   !----- Fraction of structural stem that is assumed to be above ground. -----------------!
   agf_bs  = 0.7  ! XXT
   agf_bsi = 1. / agf_bs
   
   !---------------------------------------------------------------------------------------!


   !------grasss biomass threshold for intense fire.--------------------------------------!
   !taken from Hoffman 2012 'Fuels or microclimate? Understanding the drivers of fire
   !feedbacks at savanna-forest boundaries Figures 4 and 6
   grass_biomass_threshold = 0.025 * agf_bs 
   

   !---------------------------------------------------------------------------------------!
   !   DBH/height allometry parameters.                                                    !
   !                                                                                       !
   !   WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!    !
   !   WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!    !
   !   WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!    !
   !   WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!    !
   !   WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!    !
   !   WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!    !
   !                                                                                       !
   !   These parameters have different meaning for tropical and temperate PFTs...          !
   !---------------------------------------------------------------------------------------!
   !----- DBH-height allometry intercept [m]. ---------------------------------------------!
   b1Ht(1:4)   = 0.37 * log(10.0)
   b1Ht(5)     = 0.4778
   b1Ht(6)     = 27.14
   b1Ht(7)     = 27.14
   b1Ht(8)     = 22.79
   b1Ht(9)     = 22.6799
   b1Ht(10)    = 25.18
   b1Ht(11)    = 23.3874
   b1Ht(12:13) = 0.4778
   b1Ht(14:15) = 0.37 * log(10.0)
   b1Ht(16)    = 0.37 * log(10.0)
   b1Ht(17)    = 0.37 * log(10.0)
   b1Ht(18:20) = 0.37 * log(10.0)
   b1Ht(21)    = 22.4914!Yarie+BL !21.085 Bond Lamberty!20.0 Yarie
   b1Ht(22)    = 18.0!fixed!23.259 Bond Lamberty!10.1 Yarie
   b1Ht(23:35) =  0.37 * log(10.0)
   b1Ht(23)    = 1.2436!0.8303 !0.6430    ! SLA 15 rho 8
   b1Ht(24)    = 0.3444!0.3360!0.3263    ! SLA 15 rho 8 
   b1Ht(25)    = 1.023!1.2436!0.8303    ! SLA 12 rho 6
   b1Ht(26)     = 0.98!1.023!1.2436!0.8303 ! SLA 12 rho 6
   b1Ht(27)     = -0.0682!-0.1615!0.7314!0.5343 !0.53  ! softwood E
   b1Ht(28)     = 0.6195!0.7314!0.53 ! hardwood D
   b1Ht(29) = 0.7393 !0.4052!0.38  ! not used now
   b1Ht(30) = 0.4052!0.38  ! not used now
   b1Ht(30:35)   = 0.98


   !----- DBH-height allometry slope [1/cm]. ----------------------------------------------!
   b2Ht(1:4)   = 0.64
   b2Ht(5)     = -0.75
   b2Ht(6)     = -0.03884
   b2Ht(7)     = -0.03884
   b2Ht(8)     = -0.04445 
   b2Ht(9)     = -0.06534
   b2Ht(10)    = -0.04964
   b2Ht(11)    = -0.05404
   b2Ht(12:13) = -0.75
   b2Ht(14:15) =  0.64
   b2Ht(16)    =  0.64
   b2Ht(17)    =  0.64
   b2Ht(18:20) =  0.64
   b2Ht(21)    = -0.0776!Yarie+BL fit!-0.0761!Bond-Lamberty!-0.1063 Yarie
   b2Ht(22)    = -0.0631!Yarie+BL fit!-0.0519!Bond Lamberty!-0.0634 Yarie
   b2Ht(23:35) = 0.64
   b2Ht(23)    =  0.4283!0.5145 !0.5736   ! hardwood D
   b2Ht(24)    =  0.6571!0.6462  ! hardwood D
   b2Ht(25)    =  0.4667!0.5145   ! hardwood D
   b2Ht(26)     = 0.4756!0.4629!0.5145  ! softwood D
   b2Ht(27)     = 0.8368!0.8599!0.6044 !0.6044    ! softwood E
   b2Ht(28)     = 0.5419!0.6044  ! hardwood D
   b2Ht(29)     = 0.5419!0.6892 !not used now
   b2Ht(30)     = 0.5419 !not used now
   b2Ht(30:35)  = 0.4756
   !----- Maximum Height. -----------------------------------------------------------------!
   hgt_max( 1) = 1.50
   hgt_max( 2) = 35.0
   hgt_max( 3) = 35.0
   hgt_max( 4) = 35.0
   hgt_max( 5) = 0.95  * b1Ht( 5)
   hgt_max( 6) = 0.999 * b1Ht( 6)
   hgt_max( 7) = 0.999 * b1Ht( 7)
   hgt_max( 8) = 0.999 * b1Ht( 8)
   hgt_max( 9) = 0.999 * b1Ht( 9)
   hgt_max(10) = 0.999 * b1Ht(10)
   hgt_max(11) = 0.999 * b1Ht(11)
   hgt_max(12) = 0.95  * b1Ht(12)
   hgt_max(13) = 0.95  * b1Ht(13)
   hgt_max(14) = 1.50
   hgt_max(15) = 1.50
   hgt_max(16) = 1.50
   hgt_max(17) = 35.0
   hgt_max(18:20) = 35.0
   hgt_max(21) = 20.0
   hgt_max(22) = 14.0
   hgt_max(23:35) = 35.0

   !----- Maximum DBH. --------------------------------------------------------------------!
   do ipft=1,n_pft
      min_dbh(ipft) = h2dbh(hgt_min(ipft),ipft)
      max_dbh(ipft) = h2dbh(hgt_max(ipft),ipft)
   end do
   max_dbh(21) = 30.!20. !19.3 ATT
   max_dbh(22) = 30.!13. !12.8 ATT
   max_dbh(30:35) = 30.
!   max_dbh(23:35) = 40.
   !---------------------------------------------------------------------------------------!
   !     DBH-leaf allometry.  Assign temperate PFTs outside the loop, and the tropical     !
   ! ones inside the loop.                                                                 !
   !---------------------------------------------------------------------------------------!
   !----- DBH-leaf allometry intercept [kg leaf biomass / plant * cm^(-b2Bl)]. ------------!
   b1Bl(1:4)   = 0.0
   b1Bl(5)     = 0.08
   b1Bl(6)     = 0.024
   b1Bl(7)     = 0.024
   b1Bl(8)     = 0.0454
   b1Bl(9)     = 0.0129
   b1Bl(10)    = 0.048
   b1Bl(11)    = 0.017
   b1Bl(12:13) = 0.08
   b1Bl(14:15) = 0.0
   b1Bl(16)    = 0.0
   b1Bl(17)    = 0.0
   b1Bl(18:20) = 0.0
   b1Bl(21)    = 0.0014 * 2.!Yarie+BL fit!0.006387!Yarie!0.0030 Bond Lamberty 2002!0.0028!test!
   b1Bl(22)    = 0.022 * 1.1!Yarie+BL fit!0.1208!Yarie!0.0461 Bond Lamberty 2002
   b1Bl(23:35) = 0.0
   !-----  DBH-leaf allometry slope [dimensionless]. --------------------------------------!
   b2Bl(1:4)   = 0.0
   b2Bl(5)     = 1.0
   b2Bl(6)     = 1.899
   b2Bl(7)     = 1.899
   b2Bl(8)     = 1.6829
   b2Bl(9)     = 1.7477
   b2Bl(10)    = 1.455
   b2Bl(11)    = 1.731
   b2Bl(12:13) = 1.0
   b2Bl(14:15) = 0.0
   b2Bl(16)    = 0.0
   b2Bl(17)    = 0.0
   b2Bl(18:20) = 0.0
   b2Bl(21)    = 2.5975!Yarie+BL fit!2.0!Yarie!2.3527 Bond Lamberty 2002
   b2Bl(22)    = 2.0485!Yarie+BL fit!1.2231!Yarie!1.8172 Bond Lamberty 2002
   b2Bl(23:35) = 0.0

   !------- Fill in the tropical PFTs, which are functions of wood density. ---------------!
   do ipft=1,n_pft
      if (is_tropical(ipft)) then
         select case(iallom)
         case (0,2:4)
            !---- ED-2.1 allometry. -------------------------------------------------------!
            b1Bl(ipft) = exp(a1 + c1l * b1Ht(ipft) + d1l * log(rho(ipft)))
            aux        = ( (a2l - a1) + b1Ht(ipft) * (c2l - c1l) + log(rho(ipft))          &
                         * (d2l - d1l)) * (1.0/log(dcrit))
            b2Bl(ipft) = C2B * b2l + c2l * b2Ht(ipft) + aux
         case (1)
            !---- Based on modified Chave et al. (2001) allometry. ------------------------!
            b1Bl(ipft) = C2B * exp(aleaf(1)) * rho(ipft) / aleaf(3)
            b2Bl(ipft) = aleaf(2)
         end select
      end if
   end do
   !---------------------------------------------------------------------------------------!

   b1Bl(23)  = 0.0062 * 2.
   b1Bl(24)  = 0.023 * 2. !0.0161 * 2.!0.0154 * 2.
   b1Bl(26)  = 0.008 * 2. !0.0065 * 2. !0.0063 * 2.
   b1Bl(27)  = 0.012 * 2. !0.0125 * 2. !0.0140 * 2.
   b1Bl(29)  = 0.003 * 2. !0.0075 * 2. !0.0060 * 2.
   b1Bl(30:35)  = 0.008 * 2.

   b2Bl(23)  =  1.96
   b2Bl(24)  =  1.93!1.99!2.00
   b2Bl(26)  =  2.12!2.17!2.18
   b2Bl(27)  =  1.75!1.81!1.81
   b2Bl(29)  =  2.04!2.107!2.11
   b2Bl(30:35)  = 2.12

   !---------------------------------------------------------------------------------------!
   !     DBH-stem allometry.  Assign temperate PFTs outside the loop, and the tropical     !
   ! ones inside the loop.                                                                 !
   !---------------------------------------------------------------------------------------!
   !----- DBH-stem allometry intercept [kg stem biomass / plant * cm^(-b2Bs)] -------------!
   b1Bs_small(1:4)   = 0.0 
   b1Bs_small(5)     = 1.0e-5
   b1Bs_small(6)     = 0.0792/0.8*0.9 !Drake 2010
   b1Bs_small(7)     = 0.147
   b1Bs_small(8)     = 0.1617
   b1Bs_small(9)     = 0.02648
   b1Bs_small(10)    = 0.1617
   b1Bs_small(11)    = 0.235
   b1Bs_small(12:13) = 1.0e-5
   b1Bs_small(14:15) = 0.0 
   b1Bs_small(16)    = 0.0 
   b1Bs_small(17)    = 0.0
   b1Bs_small(18:20) = 0.0
   b1Bs_small(21)    = 0.0255!Yarie+BL fit!0.1474!Bond Lamberty! 0.061 Yarie
   b1Bs_small(22)    = 0.0404!Yarie+BL fit!0.1410!Bond Lamberty!.1449 Yarie
   b1Bs_small(23:35) = 0.0
   !----- DBH-stem allometry slope [dimensionless]. ---------------------------------------!
   b2Bs_small(1:4)   = 0.0
   b2Bs_small(5)     = 1.0
   b2Bs_small(6)     = 2.4349 !Dan
   b2Bs_small(7)     = 2.238
   b2Bs_small(8)     = 2.1536
   b2Bs_small(9)     = 2.95954
   b2Bs_small(10)    = 2.4572
   b2Bs_small(11)    = 2.2518
   b2Bs_small(12:13) = 1.0
   b2Bs_small(14:15) = 0.0
   b2Bs_small(16)    = 0.0
   b2Bs_small(17)    = 0.0
   b2Bs_small(18:20) = 0.0
   b2Bs_small(21)    = 2.8440!Yarie+BL fit!2.2965!BondLamberty!2.4806 Yarie
   b2Bs_small(22)    = 2.6587!Yarie+BL fit!2.3170!BondLamberty!2.1322 Yarie
   b2Bs_small(23:35) = 0.0
   !---------------------------------------------------------------------------------------!
   !     The temperate PFTs use the same b1Bs and b2Bs for small and big trees, copy the   !
   ! values.                                                                               !
   !---------------------------------------------------------------------------------------!
   b1Bs_big(:) = b1Bs_small(:)
   b2Bs_big(:) = b2Bs_small(:)
   !------- Fill in the tropical PFTs, which are functions of wood density. ---------------!
   do ipft = 1, n_pft
      if (is_tropical(ipft)) then
         select case (iallom)
         case (0)
            !---- ED-2.1 allometry. -------------------------------------------------------!
            b1Bs_small(ipft) = exp(a1 + c1d * b1Ht(ipft) + d1d * log(rho(ipft)))
            b1Bs_big  (ipft) = exp(a1 + c1d * log(hgt_max(ipft)) + d1d * log(rho(ipft)))

            aux              = ( (a2d - a1) + b1Ht(ipft) * (c2d - c1d) + log(rho(ipft))    &
                               * (d2d - d1d)) * (1.0/log(dcrit))
            b2Bs_small(ipft) = C2B * b2d + c2d * b2Ht(ipft) + aux

            aux              = ( (a2d - a1) + log(hgt_max(ipft)) * (c2d - c1d)             &
                               + log(rho(ipft)) * (d2d - d1d)) * (1.0/log(dcrit))
            b2Bs_big  (ipft) = C2B * b2d + aux

         case (1)
            !---- Based on modified Chave et al. (2001) allometry. ------------------------!
            b1Bs_small(ipft) = C2B * exp(adead_small(1)) * rho(ipft) / adead_small(3)
            b2Bs_small(ipft) = adead_small(2)
            b1Bs_big(ipft)   = C2B * exp(adead_big(1))   * rho(ipft) / adead_big(3)
            b2Bs_big(ipft)   = adead_big(2)

         case (2:4)
            !---- Based an alternative modification of Chave et al. (2001) allometry. -----!
            b1Bs_small(ipft) = C2B * exp(zdead_small(1)) * rho(ipft) / zdead_small(3)
            b2Bs_small(ipft) = zdead_small(2)
            b1Bs_big(ipft)   = C2B * exp(zdead_big(1))   * rho(ipft) / zdead_big(3)
            b2Bs_big(ipft)   = zdead_big(2)

         end select
      end if

      !----- Assigned for both cases, although it is really only needed for tropical. -----!
      bdead_crit(ipft) = dbh2bd(max_dbh(ipft),ipft)
   end do
   !---------------------------------------------------------------------------------------!

   ! heartwood accounts for 30% of the total basal area
   ! for tropical PFTs

   b1Bs_small(2:4) = 0.3 * b1Bs_small(2:4)
   b1Bs_big(2:4)   = 0.3 * b1Bs_big(2:4)

   b2Bs_small(24)   = 2.33
   b1Bs_small(24)   = exp(-2.82) / agf_bs

   b2Bs_small(26)   = 2.28
   b1Bs_small(26)   = exp(-2.84) / agf_bs

   b2Bs_small(23)   = 2.28
   b1Bs_small(23)   = exp(-2.84) / agf_bs

   b2Bs_small(27)   = 2.32
   b1Bs_small(27)   = exp(-3.38) / agf_bs !exp(-2.89) / agf_bs

   b2Bs_small(29)   = 2.29
   b1Bs_small(29)   = exp(-4.18) / agf_bs

   b2Bs_small(30:35)   = 2.28
   b1Bs_small(30:35)   = exp(-2.84) / agf_bs

   b1Bs_big(21:35) = b1Bs_small(21:35)
   b2Bs_big(21:35) = b2Bs_small(21:35)

   b1Bsap(1:35) = exp(-2.70) / agf_bs
   b2Bsap(1:35) = 2.02

   b1Bsap(2:4) = 0.7 * b1Bs_small(2:4) / 0.3
   b2Bsap(2:4) = b2Bs_small(2:4)

   b1Bsap(24)   = exp(-1.88) / agf_bs
   b2Bsap(24)   = 2.33

   b1Bsap(23)   = exp(-1.90) / agf_bs
   b2Bsap(23)   = 2.28

   b1Bsap(26)   = exp(-1.90) / agf_bs
   b2Bsap(26)   = 2.28

   b1Bsap(27)   = exp(-2.43) / agf_bs!exp(-1.95) / agf_bs
   b2Bsap(27)   = 2.32

   b1Bsap(29)   = exp(-3.23) / agf_bs
   b2Bsap(29)   = 2.29

   b1Bsap(6) = 0.0792/0.8*0.1 
   b2Bsap(6)= 2.4349

   b1Bsap(30:32)   = exp(-1.90) / agf_bs ! * ((bark_thickness_beta(30:32) * -1.7521 + 0.7102) / 0.7)
   b2Bsap(30:32)   = 2.28

   b1Bsap(33:35)   = exp(-1.90) / agf_bs ! * ((bark_thickness_beta(33:35) * -1.7521 + 0.7102) / 0.7)
   b2Bsap(33:35)   = 2.28
   !---------------------------------------------------------------------------------------!
   !     DBH-crown allometry.  Here we use Dietze and Clark (2008) parameters for the      !
   ! temperate PFTs and Poorter et al. (2006) for the tropical PFTs.  The Poorter et al.   !
   ! coefficients are converted to be functions of DBH rather than height.                 !
   !---------------------------------------------------------------------------------------!
   !----- Intercept. ----------------------------------------------------------------------!
   b1Ca(1:4)   = exp(-1.853) * exp(b1Ht(1:4)) ** 1.888
   b1Ca(5:13)  = 2.490154
   b1Ca(14:17) = exp(-1.853) * exp(b1Ht(14:17)) ** 1.888
   b1Ca(18:20)   = exp(-1.853) * exp(b1Ht(18:20)) ** 1.888
   b1Ca(21:22) = 2.490154
   b1Ca(23:35) = exp(-1.853) * exp(b1Ht(23:35)) ** 1.888
   !----- Slope.  -------------------------------------------------------------------------!
   b2Ca(1:4)   = b2Ht(1:4) * 1.888
   b2Ca(5:13)  = 0.8068806
   b2Ca(14:17) = b2Ht(14:17) * 1.888
   b2Ca(18:20) = b2Ht(18:20) * 1.888
   b2Ca(21:22) = 0.8068806
   b2Ca(23:35) = b2Ht(23:35) * 1.888
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     DBH-Root depth allometry.  Check which allometry to use.  Notice that b?Rd have   !
   ! different meanings depending on the allometry. b?Vol is always defined because we     !
   ! may want to estimate the standing volume for other reasons (e.g. the characteristic   !
   ! diameter of branches).                                                                !
   !---------------------------------------------------------------------------------------!
   ! shape parameter of root profile
	root_beta(1:35)   = 0.96 ** (200)   ! Jackson et al. 1995
	root_beta(24)   = 0.98 ** (200)   
	root_beta(23)   = 0.97 ** (200)   
	root_beta(26)   = 0.97 ** (200)   
	root_beta(27)   = 0.96 ** (200)   
        root_beta(29)   = 0.95 ** (200)
        root_beta(30:35)   = 0.97 ** (200)   

   b1Vol(1:35)  = 0.65 * pi1 * 0.11 * 0.11
   b2Vol(1:35)  = 2.0

   select case (iallom)
   case (0:2)
      b1Rd(1)     = - 0.700
      b1Rd(2:4)   = - exp(0.545 * log(10.))
      b1Rd(5)     = - 0.700
      b1Rd(6:11)  = - exp(0.545 * log(10.))
      b1Rd(12:16) = - 0.700
      b1Rd(17)    = - exp(0.545 * log(10.))
      b1Rd(18:20) = - exp(0.545 * log(10.))
      b1Rd(21)    = -1.0!- 0.1!- 1.0!- exp(0.545 * log(10.))
      b1Rd(22)    = -0.30!- 0.0757 !- 0.300!- exp(0.545 * log(10.))
      b1Rd(23:35) = - exp(0.545 * log(10.))

      b2Rd(1)     = 0.000
      b2Rd(2:4)   = 0.277
      b2Rd(5)     = 0.000
      b2Rd(6:11)  = 0.277
      b2Rd(12:16) = 0.000
      b2Rd(17)    = 0.277
      b2Rd(18:20) = 0.277
      b2Rd(21)    = 0.0! 1.0!0.0
      b2Rd(22)    = 0.0!0.5981!0.0
      b2Rd(23:35) = 0.277
   case (3)
      !------------------------------------------------------------------------------------!
      !     Based on Kenzo et al. (2009), but we had to apply some correction to make it   !
      ! dependent on DBH rather than Diameter at 0m.                                       !
      ! Source: Kenzo, T., and co-authors: 2009.  Development of allometric relationships  !
      !            for accurate estimation of above- and below-ground biomass in tropical  !
      !            secondary forests in Sarawak, Malaysia. J. Trop. Ecology, 25, 371-386.  !
      !------------------------------------------------------------------------------------!
      b1Rd(1:35)  = -0.2185333
      b2Rd(1:35)  =  0.5436442
   case (4)
      !------------------------------------------------------------------------------------!
      !     This is just a test, not based on any paper.  This is simply a fit that would  !
      ! put the roots 0.5m deep for plants 0.15m-tall and 5 m for plants 35-m tall.        !
      !------------------------------------------------------------------------------------!
      b1Rd(1:35)  = -1.1140580
      b2Rd(1:35)  =  0.4223014
   end select
   !---------------------------------------------------------------------------------------!
   b2Rd(23:35) =  0.5436442
   b1Rd(23:30) = -0.2185333 * 0.5 * 1.0
   b1Rd(24) = -0.2185333 * 0.5 * 2.0   !* log(0.1) / log(root_beta(24)) !rd_m !2.0 
   b1Rd(23) = -0.2185333 * 0.5 * 1.6   !* log(0.1) / log(root_beta(23)) !rd_m !2.0 
   b1Rd(26) = -0.2185333 * 0.5 * 1.6   !* log(0.1) / log(root_beta(26)) !rd_m !2.0 
   b1Rd(27) = -0.2185333 * 0.5 * 1.2   !* log(0.1) / log(root_beta(27)) !rd_m !2.0 
   b1Rd(29) = -0.2185333 * 0.5 * 1.0   !* log(0.1) / log(root_beta(29)) !rd_m !2.0
   b1Rd(30:35) = -0.2185333 * 0.5 * 1.6   !* log(0.1) / log(root_beta(26)) !rd_m !2.0
   b1Rd(1) = -1.1
   b2Rd(1) = 0.

   if (write_allom) then
      open (unit=18,file=trim(allom_file),status='replace',action='write')
      write(unit=18,fmt='(209a)') ('-',n=1,209)
      write(unit=18,fmt='(18(1x,a))') '         PFT','    Tropical','         Rho'         &
                                     ,'        b1Ht','        b2Ht','        b1Bl'         &
                                     ,'        b2Bl','  b1Bs_Small','  b2Bs_Small'         &
                                     ,'    b1Bs_Big','    b1Bs_Big','        b1Bl'         &
                                     ,'        b2Bl','     Hgt_min','     Hgt_max'         &
                                     ,'     Min_DBH','     Max_DBH','   Bdead_CRT'
      write(unit=18,fmt='(209a)') ('-',n=1,209)
      do ipft=1,n_pft
         write (unit=18,fmt='(8x,i5,12x,l1,16(1x,es12.5))')                                &
                        ipft,is_tropical(ipft),rho(ipft),b1Ht(ipft),b2Ht(ipft),b1Bl(ipft)  &
                       ,b2Bl(ipft),b1Bs_small(ipft),b2Bs_small(ipft),b1Bs_big(ipft)        &
                       ,b2Bs_big(ipft),b1Ca(ipft),b2Ca(ipft),hgt_min(ipft),hgt_max(ipft)   &
                       ,min_dbh(ipft),max_dbh(ipft),bdead_crit(ipft)
      end do
      write(unit=18,fmt='(209a)') ('-',n=1,209)
      close(unit=18,status='keep')
   end if

   !---------------------------------------------------------------------------------------!
   !    Define the branching parameters, following Jrvel (2004)                          !
   !---------------------------------------------------------------------------------------!
   !----- Branching ratio -----------------------------------------------------------------!
   rbranch(1)     = 4.24
   rbranch(2:4)   = 4.23
   rbranch(5)     = 4.24
   rbranch(6:8)   = 4.44
   rbranch(9:11)  = 4.24
   rbranch(12:15) = 4.24
   rbranch(16)    = 4.24
   rbranch(17)    = 4.44
   rbranch(18:20) = 4.23
   rbranch(21)    = 4.24
   rbranch(22)    = 4.44
   rbranch(23:35) = 4.23
   !----- Diameter ratio ------------------------------------------------------------------!
   rdiamet(1)     = 5.00
   rdiamet(2:4)   = 1.86
   rdiamet(5)     = 5.00
   rdiamet(6:8)   = 2.04
   rdiamet(9:11)  = 1.86
   rdiamet(12:15) = 5.00
   rdiamet(16)    = 5.00
   rdiamet(17)    = 2.04
   rdiamet(18:20) = 1.86
   rdiamet(21)    = 1.86
   rdiamet(22)    = 2.04
   rdiamet(23:35) = 1.86
   !----- Length ratio. Jrvel used rdiamet^2/3, so do we... -----------------------------!
   rlength(1:35)  = rdiamet(1:35)**twothirds
   !----- Minimum diameter to consider [cm]. ----------------------------------------------!
   diammin(1:35)  = 1.0
   !----- Number of trunks.  Usually this is 1. -------------------------------------------!
   ntrunk(1:35)   = 1.0
   
   !---------------------------------------------------------------------------------------!
   !     The following variables are used to fit a smooth curve in the (sparse) values     !
   ! provided by Conijn (1995). This should be definitely improved...  The fitting curve   !
   ! is a + b*erf(c*bbranch+d)
   !---------------------------------------------------------------------------------------!
   conijn_a(1)     = 1.0
   conijn_a(2:4)   = 0.96305883
   conijn_a(5)     = 1.0
   conijn_a(6:11)  = 0.96305883
   conijn_a(12:15) = 1.0
   conijn_a(16)    = 1.0
   conijn_a(17)    = 0.96305883
   conijn_a(18:20)   = 0.96305883
   conijn_a(21:35) = 0.96305883

   conijn_b(1)     = 0.0
   conijn_b(2:4)   = -0.7178682
   conijn_b(5)     = 0.0
   conijn_b(6:11)  = -0.7178682
   conijn_b(12:15) = 0.0
   conijn_b(16)    = 0.0
   conijn_b(17)    = -0.7178682
   conijn_b(18:20)   = -0.7178682
   conijn_b(21:35) = -0.7178682

   conijn_c(1)     = 0.0
   conijn_c(2:4)   = 0.00490734
   conijn_c(5)     = 0.0
   conijn_c(6:11)  = 0.00490734
   conijn_c(12:15) = 0.0
   conijn_c(16)    = 0.0
   conijn_c(17)    = 0.00490734
   conijn_c(18:20)   = 0.00490734
   conijn_c(21:35) = 0.00490734

   conijn_d(1)     = 0.0
   conijn_d(2:4)   = -0.0456370
   conijn_d(5)     = 0.0
   conijn_d(6:11)  = -0.0456370
   conijn_d(12:15) = 0.0
   conijn_d(16)    = 0.0
   conijn_d(17)    = -0.0456370
   conijn_d(18:20)   = -0.0456370
   conijn_d(21:35) = -0.0456370
   return
end subroutine init_pft_alloc_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_pft_nitro_params()

use pft_coms, only: c2n_leaf,		  &
                    Vm0,		  &
                    SLA,	   	  &
                    c2n_slow,             &
                    c2n_structural,       &
                    c2n_storage,          &
                    c2n_stem,             &
                    l2n_stem,             &
                    C2B,                  &
                    plant_N_supply_scale, &
                    root_beta,            &
                    C_resorption_factor,  &
                    N_resorption_factor

implicit none

c2n_slow       = 63.0!10.0  ! Carbon to Nitrogen ratio, slow pool.
c2n_structural = 150.0 ! Carbon to Nitrogen ratio, structural pool.
c2n_storage    = 150.0 ! Carbon to Nitrogen ratio, storage pool.
c2n_stem       = 150.0 ! Carbon to Nitrogen ratio, structural stem.
l2n_stem       = 150.0 ! Carbon to Nitrogen ratio, structural stem.


plant_N_supply_scale = 0.5 

c2n_leaf(1:5)    = 1000.0 / ((0.11289 + 0.12947 *   Vm0(1:5)) * SLA(1:5)  )
c2n_leaf(6) = 51./1.2 !Dan
!c2n_leaf(6)      = 1000.0 / ((0.11289 + 0.12947 *     15.625) * SLA(6)    )
c2n_leaf(7)      = 1000.0 / ((0.11289 + 0.12947 *     15.625) * SLA(7)    )
c2n_leaf(8)      = 1000.0 / ((0.11289 + 0.12947 *       6.25) * SLA(8)    )
c2n_leaf(9)      = 1000.0 / ((0.11289 + 0.12947 *      18.25) * SLA(9)    )
c2n_leaf(10)     = 1000.0 / ((0.11289 + 0.12947 *     15.625) * SLA(10)   )
c2n_leaf(11)     = 1000.0 / ((0.11289 + 0.12947 *       6.25) * SLA(11)   )
c2n_leaf(12:15)  = 1000.0 / ((0.11289 + 0.12947 * Vm0(12:15)) * SLA(12:15))
c2n_leaf(16:17)  = 1000.0 / ((0.11289 + 0.12947 * Vm0(16:17)) * SLA(16:17))
c2n_leaf(18:20)  = 1000.0 / ((0.11289 + 0.12947 *   Vm0(18:20)) * SLA(18:20))
c2n_leaf(21)     = 16.0!20.0
c2n_leaf(22)     = 63.0
!c2n_leaf(23:35)    = 1000.0 / ((0.11289 + 0.12947 *   Vm0(23:35)) * SLA(23:35)  ) / 2. ! From Jennifer Powers Data
c2n_leaf(23:35) = exp(-0.4972 * log(SLA(23:35)) + 4.5071) ! Powers and Tiffin 2010
print*,'c2n_leaf',c2n_leaf(23:29)
C_resorption_factor(1:35) = 0.25
N_resorption_factor(1:35) = 0.48

return
end subroutine init_pft_nitro_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!   This subroutine sets up some PFT and leaf dependent properties.                        !
!------------------------------------------------------------------------------------------!
subroutine init_pft_leaf_params()
   use rk4_coms       , only : ibranch_thermo       ! ! intent(in)
   use pft_coms       , only : phenology            & ! intent(out)
                             , clumping_factor      & ! intent(out)
                             , b1Cl                 & ! intent(out)
                             , b2Cl                 & ! intent(out)
                             , c_grn_leaf_dry       & ! intent(out)
                             , c_ngrn_biom_dry      & ! intent(out)
                             , wat_dry_ratio_grn    & ! intent(out)
                             , wat_dry_ratio_ngrn   & ! intent(out)
                             , delta_c              ! ! intent(out)
   use consts_coms    , only : t3ple                ! ! intent(out) 
   use phenology_coms , only :iphen_scheme

   implicit none

   select case (iphen_scheme)
   case (-1)
      phenology(1:8)   = 0
      phenology(9:11)  = 2
      phenology(12:17) = 0
      phenology(18:20)   = 0
      phenology(21)    = 2
      phenology(22:35) = 0
   case (0,1)
      phenology(1)     = 1
      phenology(2:4)   = 1
      phenology(5)     = 1
      phenology(6:8)   = 0
      phenology(9:11)  = 2
      phenology(12:15) = 1
      phenology(16)    = 1
      phenology(17)    = 0
      phenology(18:20) = 1
      phenology(21)    = 2
      phenology(22)    = 0
      phenology(23:35) = 1
   case (2)
      phenology(1)     = 4
      phenology(2:4)   = 4
      phenology(5)     = 4
      phenology(6:8)   = 0
      phenology(9:11)  = 2
      phenology(12:15) = 4
      phenology(16)    = 4
      phenology(17)    = 0
      phenology(18:20) = 4
      phenology(21)    = 2
      phenology(22)    = 0
      phenology(23:35) = 4

   case (3)
      phenology(1)     = 4
      phenology(2:4)   = 3
      phenology(5)     = 4
      phenology(6:8)   = 0
      phenology(9:11)  = 2
      phenology(12:15) = 4
      phenology(16)    = 4
      phenology(17)    = 0
      phenology(18:20) = 3
      phenology(21)    = 2
      phenology(22)    = 0
      phenology(23:35) = 4
   case (4,5)
		! soil-moisture based drought-phenology scheme
      phenology(1)     = 4
      phenology(2:4)   = 4
      phenology(5)     = 4
      phenology(6:8)   = 0
      phenology(9:11)  = 2
      phenology(12:15) = 4
      phenology(16)    = 4
      phenology(17)    = 0
      phenology(18:20) = 3
      phenology(21)    = 2
      phenology(22)    = 0
      phenology(23:35) = 5
   end select

   clumping_factor(1)     = 1.000d0
   clumping_factor(2:4)   = 1.000d0 ! 7.350d-1
   clumping_factor(5)     = 8.400d-1
   clumping_factor(6:8)   = 7.350d-1
   clumping_factor(9:11)  = 8.400d-1
   clumping_factor(12:13) = 8.400d-1
   clumping_factor(14:15) = 1.000d0
   clumping_factor(16)    = 1.000d0 ! 8.400d-1
   clumping_factor(17)    = 1.000d0 ! 7.350d-1
   clumping_factor(18:20)  = 1.000d0 ! 7.350d-1
   clumping_factor(21)    = 8.400d-1
   clumping_factor(22)    = 7.350d-1
   clumping_factor(23:35) = 8.400d-1

   !---------------------------------------------------------------------------------------!
   !      The following parameters are second sources found in Gu et al. (2007)            !
   !---------------------------------------------------------------------------------------!
   c_grn_leaf_dry(1:35)      = 3218.0    ! Jones 1992  J/(kg K)
   c_ngrn_biom_dry(1:35)     = 1256.0    ! Forest Products Laboratory 
   wat_dry_ratio_grn(1:35)   = 2.5       ! 
   !wat_dry_ratio_grn(1:35)   = 1.5       ! Ceccato et al. 2001
   wat_dry_ratio_ngrn(1:35)  = 0.7       ! Forest Products Laboratory
   !---------------------------------------------------------------------------------------!
   !     Delta-c is found using the second term of the RHS of equation 5, assuming         !
   ! T=T3ple.  This is a simplification, but the specific heat usually varies by 3J/kg/K   !
   ! between 173K and 341K, so removing the dependence on temperature is not that bad      !
   ! assumption.                                                                           !
   !---------------------------------------------------------------------------------------!
   delta_c(1:35) = 100. * wat_dry_ratio_ngrn(1:35)                                         &
                 * (-0.06191 + 2.36e-4 * t3ple - 1.33e-2 * wat_dry_ratio_ngrn(1:35))

   !---------------------------------------------------------------------------------------!
   !     These are used to compute the crown length, which will then be used to find the   !
   ! height of the bottom of the crown.  This allometry is based on:                       !
   !                                                                                       !
   ! Poorter L., L. Bongers, F. Bongers, 2006: Architecture of 54 moist-forest tree        !
   !     species: traits, trade-offs, and functional groups. Ecology, 87, 1289-1301.       !
   !---------------------------------------------------------------------------------------!
   !----- Intercept. ----------------------------------------------------------------------!
   b1Cl(1)     = 0.99
   b1Cl(2:4)   = 0.3106775
   b1Cl(5)     = 0.99
   b1Cl(6:11)  = 0.3106775
   b1Cl(12:16) = 0.99
   b1Cl(17)    = 0.3106775
   b1Cl(18:20)   = 0.3106775
   b1Cl(21:35)   = 0.3106775
   !----- Slope. ----------------------------------------------------------------------!
   b2Cl(1)     = 1.0
   b2Cl(2:4)   = 1.098
   b2Cl(5)     = 1.0
   b2Cl(6:11)  = 1.098
   b2Cl(12:16) = 1.0
   b2Cl(17)    = 1.098
   b2Cl(18:20)   = 1.098
   b2Cl(21:35)   = 1.098 
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_pft_leaf_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!    This subroutine sets some reproduction-related parameters.                            !
!------------------------------------------------------------------------------------------!
subroutine init_pft_repro_params()

   use pft_coms , only : r_fract            & ! intent(out)
                       , st_fract           & ! intent(out)
                       , seed_rain          & ! intent(out)
                       , nonlocal_dispersal & ! intent(out)
                       , repro_min_h        ! ! intent(out)
   implicit none

   r_fract(1)              = 0.3
   r_fract(2:4)            = 0.3
   r_fract(5)              = 0.3
   r_fract(6:11)           = 0.3
   r_fract(12:15)          = 0.3
   r_fract(16)             = 0.3
   r_fract(17)             = 0.3
   r_fract(18:20)          = 0.3
   r_fract(21)             = 0.3
   r_fract(22)             = 0.3
   r_fract(23:35)          = 0.3

   st_fract(1)             = 0.0
   st_fract(2:4)           = 0.0
   st_fract(5)             = 0.0
   st_fract(6:11)          = 0.0
   st_fract(12:15)         = 0.0
   st_fract(16)            = 0.0
   st_fract(17)            = 0.0
   st_fract(18:20)         = 0.0
   st_fract(21:35)         = 0.0

   seed_rain(1:35)         = 0.00 !0.01/12.
   seed_rain(1)         = 0.001/12.
   !seed_rain(30:31)        = 0.01/12.  ! This adds 0.01 kgC/m2/year.
  ! seed_rain(33:34)        = 0.01/12.

   nonlocal_dispersal(1:5) = 1.0
!   nonlocal_dispersal(1:2) = 1.0
!   nonlocal_dispersal(3) = 0.325
!   nonlocal_dispersal(4) = 0.074


   nonlocal_dispersal(6:7) = 0.766
   nonlocal_dispersal(8)   = 0.001
   nonlocal_dispersal(9)   = 1.0
   nonlocal_dispersal(10)  = 0.325
   nonlocal_dispersal(11)  = 0.074
   nonlocal_dispersal(16)  = 1.0
   nonlocal_dispersal(17)  = 0.766
   nonlocal_dispersal(18:20) = 1.0
   nonlocal_dispersal(21)  = 0.95
   nonlocal_dispersal(22)  = 0.10
   nonlocal_dispersal(23:35) = 0.95  

   repro_min_h(1)          = 0.0
   repro_min_h(2:4)        = 5.0
   repro_min_h(5)          = 0.0
   repro_min_h(6:11)       = 5.0
   repro_min_h(12:15)      = 0.0
   repro_min_h(16)         = 0.0
   repro_min_h(17)         = 5.0
   repro_min_h(18:20)      = 5.0
   repro_min_h(21)         = 4.0
   repro_min_h(22)         = 4.0
   repro_min_h(23:35)      = 2.0!5.0


!OVERRIDE
!!repro_min_h = 20.

   return
end subroutine init_pft_repro_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!    This subroutine will assign some variables that depend on the definition of other     !
! PFT parameters.  As such, this should be the last init_pft subroutine to be called.      !
!------------------------------------------------------------------------------------------!
subroutine init_pft_derived_params()
   use decomp_coms          , only : f_labile             ! ! intent(in)
   use ed_max_dims          , only : n_pft                & ! intent(in)
                                   , str_len              ! ! intent(in)
   use consts_coms          , only : onesixth             & ! intent(in)
                                   , twothirds            ! ! intent(in)
   use pft_coms             , only : init_density         & ! intent(in)
                                   , c2n_leaf             & ! intent(in)
                                   , c2n_stem             & ! intent(in)
                                   , b1Ht                 & ! intent(in)
                                   , b2Ht                 & ! intent(in)
                                   , hgt_min              & ! intent(in)
                                   , hgt_ref              & ! intent(in)
                                   , q                    & ! intent(in)
                                   , qsw                  & ! intent(in)
                                   , sla                  & ! intent(in)
                                   , pft_name16           & ! intent(in)
                                   , hgt_max              & ! intent(in)
                                   , max_dbh              & ! intent(in)
                                   , min_recruit_size     & ! intent(out)
                                   , min_cohort_size      & ! intent(out)
                                   , negligible_nplant    & ! intent(out)
                                   , c2n_recruit          & ! intent(out)
                                   , lai_min              ! ! intent(out)
   use phenology_coms       , only : elongf_min           ! ! intent(in)
   use allometry            , only : h2dbh                & ! function
                                   , dbh2h                & ! function
                                   , dbh2bl               & ! function
                                   , dbh2bd               &
								   , dbh2bs				! ! function
   implicit none
   !----- Local variables. ----------------------------------------------------------------!
   integer                           :: ipft
   real                              :: dbh
   real                              :: huge_dbh
   real                              :: huge_height
   real                              :: balive_min
   real                              :: bleaf_min
   real                              :: bdead_min
   real                              :: balive_max
   real                              :: bleaf_max
   real                              :: bdead_max
   real                              :: min_plant_dens
   logical               , parameter :: print_zero_table = .false.
   character(len=str_len), parameter :: zero_table_fn    = 'minimum.size.txt'
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     The minimum recruitment size and the recruit carbon to nitrogen ratio.  Both      !
   ! parameters actually depend on which PFT we are solving, since grasses always have     !
   ! significantly less biomass.                                                           !
   !---------------------------------------------------------------------------------------!
   if (print_zero_table) then
      open  (unit=61,file=trim(zero_table_fn),status='replace',action='write')
      write (unit=61,fmt='(18(a,1x))')                '  PFT',        'NAME            '   &
                                              ,'     HGT_MIN','         DBH'               &
                                              ,'   BLEAF_MIN','   BDEAD_MIN'               &
                                              ,'  BALIVE_MIN','   BLEAF_MAX'               &
                                              ,'   BDEAD_MAX','  BALIVE_MAX'               &
                                              ,'   INIT_DENS','MIN_REC_SIZE'               &
                                              ,'MIN_COH_SIZE',' NEGL_NPLANT'               &
                                              ,'         SLA','     LAI_MIN'               &
                                              ,'     HGT_MAX','     MAX_DBH'
   end if
   min_plant_dens = onesixth * minval(init_density)
   do ipft = 1,n_pft

      !----- Find the DBH and carbon pools associated with a newly formed recruit. --------!
      dbh        = h2dbh(hgt_min(ipft),ipft)
      bleaf_min  = dbh2bl(dbh,ipft)
      bdead_min  = dbh2bd(dbh,ipft)
      balive_min = bleaf_min * (1.0 + q(ipft)) * 2. + dbh2bs(dbh,ipft)

      !------------------------------------------------------------------------------------!
      !   Find the maximum bleaf and bdead supported.  This is to find the negligible      !
      ! nplant so we ensure that the cohort is always terminated if its mortality rate is  !
      ! very high.                                                                         !
      !------------------------------------------------------------------------------------!
      huge_dbh    = 3. * max_dbh(ipft)
      huge_height = dbh2h(ipft, max_dbh(ipft))
      bleaf_max   = dbh2bl(huge_dbh,ipft)
      bdead_max   = dbh2bd(huge_dbh,ipft)
      balive_max  = bleaf_max * (1.0 + q(ipft)) * 2 + dbh2bs(huge_dbh,ipft)
      !------------------------------------------------------------------------------------!



      !------------------------------------------------------------------------------------!
      !    The definition of the minimum recruitment size is the minimum amount of biomass !
      ! in kgC/m is available for new recruits.  For the time being we use the near-bare  !
      ! ground state value as the minimum recruitment size, but this may change depending  !
      ! on how well it goes.                                                               !
      !------------------------------------------------------------------------------------!
      min_recruit_size(ipft) = min_plant_dens * (bdead_min + balive_min)
      !------------------------------------------------------------------------------------!



      !------------------------------------------------------------------------------------!
      !    Minimum size (measured as biomass of living and structural tissues) allowed in  !
      ! a cohort.  Cohorts with less biomass than this are going to be terminated.         !
      !------------------------------------------------------------------------------------! 
      min_cohort_size(ipft)  = 0.1 * min_recruit_size(ipft)
      !------------------------------------------------------------------------------------! 



      !------------------------------------------------------------------------------------! 
      !    The following variable is the absolute minimum cohort population that a cohort  !
      ! can have.  This should be used only to avoid nplant=0, but IMPORTANT: this will    !
      ! lead to a ridiculously small cohort almost guaranteed to be extinct and SHOULD BE  !
      ! USED ONLY IF THE AIM IS TO ELIMINATE THE COHORT.                                   !
      !------------------------------------------------------------------------------------! 
      negligible_nplant(ipft) = onesixth * min_cohort_size(ipft) / (bdead_max + balive_max)
      !------------------------------------------------------------------------------------! 


      !----- Find the recruit carbon to nitrogen ratio. -----------------------------------!
      c2n_recruit(ipft)      = (balive_min + bdead_min)                                    &
                             / (balive_min * ( f_labile(ipft) / c2n_leaf(ipft)             &
                             + (1.0 - f_labile(ipft)) / c2n_stem(ipft))                    &
                             + bdead_min/c2n_stem(ipft))
      !------------------------------------------------------------------------------------! 



      !------------------------------------------------------------------------------------!
      !     The minimum LAI is the LAI of a plant at the minimum cohort size that is at    !
      ! approaching the minimum elongation factor that supports leaves.                                !
      !------------------------------------------------------------------------------------!
      lai_min(ipft) = min(1.e-4,min_plant_dens * sla(ipft) * bleaf_min * (5. * elongf_min))
      !------------------------------------------------------------------------------------!


      if (print_zero_table) then
         write (unit=61,fmt='(i5,1x,a16,1x,16(es12.5,1x))')                                &
                                                     ipft,pft_name16(ipft),hgt_min(ipft)   &
                                                    ,dbh,bleaf_min,bdead_min,balive_min    &
                                                    ,bleaf_max,bdead_max,balive_max        &
                                                    ,init_density(ipft)                    &
                                                    ,min_recruit_size(ipft)                &
                                                    ,min_cohort_size(ipft)                 &
                                                    ,negligible_nplant(ipft)               &
                                                    ,sla(ipft),lai_min(ipft)               &
                                                    ,hgt_max(ipft),max_dbh(ipft)
      end if
      !------------------------------------------------------------------------------------!
   end do

   if (print_zero_table) then
      close (unit=61,status='keep')
   end if

   return
end subroutine init_pft_derived_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_disturb_params

   use disturb_coms , only : sm_fire                  & ! intent(in)
                           , min_new_patch_area       & ! intent(out)
                           , treefall_hite_threshold  & ! intent(out)
                           , forestry_on              & ! intent(out)
                           , agriculture_on           & ! intent(out)
                           , plantation_year          & ! intent(out)
                           , plantation_rotation      & ! intent(out)
                           , mature_harvest_age       & ! intent(out)
                           , fire_dryness_threshold   & ! intent(out)
                           , fire_smoist_threshold    & ! intent(out)
                           , fire_smoist_depth        & ! intent(out)
                           , k_fire_first,fire_year   & ! intent(out)
                           , fire_parameter           & ! intent(out)
                           , min_plantation_frac      & ! intent(out)
                           , max_plantation_dist      ! ! intent(out)
   use consts_coms  , only : erad                     & ! intent(in)
                           , pio180                   ! ! intent(in)
   implicit none
   
   !----- Minimum area that a patch must have to be created. ------------------------------!
!   min_new_patch_area = 0.5
   min_new_patch_area = 0.005

   !----- Only trees above this height create a gap when they fall. -----------------------!
   treefall_hite_threshold = 5.0 
   !treefall_hite_threshold = 5.0

   !----- Set to 1 if to do forest harvesting. --------------------------------------------!
   forestry_on = 0

   !----- Set to 1 if to do agriculture. --------------------------------------------------!
   agriculture_on = 0

   !----- Earliest year at which plantations occur. ---------------------------------------!
   plantation_year = 1960 

   !----- Number of years that a plantation requires to reach maturity. -------------------!
   plantation_rotation = 25.0

   !----- Years that a non-plantation patch requires to reach maturity. -------------------!
   mature_harvest_age = 50.0 
   
   !---------------------------------------------------------------------------------------!
   !     If include_fire is 1, then fire may occur if total (ground + underground) water   !
   ! converted to meters falls below this threshold.                                       !
   !---------------------------------------------------------------------------------------!
   fire_dryness_threshold = 0.2

   !---------------------------------------------------------------------------------------!
   !     If include_fire is 2, then fire may occur if total (ground + underground) water   !
   ! falls below a threshold defined by the total water of a soil column with average soil !
   ! moisture equal to soilcp + (slmsts-soilcp) * fire_smoist_threshold [m3_H2O/m3_gnd]    !
   ! would have.                                                                           !
   !---------------------------------------------------------------------------------------!
   fire_smoist_threshold = sm_fire

   !----- Maximum depth that will be considered in the average soil -----------------------!
   fire_smoist_depth     = -1.0

   !----- Dimensionless parameter controlling speed of fire spread. -----------------------!
   fire_parameter = 1.0

   !----- Minimum plantation fraction to consider the site a plantation. ------------------!
   min_plantation_frac = 0.125

   !--------------------------------------------------------------------------------------
   !    Maximum distance to the current polygon that we still consider the file grid point !
   ! to be representative of the polygon.  The value below is 1.25 degree at the Equator.  !
   !---------------------------------------------------------------------------------------!
   max_plantation_dist = 1.25 * erad * pio180
   !---------------------------------------------------------------------------------------!

   fire_year = .false.

   return

end subroutine init_disturb_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!    This subroutine initialises various PFT-independent variables for the leaf physiology !
! model.  Some useful references for this sub-routine:                                     !
!                                                                                          !
! - M09 - Medvigy, D.M., S. C. Wofsy, J. W. Munger, D. Y. Hollinger, P. R. Moorcroft,      !
!         2009: Mechanistic scaling of ecosystem function and dynamics in space and time:  !
!         Ecosystem Demography model version 2.  J. Geophys. Res., 114, G01002,            !
!         doi:10.1029/2008JG000812.                                                        !
! - M06 - Medvigy, D.M., 2006: The State of the Regional Carbon Cycle: results from a      !
!         constrained coupled ecosystem-atmosphere model, 2006.  Ph.D. dissertation,       !
!         Harvard University, Cambridge, MA, 322pp.                                        !
! - M01 - Moorcroft, P. R., G. C. Hurtt, S. W. Pacala, 2001: A method for scaling          !
!         vegetation dynamics: the ecosystem demography model, Ecological Monographs, 71,  !
!         557-586.                                                                         !
! - F96 - Foley, J. A., I. Colin Prentice, N. Ramankutty, S. Levis, D. Pollard, S. Sitch,  !
!         A. Haxeltime, 1996: An integrated biosphere model of land surface processes,     !
!         terrestrial carbon balance, and vegetation dynamics. Glob. Biogeochem. Cycles,   !
!         10, 603-602.                                                                     !
! - L95 - Leuning, R., F. M. Kelliher, D. G. G. de Pury, E. D. Schulze, 1995: Leaf         !
!         nitrogen, photosynthesis, conductance, and transpiration: scaling from leaves to !
!         canopies. Plant, Cell, and Environ., 18, 1183-1200.                              !
! - F80 - Farquhar, G. D., S. von Caemmerer, J. A. Berry, 1980: A biochemical model of     !
!         photosynthetic  CO2 assimilation in leaves of C3 species. Planta, 149, 78-90.    !
! - C91 - Collatz, G. J., J. T. Ball, C. Grivet, J. A. Berry, 1991: Physiology and         !
!         environmental regulation of stomatal conductance, photosynthesis and transpir-   !
!         ation: a model that includes a laminar boundary layer, Agric. and Forest         !
!         Meteorol., 54, 107-136.                                                          !
! - C92 - Collatz, G. J., M. Ribas-Carbo, J. A. Berry, 1992: Coupled photosynthesis-       !
!         stomatal conductance model for leaves of C4 plants.  Aust. J. Plant Physiol.,    !
!         19, 519-538.                                                                     !
! - E78 - Ehleringer, J. R., 1978: Implications of quantum yield differences on the        !
!         distributions of C3 and C4 grasses.  Oecologia, 31, 255-267.                     !
!                                                                                          !
!------------------------------------------------------------------------------------------!
subroutine init_physiology_params()
   use physiology_coms, only : iphysiol            & ! intent(out)
                             , c34smin_lint_co2    & ! intent(out)
                             , c34smax_lint_co2    & ! intent(out)
                             , c34smax_gsw         & ! intent(out)
                             , gbh_2_gbw           & ! intent(out)
                             , gbw_2_gbc           & ! intent(out)
                             , gsw_2_gsc           & ! intent(out)
                             , gsc_2_gsw           & ! intent(out)
                             , kookc               & ! intent(out)
                             , tphysref            & ! intent(out)
                             , tphysrefi           & ! intent(out)
                             , fcoll               & ! intent(out)
                             , compp_refval        & ! intent(out)
                             , compp_hor           & ! intent(out)
                             , compp_q10           & ! intent(out)
                             , kco2_refval         & ! intent(out)
                             , kco2_hor            & ! intent(out)
                             , kco2_q10            & ! intent(out)
                             , ko2_refval          & ! intent(out)
                             , ko2_hor             & ! intent(out)
                             , ko2_q10             & ! intent(out)
                             , klowco2             & ! intent(out)
                             , o2_ref              & ! intent(out)
                             , par_twilight_min    & ! intent(out)
                             , qyield0             & ! intent(out)
                             , qyield1             & ! intent(out)
                             , qyield2             & ! intent(out)
                             , ehleringer_alpha0c  & ! intent(out)
                             , c34smin_lint_co28   & ! intent(out)
                             , c34smax_lint_co28   & ! intent(out)
                             , c34smax_gsw8        & ! intent(out)
                             , gbh_2_gbw8          & ! intent(out)
                             , gbw_2_gbc8          & ! intent(out)
                             , gsw_2_gsc8          & ! intent(out)
                             , gsc_2_gsw8          & ! intent(out)
                             , kookc8              & ! intent(out)
                             , tphysref8           & ! intent(out)
                             , tphysrefi8          & ! intent(out)
                             , fcoll8              & ! intent(out)
                             , compp_refval8       & ! intent(out)
                             , compp_hor8          & ! intent(out)
                             , compp_q108          & ! intent(out)
                             , kco2_refval8        & ! intent(out)
                             , kco2_hor8           & ! intent(out)
                             , kco2_q108           & ! intent(out)
                             , ko2_refval8         & ! intent(out)
                             , ko2_hor8            & ! intent(out)
                             , ko2_q108            & ! intent(out)
                             , klowco28            & ! intent(out)
                             , par_twilight_min8   & ! intent(out)
                             , o2_ref8             & ! intent(out)
                             , qyield08            & ! intent(out)
                             , qyield18            & ! intent(out)
                             , qyield28            & ! intent(out)
                             , ehleringer_alpha0c8 & ! intent(out)
                             , print_photo_debug   & ! intent(out)
                             , photo_prefix        ! ! intent(out)
   use consts_coms    , only : umol_2_mol          & ! intent(in)
                             , t00                 & ! intent(in)
                             , mmdoc               & ! intent(in)
                             , mmcod               & ! intent(in)
                             , mmo2                & ! intent(in)
                             , mmdryi              & ! intent(in)
                             , Watts_2_Ein         & ! intent(in)
                             , prefsea             ! ! intent(in)
   implicit none
   !------ Local constants. ---------------------------------------------------------------!
   real, parameter :: ehl0           =  8.10e-2 ! Intercept                    (E78, eqn 2)
   real, parameter :: ehl1           = -5.30e-5 ! Linear coefficient           (E78, eqn 2)
   real, parameter :: ehl2           = -1.90e-5 ! Quadratic coefficient        (E78, eqn 2)
   real, parameter :: tau_refval_f96 =  4500.   ! Reference tau                (F96)
   real, parameter :: tau_refval_c91 =  2600.   ! Reference tau                (C91)
   real, parameter :: tau_hor        = -5000.   ! "Activation energy" for tau  (F96)
   real, parameter :: tau_q10        =  0.57    ! "Q10" term for tau           (C91)  
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     Bounds for internal carbon and water stomatal conductance.                        !
   !---------------------------------------------------------------------------------------!
   c34smin_lint_co2 = 0.5   * umol_2_mol ! Minimum carbon dioxide concentration [  mol/mol]
   c34smax_lint_co2 = 1200. * umol_2_mol ! Maximum carbon dioxide concentration [  mol/mol]
   c34smax_gsw      = 1.e+2              ! Max. stomatal conductance (water)    [ mol/m/s]
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     Many parameters in the model are temperature-dependent and we must provide a      !
   ! reference value and shape parameters.  Here we define this reference temperature.     !
   ! IMPORTANT: Some schemes use 15 C (F96-based in particular), whilst others use 25 C    !
   ! (C91-based in particular).  Be sure to make the necessary conversions of these        !
   ! references otherwise you may have surprises.                                          !
   !---------------------------------------------------------------------------------------!
   tphysref     = 15.0+t00
   tphysrefi    = 1./tphysref
   fcoll        = 0.10
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     The following parameter is the concentration of oxygen, assumed constant through- !
   ! out the integration.  The value is in mol/mol.                                        !
   !---------------------------------------------------------------------------------------!
   o2_ref       = 0.209
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     This parameter is from F80, and is the ratio between the turnover number for the  !
   ! oxylase function and the turnover number for the carboxylase function.                !
   !---------------------------------------------------------------------------------------!
   kookc        = 0.21
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !      Define the Michaelis-Mentel constants for CO2 and O2, and the CO2 compensation   !
   ! point without accounting leaf respiration (Gamma*).  Here we define the reference     !
   ! value and the compensation point parameters in a slightly different way depending on  !
   ! whether to use F96-based or C91-based physiology, and whether to use the              !
   ! pre-determined parameters or functions of the Michaelis-Mentel constant parameters.   !
   !---------------------------------------------------------------------------------------!
   !----- Common parameters. --------------------------------------------------------------!
   kco2_hor = 6000.  ! "Activation energy" for CO2 Michaelis-Mentel constant     [       K]
   kco2_q10 = 2.1    ! Q10 term for CO2 Michaelis-Mentel constant                [     ---]
   ko2_hor  = 1400.  ! "Activation energy" for O2 Michaelis-Mentel constants     [       K]
   ko2_q10  = 1.2    ! Q10 term for O2 Michaelis-Mentel constant                 [     ---]
   select case (iphysiol)
   case (0)
      !----- Use default CO2 and O2 reference values from F96. ----------------------------!
      kco2_refval   = 150. * umol_2_mol ! Reference Michaelis-Mentel CO2 coeff.  [ mol/mol]
      ko2_refval    = 0.250             ! Reference Michaelis-Mentel O2 coeff.   [ mol/mol]
      !------------------------------------------------------------------------------------!


      !------------------------------------------------------------------------------------!
      !   Use the default CO2 compensation point values from F96/M01, based on tau.        !
      !   Gamma* = [O2] / (2. * tau)                                                       !
      !------------------------------------------------------------------------------------!
      compp_refval  = o2_ref / (2. * tau_refval_f96) ! Reference value           [ mol/mol]
      compp_hor     = -tau_hor                       ! "Activation energy"       [       K]
      compp_q10     = 1. / tau_q10                   ! "Q10" term                [     ---]
      !------------------------------------------------------------------------------------!

   case (1)
      !----- Use default CO2 and O2 reference values from F96. ----------------------------!
      kco2_refval   = 150. * umol_2_mol ! Reference Michaelis-Mentel CO2 coeff.  [ mol/mol]
      ko2_refval    = 0.250             ! Reference Michaelis-Mentel O2 coeff.   [ mol/mol]
      !------------------------------------------------------------------------------------!


      !------------------------------------------------------------------------------------!
      !   Find the CO2 compensation point values using an approach similar to F80 and CLM. !
      !   Gamma* = (ko/kc) * [O2] * KCO2 / (2. * KO2)                                                       !
      !------------------------------------------------------------------------------------!
      compp_refval  = kookc * o2_ref * kco2_refval / ko2_refval ! Ref. value     [ mol/mol]
      compp_hor     = kco2_hor - ko2_hor                        ! "Act. energy"  [       K]
      compp_q10     = kco2_q10 / ko2_q10                        ! "Q10" term     [     ---]
      !------------------------------------------------------------------------------------!

   case (2)
      !------------------------------------------------------------------------------------!
      !    Use default CO2 and O2 reference values from C91.  Here we must convert the     !
      ! reference values from Pa to mol/mol, and the reference must be set to 15 C.        !
      !------------------------------------------------------------------------------------!
      !------ Reference Michaelis-Mentel CO2 coefficient [mol/mol]. -----------------------! 
      kco2_refval   = 30. * mmcod / prefsea / kco2_q10
      !------ Reference Michaelis-Mentel O2 coefficient [mol/mol]. ------------------------! 
      ko2_refval    = 3.e4 * mmo2 * mmdryi / prefsea / ko2_q10
      !------------------------------------------------------------------------------------!



      !------------------------------------------------------------------------------------!
      !   Use the default CO2 compensation point values from C91, based on tau.            !
      !   Gamma* = [O2] / (2. * tau)                                                       !
      !------------------------------------------------------------------------------------!
      compp_refval  = o2_ref * tau_q10 / (2. * tau_refval_c91) ! Reference value [ mol/mol]
      compp_hor     = - tau_hor                                ! "Activation en. [       K]
      compp_q10     = 1. / tau_q10                             ! "Q10" term      [     ---]
      !------------------------------------------------------------------------------------!
   case (3)
      !------------------------------------------------------------------------------------!
      !    Use default CO2 and O2 reference values from C91.  Here we must convert the     !
      ! reference values from Pa to mol/mol, and the reference must be set to 15 C.        !
      !------------------------------------------------------------------------------------!
      !------ Reference Michaelis-Mentel CO2 coefficient [mol/mol]. -----------------------! 
      kco2_refval   = 30. * mmcod / prefsea / kco2_q10
      !------ Reference Michaelis-Mentel O2 coefficient [mol/mol]. ------------------------! 
      ko2_refval    = 3.e4 * mmo2 * mmdryi / prefsea / ko2_q10
      !------------------------------------------------------------------------------------!



      !------------------------------------------------------------------------------------!
      !   Use the default CO2 compensation point values from C91, based on tau.            !
      !   Gamma* = [O2] / (2. * tau)                                                       !
      !------------------------------------------------------------------------------------!
      compp_refval  = kookc * o2_ref * kco2_refval / ko2_refval ! Ref. value     [ mol/mol]
      compp_hor     = kco2_hor - ko2_hor                        ! "Act. energy"  [       K]
      compp_q10     = kco2_q10 / ko2_q10                        ! "Q10" term     [     ---]
      !------------------------------------------------------------------------------------!
   end select
   !---------------------------------------------------------------------------------------! 



   !---------------------------------------------------------------------------------------!
   !    The following parameter is the k coefficient in Foley et al. (1996) that is used   !
   ! to determine the CO2-limited photosynthesis for C4 grasses.  Notice that Foley et al. !
   ! (1996) didn't correct for molar mass (the mmdoc term here).                           !
   !---------------------------------------------------------------------------------------!
   klowco2      = 18000. * mmdoc ! coefficient for low CO2                       [ mol/mol]
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     These are constants obtained in Leuning et al. (1995) and Collatz et al. (1991)   !
   ! to convert different conductivities.                                                  !
   !---------------------------------------------------------------------------------------!
   gbh_2_gbw  = 1.075           ! heat  to water  - leaf boundary layer
   gbw_2_gbc  = 1.0 / 1.4       ! water to carbon - leaf boundary layer
   gsw_2_gsc  = 1.0 / 1.6       ! water to carbon - stomata
   gsc_2_gsw  = 1./gsw_2_gsc    ! carbon to water - stomata
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     This is the minimum threshold for the photosynthetically active radiation, in     !
   ! mol/m/s to consider non-night time conditions (day time or twilight).               !
   !---------------------------------------------------------------------------------------!
   par_twilight_min = 0.5 * Watts_2_Ein ! Minimum non-nocturnal PAR.             [mol/m/s]
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     This is the quantum yield response curve, using equation 2 from E78:              !
   !     The coefficients here are different because his function used temperature in      !
   ! Celsius and here we give the coefficients for temperature in Kelvin.  We don't use    !
   ! this equation for temperatures below 0C, instead we find the equivalent at 0 Celsius  !
   ! and assume it is constant below this temperature.                                     !
   !---------------------------------------------------------------------------------------!
   qyield0            = ehl0 - ehl1 * t00 + ehl2 * t00 * t00
   qyield1            = ehl1 - 2. * ehl2 * t00
   qyield2            = ehl2
   ehleringer_alpha0c = ehl0
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !    Find the double precision version of the variables above.                          !
   !---------------------------------------------------------------------------------------!
   c34smin_lint_co28   = dble(c34smin_lint_co2  )
   c34smax_lint_co28   = dble(c34smax_lint_co2  )
   c34smax_gsw8        = dble(c34smax_gsw       )
   gbh_2_gbw8          = dble(gbh_2_gbw         )
   gbw_2_gbc8          = dble(gbw_2_gbc         )
   gsw_2_gsc8          = dble(gsw_2_gsc         )
   gsc_2_gsw8          = dble(gsc_2_gsw         )
   kookc8              = dble(kookc             )
   tphysref8           = dble(tphysref          )
   tphysrefi8          = dble(tphysrefi         )
   fcoll8              = dble(fcoll             )
   compp_refval8       = dble(compp_refval      )
   compp_hor8          = dble(compp_hor         )
   compp_q108          = dble(compp_q10         )
   kco2_refval8        = dble(kco2_refval       )
   kco2_hor8           = dble(kco2_hor          )
   kco2_q108           = dble(kco2_q10          )
   ko2_refval8         = dble(ko2_refval        )
   ko2_hor8            = dble(ko2_hor           )
   ko2_q108            = dble(ko2_q10           )
   klowco28            = dble(klowco2           )
   o2_ref8             = dble(o2_ref            )
   par_twilight_min8   = dble(par_twilight_min  )
   qyield08            = dble(qyield0           )
   qyield18            = dble(qyield1           )
   qyield28            = dble(qyield2           )
   ehleringer_alpha0c8 = dble(ehleringer_alpha0c)
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     Parameters that control debugging output.                                         !
   !---------------------------------------------------------------------------------------!
   !----- I should print detailed debug information. --------------------------------------!
   print_photo_debug = .false.
   !----- File name prefix for the detailed information in case of debugging. -------------!
   photo_prefix      = 'photo_state_'
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_physiology_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_hydro_coms

  use hydrology_coms,only:useTOPMODEL,useRUNOFF,HydroOutputPeriod, &
       MoistRateTuning,MoistSatThresh,Moist_dWT,FracLiqRunoff, &
       GrassLAIMax,inverse_runoff_time
  use ed_misc_coms, only: ied_init_mode

  implicit none

  if(ied_init_mode == 3)then
     ! Signifies a restart from an ED2 history file
     useTOPMODEL = 1
     useRUNOFF   = 0
  else
     useTOPMODEL = 0
     useRUNOFF = 0
  endif

  HydroOutputPeriod = 96 !! multiples of dtlsm

  MoistRateTuning = 1.0    

  MoistSatThresh = 0.95     
  
  Moist_dWT = 2.0 

  FracLiqRunoff = 0.5
  
  GrassLAImax = 4.0

  inverse_runoff_time = 0.1

  return
end subroutine init_hydro_coms
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_soil_coms
   use soil_coms      , only : ed_nstyp              & ! intent(in)
                             , isoilflg              & ! intent(in)
                             , nslcon                & ! intent(in)
                             , slxclay               & ! intent(in)
                             , slxsand               & ! intent(in)
                             , soil                  & ! intent(in)
                             , betapower             & ! intent(in)
                             , soil_class            & ! type
                             , soil8                 & ! intent(out)
                             , water_stab_thresh     & ! intent(out)
                             , snowmin               & ! intent(out)
                             , dewmax                & ! intent(out)
                             , soil_rough            & ! intent(out)
                             , snow_rough            & ! intent(out)
                             , tiny_sfcwater_mass    & ! intent(out)
                             , infiltration_method   & ! intent(out)
                             , soil_rough8           & ! intent(out)
                             , snow_rough8           & ! intent(out)
                             , betapower8            & ! intent(out)
                             , freezecoef            & ! intent(out)
                             , freezecoef8           ! ! intent(out)

   use grid_coms      , only : ngrids                ! ! intent(in)
   use consts_coms    , only : grav                  & ! intent(in)
                             , hr_sec                & ! intent(in)
                             , day_sec               ! ! intent(in)

   implicit none
   !----- Local variables. ----------------------------------------------------------------!
   integer            :: nsoil
   integer            :: ifm
   !----- Local constants. ----------------------------------------------------------------!
   real   , parameter :: fieldcp_K  = 0.1     ! hydraulic conduct. at field cap.   [mm/day]
   real   , parameter :: soilcp_MPa = 4.0     ! soil-water pot. for air dry soil   [   MPa]
   real   , parameter :: soilwp_MPa = 1.5     ! soil-water pot. at wil. point      [   MPa]
   real   , parameter :: sand_hcapv = 2.128e6 ! Sand vol. heat capacity            [J/m3/K]
   real   , parameter :: clay_hcapv = 2.385e6 ! Clay vol. heat capacity            [J/m3/K]
   real   , parameter :: silt_hcapv = 2.286e6 ! Silt vol. heat capacity (*)        [J/m3/K]
   real   , parameter :: air_hcapv  = 1.212e6 ! Air vol. heat capacity             [J/m3/K]
   !---------------------------------------------------------------------------------------!
   ! (*) If anyone has the heat capacity for silt, please feel free to add it in here, I   !
   !     didn't find any.  Apparently no one knows, and I've seen in other models that     !
   !     people just assume either the same as sand or the average.  Here I'm just using   !
   !     halfway.  I think the most important thing is to take into account the soil and   !
   !     the air, which are the most different.                                            !
   !---------------------------------------------------------------------------------------!


   !----- Initialise some standard variables. ---------------------------------------------!
   water_stab_thresh   = 3.0    ! Minimum water mass to be considered stable     [   kg/m2]
   snowmin             = 3.0    ! Minimum snow mass needed to create a new layer [   kg/m2]
   dewmax              = 3.0e-5 ! Maximum dew flux rate (deprecated)             [ kg/m2/s]
   soil_rough          = 0.05   ! Soil roughness height                          [       m]
   snow_rough          = 0.001  ! Snowcover roughness height                     [       m]
   tiny_sfcwater_mass  = 1.0e-3 ! Minimum allowed mass in temporary layers       [   kg/m2]
   infiltration_method = 0      ! Infiltration method, used in rk4_derivs        [     0|1]
   freezecoef          = 7.0    ! Coeff. for infiltration of frozen water        [     ---]
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !      UPDATED based on Cosby et al. 1984, soil clay and sand fractions from table 2,   !
   ! slpots, slmsts, slbs, and slcons calculated based on table 4 sfldcap calculated based !
   ! on Clapp and Hornberger equation 2 and a soil hydraulic conductivity of 0.1mm/day,    !
   ! soilcp (air dry soil moisture capacity) is calculated based on Cosby et al. 1984      !
   ! equation 1 and a soil-water potential of -3.1MPa.  soilwp (the wilting point soil     !
   ! moisture) is calculated based on Cosby et al 1985 equation 1 and a soil-water poten-  !
   ! tial of -1.5MPa (Equation 1 in Cosby is not correct it should be saturated moisture   !
   ! potential over moisture potential)  (NML, 2/2010)                                     !
   !---------------------------------------------------------------------------------------!

   !---------------------------------------------------------------------------------------!
   ! (1st line)          slpots        slmsts          slbs     slcpd        soilcp        !
   ! (2nd line)          soilwp        slcons       slcons0 soilcond0     soilcond1        !
   ! (3rd line)       soilcond2       sfldcap        albwet    albdry         xsand        !
   ! (4th line)           xclay         xsilt       xrobulk     slden                      !
   !---------------------------------------------------------------------------------------!
   soil = (/                                                                               &
      !----- 1. Sand. ---------------------------------------------------------------------!
       soil_class( -0.121,     0.395,     4.05, 1465000.,  0.026183447        &
                 ,  0.032636,  .18e-3,  0.000500000,   0.3000,       4.8000        &
                 ,      -2.7000,  0.135,        0.229,    0.352,        0.97        &
                 ,        0.030,        0.050,        1200.,    1600.              )       &
      !----- 2. Loamy sand. ---------------------------------------------------------------!
      ,soil_class( -0.090,     0.410,     4.38, 1407000.,  0.041560499        &
                 ,  0.050323046,  .16e-3,  0.000600000,   0.3000,       4.6600        &
                 ,      -2.6000,  0.150,        0.212,    0.335,        0.92        &
                 ,        0.07,        0.115,        1250.,    1600.              )       &
      !----- 3. Sandy loam. ---------------------------------------------------------------!
      ,soil_class( -0.218,     0.435,     4.9, 1344000.,  0.073495043        &
                 ,  0.085973722,  .34e-4,  0.00077000,   0.2900,       4.2700        &
                 ,      -2.3100,  0.195,        0.183,    0.307,        0.80        &
                 ,        0.18,        0.230,        1300.,    1600.              )       &
      !----- 4. Silt loam. ----------------------------------------------------------------!
      ,soil_class( -0.786,     0.485,     5.3, 1273000.,  0.150665475        &
                 ,  0.171711257,  .72e-5,  0.00001100,   0.2700,       3.4700        &
                 ,      -1.7400,  0.255,        0.107,    0.250,        0.57        &
                 ,        0.40,        0.640,        1400.,    1600.              )       &
      !----- 5. Loam. ---------------------------------------------------------------------!
      ,soil_class( -0.478,     0.451,     5.39, 1214000.,  0.125192234        &
                 ,  0.142369513,  .69e-5,  0.002200000,   0.2800,       3.6300        &
                 ,      -1.8500,  0.240,        0.140,    0.268,        0.60        &
                 ,        0.35,        0.420,        1350.,    1600.              )       &
      !----- 6. Sandy clay loam. ----------------------------------------------------------!
      ,soil_class( -0.299,     0.420,     7.12, 1177000.,  0.136417267        &
                 ,  0.150969505,  .63e-5,  0.001500000,   0.2800,       3.7800        &
                 ,      -1.9600,  0.255,        0.163,    0.260,        0.65        &
                 ,        0.31,        0.140,        1350.,    1600.              )       &
      !----- 7. Silty clay loam. ----------------------------------------------------------!
      ,soil_class( -0.356,     0.477,     7.75, 1319000.,  0.228171947        &
                 ,  0.248747504,  .17e-5,  0.00011000,   0.2600,       2.7300        &
                 ,      -1.2000,  0.322,        0.081,    0.195,        0.35        &
                 ,        0.59,        0.560,        1500.,    1600.              )       &
      !----- 8. Clayey loam. --------------------------------------------------------------!
      ,soil_class( -0.630,     0.476,     8.52, 1227000.,  0.192624431        &
                 ,  0.210137962,  .24e-5,  0.002200000,   0.2700,       3.2300        &
                 ,      -1.5600,  0.325,        0.116,    0.216,        0.48        &
                 ,        0.45,        0.340,        1450.,    1600.              )       &
      !----- 9. Sandy clay. ---------------------------------------------------------------!
      ,soil_class( -0.121283019,     0.415620,     9.538000, 1673422.,  0.182198910        &
                 ,  0.196607427,  4.314507e-6,  0.000002167,   0.2700,       3.3200        &
                 ,      -1.6300,  0.286363001,        0.144,    0.216,        0.520        &
                 ,        0.420,        0.060,        1450.,    1600.              )       &
      !----- 10. Silty clay. --------------------------------------------------------------!
      ,soil_class( -0.601312179,     0.479090,    10.461000, 1652723.,  0.263228486        &
                 ,  0.282143846,  1.055191e-6,  0.000001033,   0.2500,       2.5800        &
                 ,      -1.0900,  0.360319788,        0.068,    0.159,        0.060        &
                 ,        0.470,        0.470,        1650.,    1600.              )       &
      !----- 11. Clay. --------------------------------------------------------------------!
      ,soil_class( -0.299226464,     0.454400,    12.460000, 1692037.,  0.259868987        &
                 ,  0.275459057,  1.307770e-6,  0.000001283,   0.2500,       2.4000        &
                 ,      -0.9600,  0.353255209,        0.083,    0.140,        0.200        &
                 ,        0.600,        0.200,        1700.,    1600.              )       &
      !----- 12. Peat. --------------------------------------------------------------------!
      ,soil_class( -0.534564359,     0.469200,     6.180000,  874000.,  0.167047523        &
                 ,  0.187868805,  2.357930e-6,  0.000008000,   0.0600,       0.4600        &
                 ,       0.0000,  0.285709966,        0.070,    0.140,       0.2000        &
                 ,       0.2000,       0.6000,         500.,     300.              )       &
      !----- 13. Bedrock. -----------------------------------------------------------------!
      ,soil_class(    0.0000000,     0.000000,     0.000000, 2130000.,  0.000000000        &
                 ,  0.000000000,  0.000000e+0,  0.000000000,   4.6000,       0.0000        &
                 ,       0.0000,  0.000000001,        0.320,    0.320,       0.0000        &
                 ,       0.0000,       0.0000,           0.,       0.              )       &
      !----- 14. Silt. --------------------------------------------------------------------!
      ,soil_class( -1.047128548,     0.492500,     3.862500, 1510052.,  0.112299080        &
                 ,  0.135518820,  2.046592e-6,  0.000010600,   0.2700,       3.4700        &
                 ,      -1.7400,  0.245247642,        0.092,    0.265,        0.075        &
                 ,        0.050,        0.875,        1400.,    1600.              )       &
      !----- 15. Heavy clay. --------------------------------------------------------------!
      ,soil_class( -0.322106879,     0.461200,    15.630000, 1723619.,  0.296806035        &
                 ,  0.310916364,  7.286705e-7,  0.000001283,   0.2500,       2.4000        &
                 ,      -0.9600,  0.382110712,        0.056,    0.080,        0.100        &
                 ,        0.800,        0.100,        1700.,    1600.              )       &
      !----- 16. Clayey sand. -------------------------------------------------------------!
      ,soil_class( -0.176502150,     0.432325,    11.230000, 1688353.,  0.221886929        &
                 ,  0.236704039,  2.426785e-6,  0.000001283,   0.2500,       2.4000        &
                 ,      -0.9600,  0.320146708,        0.115,    0.175,        0.375        &
                 ,        0.525,        0.100,        1700.,    1600.              )       &
      !----- 17. Clayey silt. -------------------------------------------------------------!
      ,soil_class( -0.438278332,     0.467825,    11.305000, 1670103.,  0.261376708        &
                 ,  0.278711303,  1.174982e-6,  0.000001283,   0.2500,       2.4000        &
                 ,      -0.9600,  0.357014719,        0.075,    0.151,        0.125        &
                 ,        0.525,        0.350,        1700.,    1600.              )       &
   /)
   !---------------------------------------------------------------------------------------!

   do nsoil = 1, 8
      soil(nsoil)%soilcp = 0.1 - 0.07 * soil(nsoil)%xsand
      soil(nsoil)%xsilt = 1.0 - soil(nsoil)%xsand - soil(nsoil)%xclay
   enddo

   !---------------------------------------------------------------------------------------!
   !******** Correct soil_class table using sand and clay fractions (if provided)  ********!
   !      Based on Cosby et al 1984, using table 4 and equation 1 (which is incorrect it   !
   ! should be saturated moisture potential over moisture potential).  NML 2/2010          !
   !---------------------------------------------------------------------------------------!
   do ifm=1,ngrids
      if ( isoilflg(ifm)==2 .and. slxclay > 0. .and. slxsand > 0. .and.                    &
           (slxclay + slxsand) <= 1. ) then
         soil(nslcon)%xsand   = slxsand
         soil(nslcon)%xclay   = slxclay
         soil(nslcon)%xsilt   = 1. - slxsand - slxclay

         !----- B exponent [unitless]. ----------------------------------------------------!
         soil(nslcon)%slbs    = 3.10 + 15.7*slxclay - 0.3*slxsand

         !----- Soil moisture potential at saturation [ m ]. ------------------------------!
         soil(nslcon)%slpots  = -1. * (10.**(2.17 - 0.63*slxclay - 1.58*slxsand)) * 0.01

         !----- Hydraulic conductivity at saturation [ m/s ]. -----------------------------!
         soil(nslcon)%slcons  = (10.**(-0.60 + 1.26*slxsand - 0.64*slxclay))               &
                              * 0.0254/hr_sec
         !----- Hydraulic conductivity at saturation at top [ m/s ], for TOPMODEL style. --!
         
         !----- Soil moisture at saturation [ m^3/m^3 ]. ----------------------------------!
         soil(nslcon)%slmsts  = (50.5 - 14.2*slxsand - 3.7*slxclay) / 100.
         !----- Soil field capacity[ m^3/m^3 ]. -------------------------------------------!
         soil(nslcon)%sfldcap = soil(nslcon)%slmsts                                        &
                              *  ( (fieldcp_K/1000./day_sec)/soil(nslcon)%slcons)          &
                              ** (1. / (2.*soil(nslcon)%slbs+3.))
         !----- Dry soil capacity (at -3.1MPa) [ m^3/m^3 ]. -------------------------------!
         soil(nslcon)%soilcp  = soil(nslcon)%slmsts                                        &
                              *  ( -1.*soil(nslcon)%slpots / (soilcp_MPa * 1000. / grav))  &
                              ** (1. / soil(nslcon)%slbs)
         !----- Wilting point capacity (at -1.5MPa) [ m^3/m^3 ]. --------------------------!
         soil(nslcon)%soilwp  = soil(nslcon)%slmsts                                        &
                              *  ( -1.*soil(nslcon)%slpots / (soilwp_MPa * 1000. / grav))  &
                              ** ( 1. / soil(nslcon)%slbs)
         !---------------------------------------------------------------------------------!

         !---------------------------------------------------------------------------------!
         !     Heat capacity.  Here we take the volume average amongst silt, clay, and     !
         ! sand, and consider the contribution of air sitting in.  In order to keep it     !
         ! simple, we assume that the air fraction won't change, although in reality its   !
         ! contribution should be a function of soil moisture.  Here we use the amount of  !
         ! air in case the soil moisture was halfway between dry air and saturated, so the !
         ! error is not too biased.                                                        !
         !---------------------------------------------------------------------------------!
         soil(nslcon)%slcpd   = (1. - soil(nslcon)%slmsts)                                 &
                              * ( soil(nslcon)%xsand * sand_hcapv                          &
                                + soil(nslcon)%xsilt * silt_hcapv                          &
                                + soil(nslcon)%xclay * clay_hcapv )                        &
                              + 0.5 * (soil(nslcon)%soilcp + soil(nslcon)%slmsts)          &
                              * air_hcapv
         !---------------------------------------------------------------------------------!



         !---------------------------------------------------------------------------------!
         !     No, I'm not happy with that, but I am just trying to get some general sense !
         ! of albedo for different texture.  In reality, soil albedo is a function of many !
         ! things besides soil texture, like organic content, iron content, just to name a !
         ! few.  Getting these kinds of information is almost impossible, though.  Here I  !
         ! built a very weak linear model using robust linear model in R (rlm, package     !
         ! MASS), using data from a few papers:                                            !
         !                                                                                 !
         ! Idso, S. B., R. D. Jackson, R. J. Reginato, B. A. Kimball, F. S. Nakayama,      !
         !      1975: The dependence of bare soil albedo on soil water content.  J. Appl.  !
         !      Meteorol., 14, 109-113.  No sand/clay/silt fraction given, but it is a     !
         !      loam soil, so I used the canonical loam fraction of sand and silt.         !
         ! Matthias, A. D., A. Fimbres, E. E. Sano, D. F. Post, L. Accioly,                !
         !      A. K . Batchily, L. G. Ferreira, 2000: Surface roughness effects on soil   !
         !      albedo.  I used the average value for smooth soil and averaged across      !
         !      zenith angles.                                                             !
         ! Ten Berge, H.F.M., 1986: Heat and water transfer at the bare soil surface.      !
         !      Aspects affecting thermal imagery. PhD. Thesis Agricultural University     !
         !      Wageningen, The Netherlands.  I didn't find the actual thesis online, but  !
         !      I found a table from the Italian Centre for Industrial Agriculture         !
         !      Research and used the standard fraction of sand and clay for their soil    !
         !      types.  http://agsys.cra-cin.it/tools/solarradiation/help/Albedo.html      !
         !---------------------------------------------------------------------------------!
         soil(nslcon)%albwet = 0.02982594 + 0.21343545 * soil(nslcon)%xsand                &
                                          + 0.05314899 * soil(nslcon)%xsilt
         soil(nslcon)%albdry = 0.02008580 + 0.34730650 * soil(nslcon)%xsand                &
                                          + 0.25034280 * soil(nslcon)%xsilt
         !---------------------------------------------------------------------------------!
      end if
   end do
   !---------------------------------------------------------------------------------------!







   !----- Here we fill soil8, which will be used in Runge-Kutta (double precision). -------!
   do nsoil=1,ed_nstyp
      soil8(nsoil)%slpots    = dble(soil(nsoil)%slpots   )
      soil8(nsoil)%slmsts    = dble(soil(nsoil)%slmsts   )
      soil8(nsoil)%slbs      = dble(soil(nsoil)%slbs     )
      soil8(nsoil)%slcpd     = dble(soil(nsoil)%slcpd    )
      soil8(nsoil)%soilcp    = dble(soil(nsoil)%soilcp   )
      soil8(nsoil)%soilwp    = dble(soil(nsoil)%soilwp   )
      soil8(nsoil)%slcons    = dble(soil(nsoil)%slcons   )
      soil8(nsoil)%slcons0   = dble(soil(nsoil)%slcons0  )
      soil8(nsoil)%soilcond0 = dble(soil(nsoil)%soilcond0)
      soil8(nsoil)%soilcond1 = dble(soil(nsoil)%soilcond1)
      soil8(nsoil)%soilcond2 = dble(soil(nsoil)%soilcond2)
      soil8(nsoil)%sfldcap   = dble(soil(nsoil)%sfldcap  )
      soil8(nsoil)%albwet    = dble(soil(nsoil)%albwet   )
      soil8(nsoil)%albdry    = dble(soil(nsoil)%albdry   )
      soil8(nsoil)%xsand     = dble(soil(nsoil)%xsand    )
      soil8(nsoil)%xclay     = dble(soil(nsoil)%xclay    )
      soil8(nsoil)%xsilt     = dble(soil(nsoil)%xsilt    )
      soil8(nsoil)%xrobulk   = dble(soil(nsoil)%xrobulk  )
      soil8(nsoil)%slden     = dble(soil(nsoil)%slden    )
   end do
   betapower8  = dble(betapower)
   soil_rough8 = dble(soil_rough)
   snow_rough8 = dble(snow_rough)
   freezecoef8 = dble(freezecoef)
   return
end subroutine init_soil_coms
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine init_phen_coms
   use consts_coms   , only : erad                     & ! intent(in)
                            , pio180                   ! ! intent(in)
   use phenology_coms, only : retained_carbon_fraction & ! intent(out)
                            , elongf_min               & ! intent(out)
                            , theta_crit               & ! intent(out)
                            , dl_tr                    & ! intent(out)
                            , st_tr1                   & ! intent(out)
                            , st_tr2                   & ! intent(out)
                            , phen_a                   & ! intent(out)
                            , phen_b                   & ! intent(out)
                            , phen_c                   & ! intent(out)
                            , rad_turnover_int         & ! intent(out)
                            , rad_turnover_slope       & ! intent(out)
                            , vm_tran                  & ! intent(out)
                            , vm_slop                  & ! intent(out)
                            , vm_amp                   & ! intent(out)
                            , vm_min                   & ! intent(out)
                            , max_phenology_dist       & ! intent(out)
                            , radint                   & ! intent(out)
                            , radslp                   & ! intent(out)
                            , thetacrit                   ! ! intent(in)
   implicit none

 
   retained_carbon_fraction = 0.5
   elongf_min               = 0.02
   theta_crit               = thetacrit
   dl_tr                    = 810.0!776.0!655.0
   st_tr1                   = 284.3
   st_tr2                   = 275.15
  
   phen_a                   = -68.0!95.9!36.4!-68.0 ATT calculated for paper birtch in fairbanks
   phen_b                   = 638.0! 639.4!320.4!638.0
   phen_c                   = -0.01!-0.052!-0.01

   rad_turnover_int         = dble(radint)  !-11.3868
   rad_turnover_slope       = dble(radslp)  !0.0824

   vm_tran                  = 8.5
   vm_slop                  = 7.0
   vm_amp                   = 42.0
   vm_min                   = 18.0

   !---------------------------------------------------------------------------------------!
   !     This variable is the maximum distance between the coordinates of a prescribed     !
   ! phenology file and the actual polygon that we will still consider close enough to be  !
   ! representative.  If the user wants to run with prescribed phenology and the closest   !
   ! file is farther away from the polygon than the number below, the simulation will      !
   ! stop.                                                                                 !
   !---------------------------------------------------------------------------------------!
   max_phenology_dist       = 1.25 * erad * pio180
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_phen_coms
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!     This subroutine assigns the fusion and splitting parameters.                         !
!------------------------------------------------------------------------------------------!
subroutine init_ff_coms
   use fusion_fission_coms, only : niter_patfus       & ! intent(out)
                                 , hgt_class          & ! intent(out)
                                 , fusetol            & ! intent(out)
                                 , fusetol_h          & ! intent(out)
                                 , lai_fuse_tol       & ! intent(out)
                                 , lai_tol            & ! intent(out)
                                 , ff_nhgt            & ! intent(out)
                                 , coh_tolerance_max  & ! intent(out)
                                 , dark_cumlai_min    & ! intent(out)
                                 , dark_cumlai_max    & ! intent(out)
                                 , dark_cumlai_mult   & ! intent(out)
                                 , sunny_cumlai_min   & ! intent(out)
                                 , sunny_cumlai_max   & ! intent(out)
                                 , sunny_cumlai_mult  & ! intent(out)
                                 , light_toler_min    & ! intent(out)
                                 , light_toler_max    & ! intent(out)
                                 , light_toler_mult   & ! intent(out)
                                 , fuse_relax         & ! intent(out)
                                 , print_fuse_details & ! intent(out)
                                 , fuse_prefix        ! ! intent(out)
   use consts_coms        , only : onethird           & ! intent(out)
                                 , twothirds          ! ! intent(in)
   implicit none
   !----- Local variables. ----------------------------------------------------------------!
   real              :: exp_patfus
   real              :: exp_hgtclass
   !---------------------------------------------------------------------------------------!

   fusetol           = 0.4
   fusetol_h         = 0.5
   lai_fuse_tol      = 0.8
   lai_tol           = 1.0
   ff_nhgt           = 7
   coh_tolerance_max = 10.0    ! Original 2.0

   !----- Define the number of height classes. --------------------------------------------!
   allocate (hgt_class(ff_nhgt))
   hgt_class( 1) =  2.0
   hgt_class( 2) =  5.0
   hgt_class( 3) =  9.0
   hgt_class( 4) = 14.0
   hgt_class( 5) = 20.0
   hgt_class( 6) = 26.0
   hgt_class( 7) = 35.0

   niter_patfus       = 25
   exp_patfus         = 1. / real(niter_patfus)

   dark_cumlai_min    = 5.0
   dark_cumlai_max    = 7.0
   sunny_cumlai_min   = 0.5
   sunny_cumlai_max   = 1.0
   light_toler_min    = 0.20
   light_toler_max    = twothirds
   sunny_cumlai_mult  = (sunny_cumlai_max/sunny_cumlai_min)**exp_patfus
   dark_cumlai_mult   = (dark_cumlai_min /dark_cumlai_max )**exp_patfus
   light_toler_mult   = (light_toler_max /light_toler_min )**exp_patfus

   fuse_relax        = .false.

   !----- The following flag switches detailed debugging on. ------------------------------!
   print_fuse_details = .false.
   fuse_prefix        = 'patch_fusion_'
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_ff_coms
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
!    This subroutine assigns various parameters for the Runge-Kutta solver.  It uses many  !
! values previously assigned in other parameter initialisation, so this should be the last !
! one called.                                                                              !
!------------------------------------------------------------------------------------------!
subroutine init_rk4_params()
   use soil_coms      , only : water_stab_thresh      & ! intent(in)
                             , snowmin                & ! intent(in)
                             , tiny_sfcwater_mass     ! ! intent(in)
   use canopy_air_coms, only : leaf_drywhc            & ! intent(in)
                             , leaf_maxwhc            ! ! intent(in)
   use met_driver_coms, only : prss_min               & ! intent(in)
                             , prss_max               ! ! intent(in)
   use consts_coms    , only : wdnsi8                 ! ! intent(in)
   use rk4_coms       , only : rk4_tolerance          & ! intent(in)
                             , ibranch_thermo         & ! intent(in)
                             , maxstp                 & ! intent(out)
                             , rk4eps                 & ! intent(out)
                             , rk4eps2                & ! intent(out)
                             , rk4epsi                & ! intent(out)
                             , hmin                   & ! intent(out)
                             , print_diags            & ! intent(out)
                             , checkbudget            & ! intent(out)
                             , debug                  & ! intent(out)
                             , toocold                & ! intent(out)
                             , toohot                 & ! intent(out)
                             , lai_to_cover           & ! intent(out)
                             , rk4min_veg_temp        & ! intent(out)
                             , rk4water_stab_thresh   & ! intent(out)
                             , rk4tiny_sfcw_mass      & ! intent(out)
                             , rk4tiny_sfcw_depth     & ! intent(out)
                             , rk4leaf_drywhc         & ! intent(out)
                             , rk4leaf_maxwhc         & ! intent(out)
                             , rk4snowmin             & ! intent(out)
                             , rk4min_can_temp        & ! intent(out)
                             , rk4max_can_temp        & ! intent(out)
                             , rk4min_can_shv         & ! intent(out)
                             , rk4max_can_shv         & ! intent(out)
                             , rk4min_can_rvap        & ! intent(out)
                             , rk4max_can_rvap        & ! intent(out)
                             , rk4min_can_rhv         & ! intent(out)
                             , rk4max_can_rhv         & ! intent(out)
                             , rk4min_can_co2         & ! intent(out)
                             , rk4max_can_co2         & ! intent(out)
                             , rk4min_soil_temp       & ! intent(out)
                             , rk4max_soil_temp       & ! intent(out)
                             , rk4min_veg_temp        & ! intent(out)
                             , rk4max_veg_temp        & ! intent(out)
                             , rk4min_veg_lwater      & ! intent(out)
                             , rk4min_sfcw_temp       & ! intent(out)
                             , rk4max_sfcw_temp       & ! intent(out)
                             , rk4min_sfcw_moist      & ! intent(out)
                             , rk4min_virt_moist      & ! intent(out)
                             , effarea_heat           & ! intent(out)
                             , effarea_evap           & ! intent(out)
                             , effarea_transp         & ! intent(out)
                             , leaf_intercept         & ! intent(out)
                             , force_idealgas         & ! intent(out)
                             , supersat_ok            & ! intent(out)
                             , record_err             & ! intent(out)
                             , print_detailed         & ! intent(out)
                             , print_thbnd            & ! intent(out)
                             , errmax_fout            & ! intent(out)
                             , sanity_fout            & ! intent(out)
                             , thbnds_fout            & ! intent(out)
                             , detail_pref            ! ! intent(out)
   implicit none

   !---------------------------------------------------------------------------------------!
   !     Copying some variables to the Runge-Kutta counterpart (double precision).         !
   !---------------------------------------------------------------------------------------!
   rk4water_stab_thresh  = dble(water_stab_thresh )
   rk4tiny_sfcw_mass     = dble(tiny_sfcwater_mass)
   rk4leaf_drywhc        = dble(leaf_drywhc       )
   rk4leaf_maxwhc        = dble(leaf_maxwhc       )
   rk4snowmin            = dble(snowmin           )
   rk4tiny_sfcw_depth    = rk4tiny_sfcw_mass  * wdnsi8
   !---------------------------------------------------------------------------------------!

   !---------------------------------------------------------------------------------------!
   !    The following variables control the performance of the Runge-Kutta and Euler       !
   ! integration schemes. Think twice before changing them...                              !
   !---------------------------------------------------------------------------------------!
   maxstp      = 100000000           ! Maximum number of intermediate steps. 
   rk4eps      = dble(rk4_tolerance) ! The desired accuracy.
   rk4epsi     = 1.d0/rk4eps         ! The inverse of desired accuracy.
   rk4eps2     = rk4eps**2           ! square of the accuracy
   hmin        = 1.d-7               ! The minimum step size.
   print_diags = .false.             ! Flag to print the diagnostic check.
   checkbudget = .true.              ! Flag to check CO2, water, and energy budgets every 
                                     !     time step and stop the run in case any of these 
                                     !     budgets don't close.
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     Miscellaneous constants used in rk4_derivs.                                       !
   !---------------------------------------------------------------------------------------!
   debug         = .false.  ! Verbose output for debug                             [   T|F]
   toocold       = 1.5315d2 ! Minimum temperature for saturation specific hum.     [     K]
   toohot        = 3.5315d2 ! Maximum temperature for saturation specific hum.     [     K]
   lai_to_cover  = 1.5d0    ! Canopies with LAI less than this number  are assumed to be 
                            !     open, ie, some fraction of the rain-drops can reach
                            !    the soil/litter layer unimpeded.
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     Variables used to keep track on the error.                                        !
   !---------------------------------------------------------------------------------------!
   record_err     = .false.                  ! Compute and keep track of the errors.
   print_detailed = .false.                  ! Print detailed information about the thermo-
                                             !    dynamic state.  This will create one file
                                             !    for each patch, so it is not recommended 
                                             !    for simulations that span over one month.
   print_thbnd    = .false.                  ! Make a file with thermodynamic boundaries.
   errmax_fout    = 'error_max_count.txt'    ! File with the maximum error count 
   sanity_fout    = 'sanity_check_count.txt' ! File with the sanity check count
   thbnds_fout    = 'thermo_bounds.txt'      ! File with the thermodynamic boundaries.
   detail_pref    = 'thermo_state_'          ! Prefix for the detailed thermodynamic file
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     Assigning some default values for the bounds at the sanity check.  Units are      !
   ! usually the standard, but a few of them are defined differently so they can be scaled !
   ! depending on the cohort and soil grid definitions.                                    !
   !---------------------------------------------------------------------------------------!
   rk4min_can_temp   =  1.8400d2  ! Minimum canopy    temperature               [        K]
   rk4max_can_temp   =  3.5100d2  ! Maximum canopy    temperature               [        K]
   rk4min_can_shv    =  1.0000d-8 ! Minimum canopy    specific humidity         [kg/kg_air]
   rk4max_can_shv    =  10.200d-2  ! Maximum canopy    specific humidity         [kg/kg_air]  ! saturated air at 55 degC
   rk4max_can_rhv    =  1.1000d0  ! Maximum canopy    relative humidity (**)    [      ---]
   rk4min_can_co2    =  6.0000d1  ! Minimum canopy    CO2 mixing ratio          [ mol/mol]
   rk4max_can_co2    =  5.0000d4  ! Maximum canopy    CO2 mixing ratio          [ mol/mol]
   rk4min_soil_temp  =  1.8400d2  ! Minimum soil      temperature               [        K]
   rk4max_soil_temp  =  3.5100d2  ! Maximum soil      temperature               [        K]
   rk4min_veg_temp   =  1.8400d2  ! Minimum leaf      temperature               [        K]
   rk4max_veg_temp   =  3.5100d2  ! Maximum leaf      temperature               [        K]
   rk4min_sfcw_temp  =  1.9315d2  ! Minimum snow/pond temperature               [        K]
   rk4max_sfcw_temp  =  3.5100d2  ! Maximum snow/pond temperature               [        K]
   !.......................................................................................!
   ! (**) Please, don't be too strict here.  The model currently doesn't have radiation    !
   !      fog, so supersaturation may happen.  This is a problem we may want to address in !
   !      the future, though...                                                            !
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     Compute the minimum and maximum mixing ratio based on the specific humidity.      !
   !---------------------------------------------------------------------------------------!
   rk4min_can_rvap = rk4min_can_shv / (1.d0 - rk4min_can_shv)
   rk4max_can_rvap = rk4max_can_shv / (1.d0 - rk4max_can_shv)
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     Minimum water mass at the leaf surface.  This is given in kg/mleaf rather than   !
   ! kg/mground, so we scale it with LAI.                                                 !
   !---------------------------------------------------------------------------------------!
   rk4min_veg_lwater = -rk4leaf_drywhc            ! Minimum leaf water mass     [kg/mleaf]
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !    The minimum mass of surface water and virtual layer are given in m3/m3 rather than !
   ! kg/m2.  This is because there will be exchange between the top soil layer and the     !
   ! layers above in case the mass goes below the minimum.  Since this would make the im-  !
   ! pact of such exchange dependent on the soil depth, we assign the scale a function of  !
   ! the top layer thickness.                                                              !
   !---------------------------------------------------------------------------------------!
   rk4min_sfcw_moist = -5.0000d-4                  ! Minimum water mass allowed.
   rk4min_virt_moist = -5.0000d-4                  ! Minimum water allowed at virtual pool.
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     These variables are assigned in ed_params.f90.  Heat area should be 2.0 for all   !
   ! PFTs (two sides of the leaves exchange heat), and the evaporation area should be 1.0  !
   ! for all PFTs (only one side of the leaf is usually covered by water).  The transpir-  !
   ! ation area should be 1.0 for hypostomatous leaves, and 2.0 for symmetrical (pines)    !
   ! and amphistomatous (araucarias) leaves.  Sometimes heat and evaporation are multi-    !
   ! plied  by 1.2 and 2.2 to account for branches and twigs.  This is not recommended,    !
   ! though, because branches and twigs do not contribute to heat storage when             !
   ! ibranch_thermo is set to zero, and they are otherwise accounted through the wood area !
   ! index.                                                                                !
   !---------------------------------------------------------------------------------------!
   effarea_heat   = 2.d0 ! Heat area: related to 2*LAI
   effarea_evap   = 1.d0 ! Evaporation area: related to LAI
   !----- Transpiration.  Adjust them to 1 or 2 according to the type of leaf. ------------!
   effarea_transp(1:5)  = 1.d0 ! Hypostomatous
   effarea_transp(6:8)  = 2.d0 ! Symmetrical
   effarea_transp(9:16) = 1.d0 ! Hypostomatous
   effarea_transp(17)   = 2.d0 ! Amphistomatous
   effarea_transp(18:20)= 1.d0 ! Hypostomatous
   effarea_transp(21)   = 1.d0
   effarea_transp(22)   = 2.d0
   effarea_transp(23:35)= 1.d0
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !    This flag is used to control evaporation and transpiration when the air is         !
   ! saturated or super-saturated.  If supersat_ok is TRUE, then evaporation and           !
   ! transpiration will continue to happen even if the air is super-saturated at the       !
   ! canopy air temperature (but not at the soil or vegetation temperature).  Otherwise,   !
   ! evaporation and transpiration will be interrupted until the air becomes sub-saturated !
   ! again.  The air can still become super-saturated because mixing with the free atmo-   !
   ! sphere will not stop, but that is unlikely.                                           !
   !---------------------------------------------------------------------------------------!
   supersat_ok = .false.
   !---------------------------------------------------------------------------------------!


   !---------------------------------------------------------------------------------------!
   !     The integrator will adjust pressure every time step, including the internal ones, !
   ! to make sure the ideal gas is respected.  If set to false, it will keep pressure      !
   ! constant within on DTLSM time step, and not bother forcing the canopy air space to    !
   ! respect the ideal gas equation.                                                       !
   !---------------------------------------------------------------------------------------!
   force_idealgas = .false.
   !---------------------------------------------------------------------------------------!



   !---------------------------------------------------------------------------------------!
   !     This flag is to turn on and on the leaf interception.  Except for developer       !
   ! tests, this variable should be always true.                                           !
   !---------------------------------------------------------------------------------------!
   leaf_intercept = .true.
   !---------------------------------------------------------------------------------------!

   return
end subroutine init_rk4_params
!==========================================================================================!
!==========================================================================================!






!==========================================================================================!
!==========================================================================================!
subroutine overwrite_with_xml_config(thisnode)
   !!! PARSE XML FILE
   use ed_max_dims,   only: n_pft
   use ed_misc_coms,  only: iedcnfgf
   use hydrology_coms,only: useTOPMODEL
   use soil_coms,     only: isoilbc
   implicit none
   integer, intent(in) :: thisnode
   integer             :: max_pft_xml
   logical             :: iamhere

   if (iedcnfgf /= '') then
      inquire (file=iedcnfgf,exist=iamhere)
      if (iamhere) then

         !! FIRST, determine number of pft's defined in xml file
         call count_pft_xml_config(trim(iedcnfgf),max_pft_xml)
         if(max_pft_xml > n_pft) then

            write(unit=*,fmt='(a)') '*********************************************'
            write(unit=*,fmt='(a)') '**                                         **'
            write(unit=*,fmt='(a)') '**  Number of PFTs required by XML Config  **'
            write(unit=*,fmt='(a)') '**  exceeds the memory available           **'
            write(unit=*,fmt='(a)') '**                                         **'
            write(unit=*,fmt='(a)') '**  Please change n_pft in Module ed_max_dims **'
            write(unit=*,fmt='(a)') '**  and recompile                          **'
            write(unit=*,fmt='(a)') '**                                         **'
            write(unit=*,fmt='(a)') '*********************************************'
            call fatal_error('Too many PFTs','overwrite_with_xml_config','ed_params.f90')
         end if

         !! SECOND, update parameter defaults from XML
         call read_ed_xml_config(trim(iedcnfgf))

         !! THIRD, recalculate any derived values based on xml
         if(useTOPMODEL == 1) then
            isoilbc = 0
            print*,"TOPMODEL enabled, setting ISOILBC to 0"
         end if


         !! FINALLY, write out copy of settings
         call write_ed_xml_config()
   !      stop
      elseif (thisnode == 1) then
            write(unit=*,fmt='(a)') '*********************************************'
            write(unit=*,fmt='(a)') '**               WARNING!                  **'
            write(unit=*,fmt='(a)') '**                                         **'
            write(unit=*,fmt='(a)') '**    XML file wasn''t found. Using default **'
            write(unit=*,fmt='(a)') '** parameters in ED.                       **'
            write(unit=*,fmt='(a)') '** (You provided '//trim(iedcnfgf)//').'
            write(unit=*,fmt='(a)') '**                                         **'
            write(unit=*,fmt='(a)') '*********************************************'
      end if
   end if  !! end XML
   return
end subroutine overwrite_with_xml_config
!==========================================================================================!
!==========================================================================================!
